#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <nicklib.h>
#include <getpars.h>

#include "mcmcpars.h"  
#include "admutils.h"
#include "mcio.h"  
#include "egsubs.h"  

#define WVERSION   "410" 

// cdmin, gdmin added
// fmname, lmname added  (all pairs in region)
// maxdis added (really bugfix).  maxdis is in morgans
// Now uses egsubs.c 

#define MAXFL  50   
#define MAXSTR  512

char *trashdir = "/var/tmp" ;
int verbose = NO ;
int qtmode = NO ;
Indiv **indivmarkers;
SNP **snpmarkers ;
int numsnps, numindivs ; 
char *markername = NULL ; 
int markernum ;
SNP *xcupt ;

double plo = .001 ;
double phi = .999 ;

double reichthresh = -1.0 ;
// this is a threshold (p value) at 1 cM 
// we alter it linearly  so reichthresh of .002 at 1cM means thresh of .02 at .1 cM  
// We need to threshold in the code with a zscore, and our LRT is 2 sided 
// so we need to divide by 2 before calculating zthresh

char  *genotypename = NULL ;
char  *snpname = NULL ;
char  *snpoutfilename = NULL ;
char  *indivname = NULL ;
char *badsnpname = NULL ;

char *eglistname  = NULL ;
char *outputname = NULL ;
FILE *ofile ;
double cdmin = -1.0e6; 
double gdmin = -1.0e6; 

char *fmname = NULL ;
char *lmname = NULL ;
/** look at all markers between fmname, lmname */

char  unknowngender = 'U' ;
int   xchrom = -1 ;
int maxlag = 4 ;
double maxdis = 0.05 ; //morgans

void readcommands(int argc, char **argv) ;


int snpindx(char *snpid) ;

void cntitcase(double *xc, SNP *c1, SNP *c2) ;
void cntitx(double *xc, double *xd, SNP *c1, SNP *c2) ;
void cntitcasev(double *xc, SNP *c1, SNP *c2, Indiv **indm, int numind,  int val) ;

double lddip(double *xc) ;
double lddipx(double *xc, double *xd) ;

int main(int argc, char **argv)
{

  int **snppos ;
  int *snpindx ;
  char **snpnamelist, **indnamelist ;
  char **eglist ;
  int  lsnplist, lindlist, numeg ;
  int i,j; 
  SNP *cupt, *cupt1, *cupt2, *cupt3 ;
  Indiv *indx ;
  double gpos1,gpos2,cpos1,cpos2,gd, cd, gd100 ;
  double rthresh, zt ;
  int mpflag ;

  int ch1, ch2 ;
  int fmnum , lmnum ;
  int num, n1, n2 ;

  int nindiv = 0, e, f, lag=1  ;
  double xc[9], xd[4], xc2[9] ;
  double ychi, zscore, zthresh = 20.0 ;
  int nignore, numrisks = 1 ;

  readcommands(argc, argv) ;
  ofile = stdout; 
  if (outputname != NULL) openit(outputname, &ofile, "w") ;

  if (markername == NULL) fatalx("no markername\n") ;

  numsnps = 
    getsnps(snpname, &snpmarkers, 0.0, badsnpname, &nignore, numrisks) ;

  numindivs = getindivs(indivname, &indivmarkers) ;
  getgenos(genotypename, snpmarkers, indivmarkers, 
    numsnps, numindivs, nignore) ;

  printsnps(NULL, snpmarkers, numsnps, 
   indivmarkers, NO, NO) ;

  ZALLOC(eglist, numindivs, char *) ; 
  numeg = mkeglist(indivmarkers, numindivs, eglist) ;
  for (i=0; i<numeg; i++) {  
   printf("%3d %s\n",i, eglist[i]) ;
  }
  seteglist(indivmarkers, numindivs, eglistname);
  nindiv=0 ;
  for (i=0; i<numindivs; i++) {
    indx = indivmarkers[i] ;
    if (indx -> ignore) continue ; 
    if (eglistname == NULL) indx -> affstatus = YES ;
    if(indx -> affstatus == YES) ++nindiv  ;
  }
 printf("number of individuals in analysis %d\n", nindiv) ;
   markernum = snpindex(snpmarkers, numsnps, markername) ;
   if (markernum<0) fatalx("markername %s not found\n",markername) ;
    xcupt = snpmarkers[markernum] ;
    if (xchrom < 0) xchrom = xcupt -> chrom ;  
    if (xchrom == 23) fatalx("X chrom not implemented\n") ;

 for (i=0; i< numsnps; i++)  {  
  cupt = snpmarkers[i] ;
  if (cupt -> chrom != xchrom) continue ;
  gpos1 = cupt -> genpos ;              
  cpos1 = cupt -> physpos ;              
  fprintf(ofile, "%15s %9.3f %12.0f ", cupt -> ID, gpos1, cpos1) ;
  if (i == markernum) { 
   fprintf(ofile,  "%9s\n", "+++++++") ;
   continue ;  
  }
   cntitcase(xc, cupt, xcupt) ;
   ychi = lddip(xc) ;
   fprintf(ofile,  "%9.3f\n", ychi) ;
 }
 printf("## end of run\n") ;
}

void cntitcase(double *xc, SNP *c1, SNP *c2) 
{
 int n, i, e, f ;
 Indiv *indx ;
 n = MIN(c1->ngtypes, c2->ngtypes) ;
 vzero(xc, 9) ;
 for (i=0; i<n; i++) {  
  indx = indivmarkers[i] ;
  if (indx -> ignore) continue ;
  if (indx->affstatus != YES) continue ;
  e = getgtypes(c1, i) ;
  f = getgtypes(c2, i) ;
  if (e<0) continue ;
  if (f<0) continue ;
  ++xc[3*e+f] ;
 }
}

void cntitx(double *xc, double *xd, SNP *c1, SNP *c2) 
{
 int n, i, e, f ;
 Indiv *indx ;
 n = MIN(c1->ngtypes, c2->ngtypes) ;
 vzero(xc, 9) ;
 vzero(xd, 9) ;
 for (i=0; i<n; i++) {  
  indx = indivmarkers[i] ;
  if (indx->affstatus != YES) continue ;
  if (indx->gender == 'F') {
  e = getgtypes(c1, i) ;
  f = getgtypes(c2, i) ;
   if (e<0) continue ;
   if (f<0) continue ;
   ++xc[3*e+f] ;
  }
  if (indx->gender == 'M') {
   e = getgtypes(c1, i) ;
   f = getgtypes(c2, i) ;
   if (e<0) continue ;
   if (f<0) continue ;
   e = e/2 ;
   f = f/2 ;
   ++xd[2*e+f] ;
  }
 }
}

void checkgeno() 
{
  int i, k, x ;
  int n, nbad, ngood ;
  SNP *cupt ;

  nbad = ngood = 0 ;
  for (i=0; i<numsnps; i++) {  
   cupt = snpmarkers[i] ;
   if (cupt -> isfake) continue ;
   if (cupt -> ignore) continue ;
   n = cupt -> ngtypes ;
   if (n<=0) continue ;
   for (k=0; k<n; k++) {
    x = getgtypes(cupt, k) ;
    if (x>2) {  
     printf("bad genotype: %s %4d %4d\n", cupt->ID, k,x) ;
     putgtypes(cupt, k, -1) ;
     ++nbad ;
    }
    else  {
     ++ngood ;
    }
   }
  }
  printf("checkgeno. good: %d  bad:  %d\n",ngood, nbad) ;
/**
  if (nbad>0) fatalx("checkgeno fails\n") ;
*/
}

void
countgtypes(SNP *cupt, int *xcnt)  
{
   int k, numstates, ignore, refcnt, varcnt, a ;
   int maxcnt ;
   Indiv *indx ;

   ivzero(xcnt, 4) ;
   if (cupt -> isfake) return ;
   if (cupt -> ignore) return ;
   if (cupt -> gtypes == NULL) return ;
   if (cupt -> ngtypes == 0) return ;
   for (k=0; k< numindivs; ++k) {  
    indx = indivmarkers[k];
    gettln(cupt, indx, NULL, NULL, &numstates, &ignore) ;
    if (ignore == YES) continue ;
    a = 0 ;
    if (indx -> affstatus) a = 2 ;  
    refcnt = getgtypes(cupt, k) ;
    if (refcnt<0) continue ;
    maxcnt = 2  ;
    if (numstates==2) {  
     maxcnt = 1 ;
     refcnt /= 2 ;
    }
    varcnt = maxcnt-refcnt ;
    xcnt[a] += refcnt ;
    xcnt[a+1] += varcnt ;
   }
}

void readcommands(int argc, char **argv) 

{
  int i,haploid=0;
  char *parname = NULL ;
  phandle *ph ;
  char str[5000]  ;
  char *tempname ;
  int n ;

  while ((i = getopt (argc, argv, "p:vV")) != -1) {

    switch (i)
      {

      case 'p':
	parname = strdup(optarg) ;
	break;

      case 'v':
	printf("version: %s\n", WVERSION) ; 
	break; 

      case 'V':
	verbose = YES ;
	break; 

      case '?':
	printf ("Usage: bad params.... \n") ;
	fatalx("bad params\n") ;
      }
  }

         
   pcheck(parname,'p') ;
   printf("parameter file: %s\n", parname) ;
   ph = openpars(parname) ;
   dostrsub(ph) ;

/**
DIR2:  /fg/nfiles/admixdata/ms2
SSSS:  DIR2/outfiles 
genotypename: DIR2/autos_ccshad_fakes
eglistname:    DIR2/eurlist  
output:        eurout
*/

   getstring(ph, "genotypename:", &genotypename) ;
   getstring(ph, "snpname:", &snpname) ;
   getstring(ph, "indivname:", &indivname) ;
   getstring(ph, "eglistname:", &eglistname) ;
   getstring(ph, "output:", &outputname) ;
   getint(ph, "chrom:", &xchrom) ;
   getstring(ph, "badsnpname:", &badsnpname) ;
   getstring(ph, "markername:", &markername) ;

}

void setfliplist(char *flistname, char **flist, int *numfl)
{
  FILE *fgfile ;
  char line[MAXSTR] ;
  char *spt[MAXFF] ;
  char *sx ;
  int nsplit, i, num=0 ;

  *numfl = 0; 
  if (flistname == NULL) return ;
  openit(flistname, &fgfile, "r") ;
  while (fgets(line, MAXSTR, fgfile) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   sx = spt[0] ;
   if (sx[0] == '#') continue ;
   if (num>=MAXFL) 
  fatalx("too many flips\n") ;
   flist[num] = strdup(sx) ;
   ++num ;
   freeup(spt, nsplit) ;
  }
  fclose(fgfile) ;
  *numfl = num ;
}



int snpindx(char *snpid) 
/* hash table would be good here */
{
     int k ; 
     for (k=0; k< numsnps; k++) {  
      if (strcmp(snpmarkers[k] -> ID, snpid) == 0) return k ;
     }
     return -1 ;
}

void flipallele(SNP *cupt) 
// flip reference, variant counts
{
   int i, n, g ;
   n = cupt -> ngtypes ;
   for (i=0; i<n; i++) {  
    g = cupt -> gtypes[i] ;
    if (g<0) continue ;
    cupt ->gtypes[i] = 2-g ;
   }
}
void cntitcasev(double *xc, SNP *c1, SNP *c2, Indiv **indm, int numind,  int val) 
{
 int n, i, e, f ;
 Indiv *indx ;
 n = MIN(c1->ngtypes, c2->ngtypes) ;
 vzero(xc, 9) ;
 for (i=0; i<numind; i++) {  
  indx = indm[i] ;
  if (indx->affstatus != val) continue ;
  if (indx-> ignore) continue ;
  e = getgtypes(c1, i) ;
  f = getgtypes(c2, i) ;
  if (e<0) continue ;
  if (f<0) continue ;
  ++xc[3*e+f] ;
 }
}

