#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <nicklib.h>
#include <getpars.h>

#include "admutils.h"
#include "mcio.h"  

#define WVERSION   "2.00" 

#define MAXFL  50   
#define MAXSTR  512

char *trashdir = "/var/tmp" ;
int verbose = NO ;
Indiv **indivmarkers;
SNP **snpmarkers ;
int numsnps, numindivs ; 


double plo = .001 ;
double phi = .999 ;

char  *genotypename = NULL ;
char  *snpname = NULL ;
char  *snpoutfilename = NULL ;
char  *indivname = NULL ;
char *eglistname  = NULL ;
char *outputname = NULL ;
FILE *ofile ;


void readcommands(int argc, char **argv) ;

int mkeglist(char **eglist);
void seteglist(char *eglistname, char **eglist, int numeg, int val) ;

void cntit(double *xc, SNP *c1, SNP *c2) ;

double lddip(double *xc) ;
int countgt(SNP *cupt, int val, int *x0, int *x1) ;

int main(int argc, char **argv)
{

  int **snppos ;
  int *snpindx ;
  char **snpnamelist, **indnamelist ;
  char **eglist ;
  char **aglist ;
  int  lsnplist, lindlist, numeg, numag ;
  int i,j,k; 
  SNP *cupt, *cupt1, *cupt2 ;
  Indiv *indx ;
  char str[MAXSTR] ;

  int nindiv = 0, e, f, lag=1  ;
  int n, n0, n1, g, chrom, gender ;
  int ec0, ec1, ac0, ac1, num1, num2 ;
  double xc[9], xd[4], xc2[9] ;
  double cc[4] ;
  double ychi ;
  char *fliplist[MAXFL] ;
  int numfl ;
  int nignore, numrisks = 1 ;

  readcommands(argc, argv) ;

  numsnps = 
    getsnps(snpname, &snpmarkers, 0.0, NULL, &nignore, numrisks) ;

  numindivs = getindivs(indivname, &indivmarkers) ;

  getgenos(genotypename, snpmarkers, indivmarkers, 
    numsnps, numindivs, nignore) ;

  setstatus(indivmarkers, numindivs, "Case") ;

  printf("##numsnps: %d\n", numsnps) ;
  printf("##numindivs: %d\n", numindivs) ;

  ofile = stdout ;
  if (outputname != NULL) openit(outputname, &ofile, "w") ;

  for (j=0; j<numsnps; j++) {  

   cupt = snpmarkers[j] ;
   chrom = cupt -> chrom ;

   num1 = countgt(cupt, YES, &ec0, &ec1) ;
   num2 = countgt(cupt, NO, &ac0, &ac1) ;

   if (num1==0) continue ;
   if (num2==0) continue ;

   cc[0] = ec0 ; cc[1] = ec1 ;
   cc[2] = ac0 ; cc[3] = ac1 ;
   vsp(cc, cc, .001, 4) ;
   ychi = conchi(cc, 2, 2) ;
   fprintf(ofile, "%25s %3d", cupt->ID, chrom) ;
   fprintf(ofile, " %5d %5d %9.3f", num1, num2, ychi) ;
   fprintf(ofile, "\n") ;
  }
  return 0 ;
  if (outputname != NULL) fclose(ofile) ;
}

int countgt(SNP *cupt, int val, int *x0, int *x1)
{
   int c0, c1, n, i, num ; 
   int n0, n1, g ;
   int chrom ;
   Indiv *indx ;

   chrom = cupt -> chrom ;
   n = cupt -> ngtypes ;
   num = c0 = c1 = 0 ;
   for (i=0; i<n; i++) {
    indx = indivmarkers[i] ;

    g = cupt->gtypes[i] ;
    if (g<0) continue ;
    n0 = g ;  
    n1 = 2-g ;
    if (indx -> affstatus != val) {  
     continue ;
    }
    if ((chrom==23) && ((indx->gender=='M'))) {  
     n0 /= 2 ;
     n1 /= 2 ;
    }
    c0 += n0 ;
    c1 += n1 ;
    ++num ;
   }

   *x0 = c0 ; 
   *x1 = c1 ;
   return num;
}

int mkeglist(char **eglist)
{
  Indiv *indx ;
  int i, k, numeg=0 ;
  for (i=0; i<numindivs; i++) { 
   indx = indivmarkers[i] ;
    k = indxindex(eglist, numeg,  indx->egroup) ;
    if (k<0)   { 
     eglist[numeg] = strdup(indx->egroup) ;
     ++numeg ;
    }
  }
  return numeg ;
}

void checkgeno() 
{
  int i, k, x ;
  int n, nbad, ngood ;
  SNP *cupt ;

  nbad = ngood = 0 ;
  for (i=0; i<numsnps; i++) {  
   cupt = snpmarkers[i] ;
   if (cupt -> isfake) continue ;
   if (cupt -> ignore) continue ;
   n = cupt -> ngtypes ;
   if (n<=0) continue ;
   for (k=0; k<n; k++) {
    x = cupt -> gtypes[k] ;
    if (x>2) {  
     printf("bad genotype: %s %4d %4d\n", cupt->ID, k,x) ;
     cupt -> gtypes[k]  = -1 ;
     ++nbad ;
    }
    else  {
     ++ngood ;
    }
   }
  }
  printf("checkgeno. good: %d  bad:  %d\n",ngood, nbad) ;
/**
  if (nbad>0) fatalx("checkgeno fails\n") ;
*/
}

void
countgtypes(SNP *cupt, int *xcnt)  
{
   int k, numstates, ignore, refcnt, varcnt, a ;
   int maxcnt ;
   Indiv *indx ;

   ivzero(xcnt, 4) ;
   if (cupt -> isfake) return ;
   if (cupt -> ignore) return ;
   if (cupt -> gtypes == NULL) return ;
   if (cupt -> ngtypes == 0) return ;
   for (k=0; k< numindivs; ++k) {  
    indx = indivmarkers[k] ;
    gettln(cupt, indx, NULL, NULL, &numstates, &ignore) ;
    if (ignore == YES) continue ;
    a = 0 ;
    if (indx -> affstatus) a = 2 ;  
    refcnt = cupt -> gtypes[k] ;
    if (refcnt<0) continue ;
    maxcnt = 2  ;
    if (numstates==2) {  
     maxcnt = 1 ;
     refcnt /= 2 ;
    }
    varcnt = maxcnt-refcnt ;
    xcnt[a] += refcnt ;
    xcnt[a+1] += varcnt ;
   }
}

void readcommands(int argc, char **argv) 

{
  int i,haploid=0;
  char *parname = NULL ;
  phandle *ph ;
  char str[5000]  ;
  char *tempname ;
  int n ;

  while ((i = getopt (argc, argv, "p:vV")) != -1) {

    switch (i)
      {

      case 'p':
	parname = strdup(optarg) ;
	break;

      case 'v':
	printf("version: %s\n", WVERSION) ; 
	break; 

      case 'V':
	verbose = YES ;
	break; 

      case '?':
	printf ("Usage: bad params.... \n") ;
	fatalx("bad params\n") ;
      }
  }

         
   pcheck(parname,'p') ;
   printf("parameter file: %s\n", parname) ;
   ph = openpars(parname) ;
   dostrsub(ph) ;

   getstring(ph, "genotypename:", &genotypename) ;
   getstring(ph, "snpname:", &snpname) ;
   getstring(ph, "indivname:", &indivname) ;
   getstring(ph, "output:", &outputname) ;

}

void setfliplist(char *flistname, char **flist, int *numfl)
{
  FILE *fgfile ;
  char line[MAXSTR] ;
  char *spt[MAXFF] ;
  char *sx ;
  int nsplit, i, num=0 ;

  *numfl = 0; 
  if (flistname == NULL) return ;
  openit(flistname, &fgfile, "r") ;
  while (fgets(line, MAXSTR, fgfile) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   sx = spt[0] ;
   if (sx[0] == '#') continue ;
   if (num>=MAXFL) 
  fatalx("too many flips\n") ;
   flist[num] = strdup(sx) ;
   ++num ;
   freeup(spt, nsplit) ;
  }
  fclose(fgfile) ;
  *numfl = num ;
}

void seteglist(char *eglistname, char **eglist, int numeg, int val)
{
  FILE *egfile ;
  char line[MAXSTR] ;
  char *spt[MAXFF] ;
  char *sx ;
  Indiv *indx ;
  int nsplit, i ;

  if (eglistname == NULL) return ;
  openit(eglistname, &egfile, "r") ;
  while (fgets(line, MAXSTR, egfile) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   sx = spt[0] ;
   for (i=0; i<numindivs; i++) {
    indx = indivmarkers[i] ;
    if (strcmp(indx->egroup, sx) == 0)
     indx -> affstatus = val ;
   }
   freeup(spt, nsplit) ;
  }
  fclose(egfile) ;
}


void flipallele(SNP *cupt) 
// flip reference, variant counts
{
   int i, n, g ;
   n = cupt -> ngtypes ;
   for (i=0; i<n; i++) {  
    g = cupt -> gtypes[i] ;
    if (g<0) continue ;
    cupt ->gtypes[i] = 2-g ;
   }
}
