#include "mcsubs.h" 
#include "gammas.h"

extern char *gammafiles ; 
extern int keeptrashfiles ;
extern int magicnum ;

static   int  f0, f1, f2 = -1 ; /* f descriptors */
static   int  a0, aa1, a2 = -1 ; /* f descriptors */
static   int  b0, bb1, b2 = -1 ; /* f descriptors */


static char *alp0name ;
static char *alp1name ;
static char *alp2name ;
static char *bet0name ;
static char *bet1name ;
static char *bet2name ;
static char *gam0name ;
static char *gam1name ;
static char *gam2name ;

static char *ffname[9] ;
static int ffdes[9] ;

static SNP **snpx  = NULL;

static int nums, numind ;

void dumpgams(SNP **snpmarkers, int numsnps, int numindivs, char *scratchdir)
{
    double scbeta, scalpha ;
    double theta, lambda ;
    SNP *cupt ;
    Indiv *indx ;
    double **betas ;
    double **alphas ;
    int indiv, i, j, k, kk, chrom, fc, lc, numstates, ignore ;
    int len, ind ;
    int fdes, t, jj, ret ;
    double **abg, gg[3], y, *xx ;
    char str[512], *tname, *fbase ;
    int xdebug ;

   printf("dumpgams called %d %d\n", numsnps, numindivs) ;
   if (snpx == NULL) snpx =  snpmarkers ;

   betas = initarray_2Ddouble(numsnps, 3, 1.0);
   alphas = initarray_2Ddouble(numsnps, 3, 0.0);

   makedir(scratchdir) ;  // No op if scratchdir exists 

   setgfiles(snpmarkers, numsnps, numindivs, scratchdir, gammafiles, 0) ;

/**
 loop on chromosome
 compute alpha beta gamma 
 tilt. 
 By computing 1 chromosome at a time 
 we reduce memory requrements here 
*/
 
   for (chrom = 1; chrom <= 23 ; ++chrom) {
    fc = fchrom[chrom] ;  lc = lchrom[chrom] ;
    if (lc < 0) continue ;
    len = lc-fc+1 ;
    if (len <= 0) continue ;
    abg = initarray_2Ddouble(9, len*numindivs, 0.0);
    cupt = snpmarkers[fc] ;

    for (indiv = 0; indiv < numindivs; ++indiv) { 

     indx =  indivmarkers[indiv] ;
     gettln(cupt, indx, &theta, &lambda, &numstates, &ignore) ;

     scbeta = HMMbeta(snpmarkers,indiv, theta, lambda, fc, lc,   
      betas, NULL, numstates, NULL) ;

     xdebug = debug ;
     if ((indiv== -1) & (chrom==23))  {  
       debug = YES ;
     }
     else debug = NO ;

     scalpha  = HMMalpha(snpmarkers,indiv, theta, lambda, fc, lc,   
      alphas, NULL, numstates, NULL) ;

     if (debug) {   
      printf("zz1 %d %d %9.3f %9.3f\n", numstates, ignore, theta, lambda) ;
      jj = 0 ; 
      xx = alphas[jj] ;
      printmat(xx, 1, 3) ;
      abort() ;  
     }
    
     debug = xdebug ;
      
     for (kk= fc; kk<= lc; ++kk) {  
      jj = kk-fc ;
      gg[2] = 0.0 ;

      xx = alphas[jj] ;
      y = asum(xx, numstates) ;
      if (y>0.0) vst(xx, xx, 1.0/y, numstates) ;

      xx = betas[jj] ;
      y = asum(xx, numstates) ;
      if (y>0.0) vst(xx, xx, 1.0/y, numstates) ;

      vvt(gg, alphas[jj], betas[jj], numstates) ;
      y = asum(gg, 3) ;
      if (y>0.0) vst(gg, gg, 1.0/y, 3) ;

      ind = jj*numindivs + indiv ; 
      if ( (chrom==1) && (jj==-1) && (indiv<5)) { 
        printf("zzq theta: %15.9f\n", theta) ;
        printf("zzq0 %d %d %15.3e %15.3e %15.3e\n", 
         jj, indiv, 
         alphas[jj][0], betas[jj][0], gg[0]) ;

        printf("zzq1 %d %d %15.3e %15.3e %15.3e\n", 
         jj, indiv, 
         alphas[jj][1], betas[jj][1], gg[1]) ;

        printf("zzq2 %d %d %15.3e %15.3e %15.3e\n", 
         jj, indiv, 
         alphas[jj][2], betas[jj][2], gg[2]) ;
      }
      for (j=0; j<=2; j++) { 
       abg[j] [ind]   = alphas[jj][j] ;
       abg[j+3] [ind] = betas[jj][j] ;
       abg[j+6] [ind] = gg[j] ;
      }
     }
    }

    for (k=0; k<9; k++) {
     fdes = ffdes[k] ;
     ret = write(fdes, abg[k], numindivs * len * sizeof (double) ) ;
     if (ret<0) {  
      perror("write failure") ;
      fatalx("bad write: chrom: %d numindivs: %d len: %d\n", chrom, numindivs, len) ;
     } 
    }
    free2D(&abg, 9) ;
   }

  free2D(&alphas, numsnps) ;
  free2D(&betas, numsnps) ;

 for (k=0; k<9; k++) {
   tname = ffname[k] ;
   fdes = ffdes[k] ;
   close(fdes) ;
   t = chmod(tname, 0777) ;
   if (t<0) fatalx("bad chmod %s\n", tname) ;
   ffdes[k] = fdes = open(tname, O_RDONLY) ;
   if (fdes<0) {
    perror("dumpgams") ;
    fatalx("open failed (on read) for %s\n", str) ;
   }
 }


}

void unlinktemp() 

/** 
 destructor for all temporaries 
*/

{

  int i ;

  if (keeptrashfiles) {
   printf("trash files kept\n") ;
   return ;
  }

  for (i=0; i<9; i++)  {
   unlink(ffname[i]) ;
  }
  printf("trash files deleted\n") ;
}


void floadgams(int fd, int markernum, double *gammas, 
   int *numindivs, SNP *marker)
{
  int i,t,numcounted,numloaded=0;

  numloaded = *numindivs ;
  marker = snpmarkers[markernum] ;
  if (fd<0) fatalx("(floadgams) fd not set\n") ;
  t = lseek(fd, numloaded*markernum*sizeof(double), SEEK_SET) ;
  if (t<0) {
    perror("floadgams") ;
    fatalx("bad seek for record %d\n", markernum) ;
  }
  t = read(fd, gammas, numloaded * sizeof (double) ) ;
  if (t<0) {
    perror("floadgams") ;
    fatalx("bad read for record %d\n", markernum) ;
  }
  
}

void setgfiles(SNP **snpm, int numsnps, int numindivs, char *scdir, char *gammfiles, int mode) 
/** 
 mode 0  write 
 mode 1  read 
*/
{
   int fdes, t, k, pid ;
   char str[512], *tname, *fbase ;

   snpx = snpm ; 
   nums = numsnps ;
   numind = numindivs ;

   if (gammfiles == NULL) {  
     sprintf(str, "ffname:%d", pid) ;
     fbase = strdup(str) ;
   }
   else fbase = gammfiles ;
   

   for (k=0; k<9; k++) {  
    sprintf(str, "%s/%s.%d", scdir, fbase, k) ;
    ffname[k] = strdup(str) ;
   }

   alp0name = ffname[0] ;
   alp1name = ffname[1] ;
   alp2name = ffname[2] ;
   bet0name = ffname[3] ;
   bet1name = ffname[4] ;
   bet2name = ffname[5] ;
   gam0name = ffname[6] ;
   gam1name = ffname[7] ;
   gam2name = ffname[8] ;

   if (gammafiles == NULL) {  
     free(fbase)  ;
   }


  for (k=0; k<9; k++) {
   
   tname = ffname[k] ;

   if (mode==0) {
    ridfile(tname) ;
    ffdes[k] = fdes = open(tname, O_CREAT | O_TRUNC | O_RDWR, 006);
   }

   if (mode==1) {
    ffdes[k] = fdes = open(tname, O_RDONLY) ;
   }
 
   if (fdes<0) {
    perror("dumpgams") ;
    fatalx("open failed for %s\n", tname) ;
   }
   if (verbose) 
    printf("file %s opened\n", tname) ;
  }
/* 
 assign symbolic names 
*/
 a0 = ffdes[0] ;
 aa1 = ffdes[1] ;
 a2 = ffdes[2] ;
 b0 = ffdes[3] ;
 bb1 = ffdes[4] ;
 b2 = ffdes[5] ;
 f0 = ffdes[6] ;
 f1 = ffdes[7] ;
 f2 = ffdes[8] ;
 
}


void gammastar(double *xx, int j, int i, int numstates) 
{
/* gamma(i,a,-j) 
  as in checking an admixture map
  j isd markernum 
*/
 double gg[3], X[3], y ;

 tcalcX(snpx[j], i, X, numstates) ;
 y = asum(X, numstates) ;
 if (y<=0.0) {  
   vclear(gg, 1.0, numstates) ;
 }
 else {
  grabgam(gg, j, i, 2) ; 
  vsp(X,X,1.0e-6,numstates) ; /* add epsilon */
  vvd(gg, gg, X, numstates) ;  
 }
 y = asum(gg, numstates) ; 
 if (y<=0.0) vclear(xx, -1.0, numstates) ;
 else vst(xx, gg, 1.0/y, numstates) ;
}

void alphastar(double *xx, int j, int i, int numstates) 
{
/* alphastar(i,a) 
  as in checking an admixture map
*/
 double gg[3], X[3], y ;

 vzero(xx, 3) ;
 tcalcX(snpx[j], i, X, numstates) ;

 y = asum(X, numstates) ;
 if (y<=0.0) {  
   vclear(gg, 1.0, numstates) ;
 }
 else {
  grabgam(gg, j, i, 0) ;  /* alpha */
  vsp(X,X,1.0e-6,numstates) ; /* add epsilon */
  vvd(gg, gg, X, numstates) ;  
 }
 y = asum(gg, numstates) ; 
 if (y<=0.0) vclear(xx, -1.0, numstates) ;
 else vst(xx, gg, 1.0/y, numstates) ;
}

void betax(double *xx, int j, int i, int numstates) 
{
/* 
  just grabs beta vector (normalized) 
*/
 double gg[3], X[3], y ;

 vzero(xx, 3) ;
 grabgam(gg, j, i, 1) ;  /* beta */
 y = asum(gg, numstates) ; 
 if (y<=0.0) vclear(xx, -1.0, numstates) ;
 else vst(xx, gg, 1.0/y, numstates) ;
}


void grabgam(double *vv, int markernum, int indiv, int mode) 
/** 
 vv[i] is gamma (alpha, beta) for markernum and indiv 
*/

{
/** 
 a buffered version 
 should be reasonably efficient in 
 practice 
*/
#define NBUFF 12 
   int fdes[3]  ;
   int i, k, kind, itag ;
   double *tt ;
   SNP *cupt ;
   static double *buff[NBUFF] ;
   static int kbuff[NBUFF] ;
   static int fcall = 0, knext ;

   if (fcall==0) { 
    for (i=0; i< NBUFF; i++)  {
     ZALLOC(buff[i], numind, double) ;
     kbuff[i] = -1 ; /* empty */
    }
    knext = 0 ;
    fcall = 1 ; 
   }

   switch (mode)  {
    case 0:  /* alpha */
     fdes[0] = a0; fdes[1] = aa1; fdes[2] = a2 ;
     break ;
    case 1:  /* beta */
     fdes[0] = b0; fdes[1] = bb1; fdes[2] = b2 ;
     break ;
    case 2:  /* gamma */
     fdes[0] = f0; fdes[1] = f1; fdes[2] = f2 ;
     break ;
    default:  fatalx("(grabgam) bad mode %d\n", mode) ;
   }
/* tag = (markernum, fdindex,  mode)  */
  cupt = snpmarkers[markernum] ;
  for (k=0; k<3; k++)  {
   itag = markernum ;
   itag = 3*itag + k ;
   itag = 3*itag + mode ;
   if (itag<0) fatalx("(grabgam) overflow\n") ;
   kind = findfirst(kbuff, NBUFF, itag) ;  
   if (kind<0) {  
     kind = knext ;
     ++knext ;
     knext %= NBUFF ;
     tt = buff[kind] ;
     floadgams(fdes[k],markernum,tt,&numind,cupt);
     kbuff[kind] = itag ;
   }
   tt = buff[kind] ;
   vv[k] = tt[indiv] ;
  }
}
void getfd (int *x0, int *x1, int *x2)  
{

  *x0 = f0 ;
  *x1 = f1 ;
  *x2 = f2 ;

}

void tcalcX(SNP *cupt, int i, double *X, int numstates)
/** 
 i is indiv number 
*/
{
  int x,k;
  double PopA_freq, PopB_freq;

  if ((cupt -> markernum) == magicnum) {  
    magicgg(X, i) ;
    return ;
  }

  if ((cupt -> isfake) || (cupt->ignore) || (cupt -> isrfake)) { 
   vclear(X, 1.0, numstates) ;
   return ;
  }

  if (cupt -> ngtypes == 0) {  
   vclear(X, 1.0, numstates) ; 
   return ; 
  } 

  x = getgtypes(cupt, i) ; 
  if ((x==1) && (numstates==2)) x = -1 ;

  if (x<0) {  
   vclear(X, 1.0, numstates) ;
   return ; 
  }

  PopA_freq = cupt -> adm_PopA_freq;
  PopB_freq = cupt -> adm_PopB_freq;

  if ((PopA_freq<0) || (PopB_freq<0)) 
    fatalx("marker freqs %s not set\n", cupt -> ID) ;

  if (numstates == 3) {
    switch(x) {
      
    case 2:
      X[0] = PopA_freq * PopA_freq;
      X[1] = PopA_freq * PopB_freq;
      X[2] = PopB_freq * PopB_freq;
      break;
      
    case 1:
      X[0] = 2 * PopA_freq * (1 - PopA_freq) ;
      X[1] = PopB_freq * (1 - PopA_freq) + 
	(1 - PopB_freq) * PopA_freq ;
      X[2] =  2 * PopB_freq * (1 - PopB_freq) ;
      break;
      
    case 0:
      X[0] = (1 - PopA_freq) * (1 - PopA_freq);
      X[1] = (1 - PopA_freq) * (1 - PopB_freq);
      X[2] = (1 - PopB_freq) * (1 - PopB_freq);
      break;
  
    default:
      fatalx(" %s %d warning genotype unset\n", cupt->ID, i) ;
      break;
    }
  }
  else if (numstates == 2) {
    switch(x) {

    case 2:
      X[0] = PopA_freq;
      X[1] = PopB_freq;
      break;

    case 0:
      X[0] = (1 - PopA_freq);
      X[1] = (1 - PopB_freq);
      break;
  
    default:
      fatalx(" %s %d warning genotype unset\n", cupt->ID, i) ;
      break;
    }
  }
}


double HMMbeta
 (
 SNP **tsnpm,
 int indiv,
 double theta, 
 double lambda, 
 int first, 
 int last, 
 double **betas,
 double *binit, 
 int numstates, double *scale )

/* returns log10 score */

/** 
 dangerous bend.  It is betas[count][k] 
 but in HMMwindow gammas[k][count] 

 we maybe should fix this
*/


{

  int j,k,x,y,count ;

  double  *betaexp ;
  double poisson, temp,  *beta_star, tsum, glen;
  double loglike ;
  double www[3], ta[3], tb[3], y1, y2 ; /* debug */
  double ttt[3]  ;
  register double yy, yyp ;
  int nummarkers ;
  int dowork = YES ;  /* no we just skip */
  int numsta = 3 ;
  int ignore, chrom ;

  double X[3], **M ;
  double *bc ;
  double *aaa, *bbb ;
  SNP *cupt ;
  Indiv *indx ;

  count = nummarkers = last-first+1 ;

  cupt = tsnpm[first] ;
  chrom = cupt -> chrom ;
  indx = indivmarkers[indiv] ;
  ignore = indx -> ignore ;


  if (ignore == YES) dowork = NO ;

  if ((binit != NULL) && (binit[0] < 0.0))   dowork = NO ;

  if (dowork == NO)        {
   for (k=0; k<count; k++)  {  
    vclear(betas[k], -1.0, numstates) ;
   }
   return 0.0 ;
  }

  vzero(ttt, 3) ;
  sett1(ttt, theta, numstates) ;

  ZALLOC(betaexp,count,double);
  ZALLOC(beta_star,numsta,double);

  M = initarray_2Ddouble(numsta,numsta,0.0);

  /* ###############  Beta Pass ################# */

  
  for (j=last; j>=first; --j) {

    count = j-first ;
    if (j == last) {
      if (binit != NULL)  {
       copyarr(binit, betas[count], numstates) ;
      }
      else  {
       vclear(betas[count], 1.0, numstates) ;
      }
      betaexp[count]=0.0;

    }
    else {
       glen = (snpmarkers[j+1] -> genpos) - (snpmarkers[j] -> genpos);
/* this is in morgans */ 
       temp = lambda * glen ;
       if (temp<20.0) {
        poisson = exp(-temp) ;
       }
       else poisson = 0.0 ;
       if (lambda>500) poisson = 0.0 ;

       calcM(poisson, theta, M, numstates) ;
  
      bc = betas[count] ;

      vzero(bc, numstates) ;
      for (x=0; x<numstates; x++) {
	for (y=0; y<numstates; y++) {
	  bc[x] += M[y][x] * beta_star[y];
	}
      }
      betaexp[count] = betaexp[count+1];
    }

    tsum = asum(betas[count], numstates) ;

    if (tsum <= 0.0) 
      fatalx("can't modify exponent -- negative or zero beta value!\n");

    if ((tsum < .0001) || (tsum>10000) || (count == 0)) {
 
     vst(betas[count], betas[count], 1.0/tsum, numstates) ;
     betaexp[count] += log10(tsum) ;

    }
    
      
     cupt = tsnpm[j] ;
     if (!(cupt -> isfake) || (j == magicnum))  {
      tcalcX(cupt, indiv, X, numstates) ;
      vvt(beta_star, betas[count], X, numstates) ;
     }
     else copyarr(betas[count], beta_star, numstates) ;

  }

  loglike =  betaexp[0] + log10(vdot(ttt, beta_star, numstates)) ; 

  count = nummarkers ;
  if (scale != NULL) copyarr(betaexp, scale, count) ;

  free2D(&M, numsta) ;

  free(betaexp);
  free(beta_star);

  return loglike ;

}

double HMMalpha
 (
 SNP **tsnpm,
 int indiv,
 double theta, 
 double lambda, 
 int first, 
 int last, 
 double **alphas,
 double *ainit, 
 int numstates , double *scale)

/* returns log10 score */

/** 
 dangerous bend.  It is alphas[count][k] 
 but in HMMwindow gammas[k][count] 

 we maybe should fix this
*/


{

  int j,k,x,y,count ;

  double  *alphaexp ;
  double poisson = 0.0, temp, *alpha_star,  temp2, glen;
  double loglike ;
  double www[3], ta[3], tb[3], y1, y2, y3 ; /* debug */
  double ttt[3], yy ;
  int nummarkers ;
  int dowork = YES ;  /* no we just skip */
  int numsta = 3 ;

  double X[3];
  double *aaa, *bbb ;
  SNP *cupt ;
  Indiv *indx ;

  double **M ;
  
  int chrom, ignore ;

  count = nummarkers = last-first+1 ;

  cupt = tsnpm[first] ;
  chrom = cupt -> chrom ;
  indx = indivmarkers[indiv] ;
  ignore = indx -> ignore ;

  if (ignore == YES) dowork = NO ;

  if ((ainit != NULL) && (ainit[0] < 0.0))   dowork = NO ;

  if (dowork == NO)  {
   for (k=0; k<count; k++)  {  
    vclear(alphas[k], -1.0, numstates) ;
   }
   return 0.0 ;
  }

  sett1(ttt, theta, numstates) ;


  ZALLOC(alphaexp,count,double);
  ZALLOC(alpha_star,numsta,double);


  M = initarray_2Ddouble(numsta,numsta,0.0);

  /* ###############  Alpha Pass ################# */
  
  for (j=first; j<=last; ++j) {

    count = j-first ;
    if (j == first) {
      if (ainit != NULL) {
       copyarr(ainit, alpha_star, numstates) ;
      }
      else  {
       copyarr(ttt, alpha_star, numstates) ;
      }
      alphaexp[count]=0.0;
    }
    else {
       glen = (snpmarkers[j] -> genpos) - (snpmarkers[j-1] -> genpos);
/* this is in morgans */ 
       temp = lambda * glen ;
       if (temp<20.0) {
        poisson = exp(-temp) ;
       }
       else poisson = 0.0 ;
       if (lambda>500) poisson = 0.0 ;

    calcM(poisson, theta, M, numstates) ;

    /* create alpha_star vector by matrix multiplication */
       vzero(alpha_star, 3) ;
       for (x=0; x<numstates; x++) {
	for (y=0; y<numstates; y++) {
	  alpha_star[x] += M[x][y] * alphas[count-1][y];
/* x is new state y is old */
	}
       }

      alphaexp[count] = alphaexp[count-1];
    }
    
     cupt = tsnpm[j] ;
     if (!(cupt -> isfake) || (j == magicnum))  {
      tcalcX(cupt, indiv, X, numstates) ;
      if ((ainit != NULL) && (j==first)) vclear(X, 1.0, numstates) ;
      vvt(alphas[count], alpha_star, X, numstates) ;
     }
     else { 
      copyarr(alpha_star, alphas[count],  numstates) ;
      vclear(X, 1.0, numstates) ;
     }

    temp = asum(alphas[count], numstates) ;
    if (temp <= 0.0) 
      fatalx("can't modify exponent -- negative or zero alpha value!\n");

    if ((temp < .0001) || (temp>10000) || (j == last)) {
 
     vst(alphas[count], alphas[count], 1.0/temp, numstates) ;
     alphaexp[count] += log10(temp) ;
    }

/**
    if ((debug) && (count<=3)) { 
     printf("zwww %d %12.6f\n", count, theta) ;
     print3(alphas[count]) ;
     print3(alpha_star) ;
     print3(X) ;
    }
*/

  }

  if (scale != NULL) copyarr(alphaexp, scale, count) ;
  loglike =  alphaexp[count]  ; 

  count = nummarkers ;
  free2D(&M, numsta) ;

  free(alphaexp);
  free(alpha_star);

  return loglike ;

}



void calcM(double poisson, double theta, double **M, int numstates)
{
  double PrCC, PrAA;

  PrCC = poisson + (1-poisson)*theta;	
  PrAA = poisson + (1-poisson)*(1-theta);
  
  if (numstates == 3) {
    M[0][0] = PrAA * PrAA;
    M[1][0] = 2 * PrAA * (1-PrAA);
    M[2][0] = (1-PrAA) * (1-PrAA);
    
    M[0][1] = PrAA * (1-PrCC);
    M[1][1] = (1-PrAA) * (1-PrCC) + PrAA * PrCC;
    M[2][1] = (1-PrAA) * PrCC;
    
    M[0][2] = (1-PrCC) * (1-PrCC);
    M[1][2] = 2 * PrCC * (1-PrCC);
    M[2][2] = PrCC * PrCC;
  }
  else if (numstates == 2) {
    M[0][0] = PrAA;
    M[1][0] = (1-PrAA);
    M[0][1] = (1-PrCC);
    M[1][1] = PrCC;
  }
}
