#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <nicklib.h>
#include <getpars.h>
#include "mcmcpars.h"  
#include "mcio.h"  
#include "admutils.h" 


#define WVERSION   "1320" 
// badpairs array now bitpacked
// missing snps written to output as bad
// check made on too many pairs
// bad bug fixed
// no explicit limit on maxpairs
// bug fixed (linux mutxx)
// bug fixed (nsplit = 0 in addsc)

#define MAXSTR  512


SNP **snpmarkers ;
int numsnps ; 
int qtmode = NO ;
int verbose = NO ;
int smode = NO ; /* changes score */
int randommode = NO ;
int seed = 0 ;
// smode YES => snps with larger counts preferred */
char *trashdir = "/var/tmp" ;

char *snpname = NULL ;
char *badpairsname = NULL ;
char *oname = NULL ;
char *badsnpname = NULL ;
char *goodsnpname = NULL ;

int *badpairs ;
int lenbadpairs ;
int maxpairs = 0 ;

void readcommands(int argc, char **argv) ;
int getmarkers(char *name, char **mname, double *sc) ;

double entrop(double *a, int n)  ;
double xxlog2(double t) ;
double mutx(double *dd) ;
double mutxx(double *dd, int m , int n) ;

int getpairs(char *name, char **mname, int lmname, int *b1, int *b2) ;
double gc(int *ilist, double *score, int n)  ;

void
loadsc(SNP **snpm, int numsnps, double *score) ;

void
addsc(char *name, char **mname, int numsnps, 
 double *score, double addval, FILE *fff);
 
void putbadp(int j1, int j2, int n, int val)  ;
int  getbadp(int j1, int j2, int n)   ;         

int main(int argc, char **argv)
{
  SNP *cupt ;
  char **mname ;
  double *score ;
  double *sc ;
  int i, k, nmark, nbadp, nbad ;
  int j1, j2, nignore, n ;
  int *b1, *b2 ;
  int  *bindx, *cindx, *ilist ;
  FILE *ofile ;

  readcommands(argc, argv) ;
  if ((randommode) && (seed==0)) {
   seed = seednum() ;
   printf("seed: %d\n", seed) ;
  }

  if (randommode) SRAND(seed) ;

  ofile = stdout; 
  if (oname != NULL) openit(oname, &ofile, "w") ;

  nmark = numsnps = 
    getsnps(snpname, &snpmarkers, 0.0, 
    NULL, &nignore, 1) ;

  maxpairs = numlines(badpairsname) ;

  ZALLOC(score, numsnps, double) ;
  ZALLOC(mname, numsnps, char *) ;
  ZALLOC(b1, maxpairs, int) ;
  ZALLOC(b2, maxpairs, int) ;
  for (i=0; i< numsnps; i++) {
   mname[i] = strdup(snpmarkers[i] -> ID) ;
  }
  loadsc(snpmarkers, numsnps, score) ;
  addsc(badsnpname, mname, numsnps, score, -10000, ofile) ;
  for (i=0; i<numsnps; i++) { 
   if (score[i] < -5000) snpmarkers[i] -> ignore = YES ;
  }
  addsc(goodsnpname, mname, numsnps, score, 1000, ofile) ;
  nbadp = getpairs(badpairsname, mname, nmark, b1, b2) ;
  printf("numsnps:  %d  nbadp: %d\n", nmark, nbadp) ;
  fflush(stdout) ;

  for (i=0; i<nmark; i++)  {  
    if (verbose) 
     printf("%5d %25s %9.3f\n", i, mname[i], score[i]) ;
  }
// bindx maps variable i in bad list -> variable k
// cindx is inverse map

  n = MAX(nbadp, nmark) ;
  ZALLOC(bindx, nmark, int) ;
  ZALLOC(cindx, nmark, int) ;

  ivclear(bindx, -1, nmark) ;
  ivclear(cindx, -1, nmark) ;

  nbad = 0 ;
  for (i=0; i<nbadp; i++) {  
   k = b1[i] ;
   if (cindx[k] < 0 ) {  
    cindx[k] = nbad ;
    bindx[nbad] = k ;
    ++nbad ;
   }
   k = b2[i] ;
   if (cindx[k] < 0 ) {  
    cindx[k] = nbad ;
    bindx[nbad] = k ;
    ++nbad ;
   }
  }
  lenbadpairs = (nbad*nbad)/32 ;  
  ++lenbadpairs ;
  printf("allocating badpairs:  nbad: %d  lenbadpairs: %d\n", nbad,  lenbadpairs) ;
  fflush(stdout) ;
  ZALLOC(badpairs, lenbadpairs, int) ;
  ZALLOC(ilist, nbad, int) ;
  ZALLOC(sc, nbad, double) ;
  for (i=0; i<nbadp; i++) {  
   j1 = cindx[b1[i]] ;
   j2 = cindx[b2[i]] ;
    putbadp(j1, j2, nbad, 1) ;            
    putbadp(j2, j1, nbad, 1) ;            
  }

  for (i=0; i<nbad; i++) {  
   k = bindx[i] ;
   sc[i] = score[k] ;
  }

  printf("## nbad: %d\n", nbad) ;
  fflush(stdout) ;
  gc (ilist, sc, nbad) ;

  for (i=0; i<nbadp; i++) {  
   j1 = cindx[b1[i]] ;
   j2 = cindx[b2[i]] ;
   printf("%4d %15s %15s",i, mname[b1[i]], mname[b2[i]]) ;
   printf(" %9.3f  %9.3f", score[b1[i]], score[b2[i]]) ;
   printf("%2d %2d\n", ilist[j1], ilist[j2]) ;
  }

  for (i=0; i<nbad; i++) {  
   if (ilist[i]>0) continue ;
   k = bindx[i] ;
   snpmarkers[k] -> ignore = YES ;
  }
  for (k=0; k<numsnps; k++) {  
   cupt = snpmarkers[k] ;
   if (!(cupt -> ignore)) continue ;
   fprintf(ofile, "%s\n", cupt -> ID) ;
// printf("zzbug: %d %s\n", k, cupt -> ID) ;
  }
  if (oname != NULL) fclose(ofile) ;
  printf("## end of run\n") ;  fflush(stdout) ;
  return 0 ;
}

void readcommands(int argc, char **argv) 

{
  int i;
  char *parname = NULL ;
  phandle *ph ;
  char str[5000]  ;
  char *tempname ;
  int n ;

  while ((i = getopt (argc, argv, "p:vV")) != -1) {

    switch (i)
      {

      case 'p':
	parname = strdup(optarg) ;
	break;

      case 'v':
	printf("version: %s\n", WVERSION) ; 
	break; 

      case 'V':
	verbose = YES ;
	break; 

      case '?':
	printf ("Usage: bad params.... \n") ;
	fatalx("bad params\n") ;
      }
  }

         
   pcheck(parname,'p') ;
   printf("parameter file: %s\n", parname) ;
   ph = openpars(parname) ;
   dostrsub(ph) ;
  
   writepars(ph) ;
   
/**
## params for gpick
ADBASE:  /fg/nfiles/admixdata
DIR:  /fg/nfiles/admixdata/ms5
DIRX:  /fg/nfiles/admixdata/ms5x
SCRATCH:  DIR/outfiles 
SADIR:    ADBASE/samplefiles
snpname: SADIR/snpxvdatafile
## In main run 
## never pick badsnp
## pick good snp if possible
badsnp:   badsnp
goodsnp:  gpicked
badpairs: badpairs5
output:   glist 
*/
   getstring(ph, "snpname:", &snpname) ;
   getstring(ph, "badsnp:", &badsnpname) ;
   getstring(ph, "goodsnp:", &goodsnpname) ;
   getstring(ph, "badpairs:", &badpairsname) ;
   getstring(ph, "output:", &oname) ;
   getint(ph, "smode:", &smode) ;
   getint(ph, "seed:", &seed) ;
   getint(ph, "randommode:", &randommode) ;
   closepars(ph) ;

}
void
addsc(char *name, char **mname, int numsnps, 
 double *score, double addval, FILE *ofile)
{

  FILE *fff ;
  char line[MAXSTR] ;
  char *spt[MAXFF] ;
  char *sx ;
  int nsplit, i, j, num=0 ;
  int k1 ;

  if (name == NULL) return;
  if (strcmp(name, "NULL") == 0) return;
  num = 0; 
  if (mname == NULL) fatalx("no badpairsname\n") ; ;
  openit(name, &fff, "r") ;
  while (fgets(line, MAXSTR, fff) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   if (nsplit<1) { 
     continue ;
   }
   sx = spt[0] ;
   if (sx[0] == '#') {
    freeup(spt, nsplit) ;
    continue ;
   }
   k1 = indxindex(mname, numsnps, spt[0]) ;
   if (k1<0) { 
     printf("missing SNP %s %f\n", line, addval)  ;
     fprintf(ofile, "%s\n", spt[0]) ;
     continue ;
   }
   score[k1] += addval ;
  }
  fclose (fff) ;
}

int getpairs(char *name, char **mname, int lmname, int *b1, int *b2) 
{
  FILE *fff ;
  char line[MAXSTR] ;
  char *spt[MAXFF] ;
  char *sx ;
  int nsplit, i, j, num=0 ;
  int k1, k2 ;

  num = 0; 
  if (mname == NULL) fatalx("no badpairsname\n") ; ;
  openit(name, &fff, "r") ;
  while (fgets(line, MAXSTR, fff) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   if (nsplit<2) { 
     printf("bad line %s\n", line)  ;
     continue ;
   }
   sx = spt[0] ;
   if (sx[0] == '#') {
    freeup(spt, nsplit) ;
    continue ;
   }
   k1 = indxindex(mname, lmname, spt[0]) ;
   k2 = indxindex(mname, lmname, spt[1]) ;
   if ((k1<0) || (k2<0)) { 
     printf("missing SNP %s\n", line)  ;
     abort() ;
   }
   if (num >= maxpairs) { 
    fatalx("too many pairs\n") ;
   }
   b1[num] = k1 ;
   b2[num] = k2 ;
   ++num ;
   freeup(spt, nsplit) ;
  }
  fclose(fff) ;
  return num ;
}
void loadsc(SNP **snpm, int numsnps, double *sc) 
{

  SNP *cupt ;
  double dd[4] ;
  double theta = 0.2 ;
  double tt[4], pa, pe, qa, qe, pt, qt  ;
  double xa, xe, xx ;
  int num ;

  pt = theta ; qt = 1.0 - pt ;
  for (num= 0; num < numsnps; ++num) {

   cupt = snpm[num] ;

   tt[0] = cupt -> popA_nn[0] ;
   tt[1] = cupt -> popA_nn[1] ;
   tt[2] = cupt -> popB_nn[0] ;
   tt[3] = cupt -> popB_nn[1] ;

   xa = asum(tt,2) ;
   xe = asum(tt+2,2) ;
   pa = tt[0]/xa ;
   pe = tt[2]/xe ;
   qa = 1.0 - pa ;
   qe = 1.0 - pe ;

   dd[0] = qt*pa ;
   dd[1] = qt*qa ;
   dd[2] = pt*pe ;
   dd[3] = pt*qe ;

   if (randommode) {  
    sc[num] = 100.0*DRAND()  ;
   }
   else {
    sc[num] = mutx(dd) ;
   }
   if (smode) sc[num] *= MIN(xe, xa) ;
   
  }

}

double mutxx(double *dd, int m, int n) 
{
   double *aa, *bb ; 
   double y ;

   ZALLOC(aa, m, double) ;
   ZALLOC(bb, n, double) ;

   rrsum(dd, bb, m, n) ;
   ccsum(dd, aa, m, n) ;
/**
   if ( ((m*n)<=4)  && verbose) {  
    printf("m: %d n: %d\n",m,n) ;
    printmat(dd,m,n) ;
    printf("\n") ;
    printmat(aa,1,m) ;
    printmat(bb,1,n) ;
   }
*/
   y = entrop(aa,m) + entrop(bb,n) - entrop(dd,m*n) ;

  free(aa) ;  free(bb) ;
  return y ;

}
double mutx(double *dd) 
/** mutual inf from 2 x 2 */
{
    double y ;
    y = mutxx(dd, 2,2 ) ;
    return y  ;
}
double gc(int *ilist, double *score, int n) 
/** 
 a greedy algorithm 
 outputting markers in turn 
*/
{
   double *sc ; 
   int *inda ; 
   int *badx ;

   int i, k, j ;
   double y ;

   ZALLOC(sc, n, double) ;
   ZALLOC(inda, n, int) ;
   ZALLOC(badx, n, int) ;

   vst(sc, score, -1.0, n) ;
   sortit(sc, inda, n) ;

   y = 0.0 ;
   ivzero(ilist, n) ; 
   for (i=0; i<n; i++) {  
    k = inda[i] ;
    if (badx[k] >0) continue ;
     y  += score[k] ;
     for (j=0; j<n; j++) {  
      badx[j] += getbadp(k,j,n) ;
     }
     ilist[k] = 1; 
// ilist[k] are the good guys we want.  Others are bad
   }

   free(sc) ;
   free(inda) ;
   free(badx) ;

   return y ;

}
void
putbadp(int j1, int j2, int n, int val)             
{

   int a, b, x, msk ;
   x = j1*n+j2 ;
   a = x/32 ;
   b = x % 32 ;

   msk = val <<  b ;  

   badpairs[a] |= msk ;

}
int
getbadp(int j1, int j2, int n)             
{

   int a, b, x, msk, val, t ;
   x = j1*n+j2 ;
   a = x/32 ;
   b = x % 32 ;

   t = badpairs[a] >> b ;  
   val = t & 1 ;

   return val ;

}
