#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

#include <nicklib.h>
#include <getpars.h>

#include "admutils.h"
#include "mcio.h"  
#include "logsubs.h" 
#include "regsubs.h" 
#include "emsubs.h" 
#include "mcmcpars.h"  

#define WVERSION   "1000" 

#define MAXSTR  512
#define MAXHAPSIZE  4
#define HAPADMIN 7  
/* number of nonhap fileds in output array */

char *trashdir = "/var/tmp" ;
int *fchrom, *lchrom ;
int noxdata = YES ;

int verbose = NO ;
int qtmode = NO ;

Indiv **indivmarkers;
Indiv **parindivmarkers ;
SNP **snpmarkers ;
int numsnps, numindivs, parnumindivs ; 

int numgvals = 4 ;

char  *genotypename = NULL ;
char  *snpname = NULL ;
char  *snpoutfilename = NULL ;
char  *hapfilename = NULL ;
char  *indivname = NULL ;
char *outputname = NULL ;
char *taglistname = NULL ;
char *tagfilehead = NULL ;
char  *pargenotypename = NULL ;
char  *parindivname = NULL ;

int lotag = 401 ;
int hitag = 413 ;
int lopos, hipos ; 
int xchrom = -1 ;
double maxgdis = .002 ; // haplotype width  in Morgans
double hapthresh[MAXHAPSIZE] ;
int *lx, *rx ;  //  left, right allowable window
int xxreg[MAXHAPSIZE] ;
int xxlhap = -1 ;
int maxlhap = 99 ;

double ***gammavecs ;  // (tag, indivnum, 3)

int numtags ;

void readcommands(int argc, char **argv) ;
void setfc(SNP **snpm, int numsnps)  ;
void gettags(SNP **snpm, int numsnps, char *taglistname)  ;
int  settags(SNP *cupt) ;
void loadp(SNP *parcupt, int *na, int *nb)  ;
void loadpx(int pat, int *patarr, int *gtypes, int n,  int *na, int *nb) ;
int load1(double **gams, int *gobs, int *iscase, SNP *cupt)  ;
int load1x(int pat, int *patarr, int *gtypes, int *idnum, int n,
  double **gams, int *gobs, int *iscase)  ;
void splitc2(int *nn, int g) ;
void iswap(int *a, int *b) ;
void xcalc(int x, double pa, double pb, double *X) ;
double logrx(double **gams, int *gobs, int *iscase, int neq) ;
double logscor(double **gams, int *gobs, int *iscase, 
 int neq, double pa, double pe) ;
void printscoretable(FILE *ofile, int *scoretable, int nrow, int ncol, int row)   ;
int grabgeasy(int *genos, int *gtypes, int *idnum, int *mreg, int n)  ;
void phaseit(int *haplos, int *genos, int ndip, int len)  ;
void emdoit(int *odata, int *udata, int nt, int len, int *base1, int *base2)  ;
int codeit2x(int *ttt, int len);
void uncodeit2x(int *ttt, int len, int kk);
void printsnpsx(char *snpoutfilename, SNP **snpm, int num, int printfake, int printvalids) ;
void gethaps(char ***haplist, int *haplen, int maxn, int maxl, char *haplname)  ;
double logsceasy(int *gobs, int *iscase, int neq)  ;
void loadcc(double *ch2, int *gobs, int *iscase, int neq)  ;



int main(int argc, char **argv)
{
  int i, j, k, a, b, n, npow, ihap, lhap, itable ;
  char str[MAXSTR] ;
  int nignore ;
  Indiv *indx ;
  SNP *cupt ;
  double *gpt ;
  int tag, tagnum ; 
  double yscore, y, yl1, yl2, y1, y2 ;
  int fc, lc ;
  int ncase, neur, neq    ;
  int na[2], nb[2] ;
  int *gobs, kmax, kmin ;    
  int *iscase ;
  int *ww ;
  double **gams, pa, pe, ggr[3], gran[3], ans[3] ;

  int totnumindivs, maxhapsize, hapsize ;  
  int *genos, *haplos, *scoretable, nscoretable, maxscoretable ;
  int *gtypes, *hpat, *hcount, *idnum ;
   
  int scrowsize = MAXHAPSIZE + HAPADMIN ;
  int *sc ;
  FILE *ofile, *hfile ;
  int mreg[MAXHAPSIZE] ;
  int xreg[MAXHAPSIZE] ;
  int ndip, nhap, nx ;
  int x, kk, xleft, xright ;
  char *** haplist  ;
  int *haplen ; 
  int nlist, len, kpt ;
  double ch2[4], ychi, cx[4] ;

  hapthresh[0] =  5.0 ;
  hapthresh[1] = 10.0 ;
  hapthresh[2] = 15.0 ;
  hapthresh[3] = 20.0 ;

  readcommands(argc, argv) ;

  if (outputname != NULL) 
   openit(outputname, &ofile, "w") ;
  else ofile = stdout ;
  if (hapfilename  == NULL) 
   fatalx("hapfilename not set\n") ;


  numsnps = 
    getsnps(snpname, &snpmarkers, 0.0, NULL, &nignore, 1) ;

  setfc(snpmarkers, numsnps) ;
  fflush(stdout) ;

   numindivs = getindivs(indivname, &indivmarkers) ;
   ncase = setstatus(indivmarkers, numindivs, "Case") ;

   getgenos(genotypename, snpmarkers, indivmarkers, 
    numsnps, numindivs, nignore) ;

  rmsnps(snpmarkers, numsnps, nignore) ;  //  rid ignorable snps

  gethpos(&fc, &lc, snpmarkers, numsnps, xchrom, lopos, hipos)   ;
  neur = parnumindivs = 0 ;

  
  printf("numindivs: %d  parnumindivs: %d\n", 
   numindivs, parnumindivs) ;
  printf("ncase: %d  PopB: %d  PopA: %d\n", 
    ncase, neur, parnumindivs-neur) ;
 

  totnumindivs = numindivs + parnumindivs ;
  maxhapsize = MAXHAPSIZE ;

  ZALLOC(genos, totnumindivs*maxhapsize, int) ;
  ZALLOC(gtypes, totnumindivs, int) ;
  ZALLOC(idnum, totnumindivs, int) ;
  ZALLOC(haplos, 2*totnumindivs*maxhapsize, int) ;
  ZALLOC(hpat, 2*totnumindivs, int) ;
  n = (int) pow(2.0, (double) maxhapsize) ;
  ZALLOC(hcount, n, int) ;
  maxscoretable = 50000 ;  
  ZALLOC(scoretable, maxscoretable*scrowsize, int) ;
  ivclear(scoretable, -1, maxscoretable*scrowsize) ;
  nscoretable = 0 ;


  ZALLOC(gobs, numindivs, int) ;
  ZALLOC(iscase, numindivs, int) ;

  tagnum = -1 ;

  gran[0] = gran[2] = 1.0 ;
  gran[1] = 2.0 ; bal1(gran,3) ;
  lhap = 1 ;  //  haplotype length 

  ZALLOC(lx, lc+1, int) ;
  ZALLOC(rx, lc+1, int) ;
  ZALLOC(ww, lc+1, int) ;

// set rx, lx
  ivclear(lx, 999999, lc+1) ;
  ivclear(rx, -1, lc+1) ;
  ivclear(ww, -1, lc+1) ;

  for (i=fc; i<=lc; i++) { 
   y1 = snpmarkers[i] -> genpos ;

   for (k=i ; k <=lc ; ++k)  {  
    y2 = snpmarkers[k] -> genpos ;
    if (y2 > (y1+maxgdis)) break ;
    rx[i] = k ;
   }
   for (k=i ; k >=fc ; --k)  {  
    y2 = snpmarkers[k] -> genpos ;
    if (y2 < (y1-maxgdis)) break ;
    lx[i] = k ;
   }
  }

  printsnpsx(snpoutfilename, snpmarkers, numsnps, NO, YES) ;
  if (maxlhap==0) {  
   printf("end of run\n") ;
   return 0 ;
  }

  nlist = numlines(hapfilename) ;
  ZALLOC(haplist, nlist, char **) ;
  ZALLOC(haplen, nlist, int) ;

  for (i=0; i<nlist; i++)  {  
   ZALLOC(haplist[i], maxlhap, char *) ; 
  }
  gethaps(haplist, haplen, nlist, maxlhap, hapfilename) ;

  for (kpt=0; kpt<nlist; kpt++)  {  
   len = haplen[kpt] ;
   if (len<1) continue ;
   xxlhap = len ;
   for (j=0; j<len; j++) {  
    xxreg[j] = snpindex(snpmarkers, numsnps, haplist[kpt][j]) ; 
   }

   lhap = xxlhap ;
   copyiarr(xxreg, mreg, lhap) ; 
   isortit(mreg, NULL, lhap) ;
    printf("xxmreg ") ;  printimat(mreg, 1, lhap) ;
    a = mreg[0] ;
    if (a<0) continue ;
    cupt = snpmarkers[a] ;
    npow = (int) pow(2.0, (double) lhap) ;
    ndip = grabgeasy(genos, gtypes, idnum, mreg, lhap) ;
    if (ndip<=0) return 0 ;
    phaseit(haplos, genos, ndip, lhap) ;
    nhap = ndip*2 ;
    for (i=0; i<nhap; i++) {
     hpat[i] = codeit2x(haplos+i*lhap, lhap) ;
    }
    countcat(hpat, nhap, hcount, npow) ;
    loadpx(0, hpat, gtypes, -1, na, nb) ;  // count pattern in moderns

    for (ihap=0; ihap<npow; ihap++)  {  
     if ((lhap==1) && (ihap>0)) continue ;
     if (lhap>1) {
      if (hcount[ihap]<10) continue ; // 10 should be parameter
      if (hcount[ihap]>(nhap-10)) continue ; // 10 should be parameter
     }

     loadpx(ihap, hpat, gtypes, -1, na, nb) ;  // count pattern in moderns
     neq = load1x(ihap, hpat, gtypes, idnum, ndip, NULL, gobs, iscase) ;
     loadcc(ch2, gobs, iscase, neq)  ;
     if (neq == 0) continue ;
     ivmaxmin(gobs, neq, &kmax, &kmin) ;
     if (kmax>2) fatalx("bad genotype\n") ;
     vsp(cx, ch2, 0.5, 4) ;
     yscore = conchi(cx, 2, 2) ;          
     printf("scorehap:") ;
     for (k=0; k<lhap; k++) { 
      printf(" %4d",mreg[k]) ;
     }
     printf("  %4d", ihap) ;
     printf(" %9.3f", yscore) ;
     for (k=0; k<lhap; k++) { 
      b = mreg[k] ;
      cupt = snpmarkers[b] ;
      printf(" %12s",cupt -> ID) ;
     }
     printf("\n") ;
     printmat(ch2, 2, 2) ;

     sc = scoretable + nscoretable*scrowsize ;
     sc[0] = nnint(1000*yscore) ;
     sc[1] = lhap ;  // # markers ;
     sc[2] = ihap ;  // haplotype pattern
     copyiarr(na, sc+3, 2) ;
     copyiarr(nb, sc+5, 2) ;
     for (k=0; k<lhap; k++) {  
      sc[7+k] = mreg[k] ;
     }
     ++nscoretable ;
    }
  }
    printscoretable(ofile, scoretable, nscoretable, scrowsize, -1) ;
    printf("end of run\n") ;
    return 0 ;


  for (i=fc; i<=lc; i++) { 
   cupt = snpmarkers[i] ;
   
   settags(cupt) ;
   neq = load1(gams, gobs, iscase, cupt) ;
   if (neq == 0) continue ;
   ivmaxmin(gobs, neq, &kmax, &kmin) ;
   if (kmax>2) fatalx("bad genotype\n") ;

   yscore = logsceasy(gobs, iscase, neq) ; 
   printf("score: %6d %20s %9.3f  %d\n", i, cupt -> ID, yscore, intsum(iscase, neq)) ;
   if (yscore<hapthresh[lhap-1]) continue ; 
   if (nscoretable>=maxscoretable) fatalx("overflow in scoretable\n") ;
   sc = scoretable + nscoretable*scrowsize ;
   sc[0] = nnint(1000*yscore) ;
   sc[1] = 1 ;  // # markers ;
   sc[2] = 0 ;  // haplotype pattern
   copyiarr(na, sc+3, 2) ;
   copyiarr(nb, sc+5, 2) ;
   sc[7] = i ;
   ++nscoretable ;
  }
  printscoretable(ofile, scoretable, nscoretable, scrowsize, -1) ;
  nscoretable = idedupe(scoretable, nscoretable, scrowsize) ;
  nx = 0 ;
  for (lhap =2 ; lhap <= maxhapsize; ++lhap)  {
   if (lhap>maxlhap) break ;
   for (itable=0; itable<nscoretable; ++itable)  {  
    sc = scoretable + itable*scrowsize ;
    if (sc[1] != (lhap-1)) continue ;
    copyiarr(sc+7, xreg, lhap-1) ;

    xright = rx[xreg[0]] ;
    xleft  = lx[xreg[lhap-2]] ;  // window sizes

    ++nx ;
    for (j=0; j<lhap-1; j++) {  
     k = xreg[j] ; 
     ww[k] = nx ;
    }

   for (kk=xleft; kk<=xright; ++kk) {  
    if (ww[kk] == nx) continue ;
    copyiarr(xreg, mreg, lhap-1) ;
    mreg[lhap-1] = kk ;  
    isortit(mreg, NULL, lhap) ;
    a = mreg[0] ;
    cupt = snpmarkers[a] ;
    npow = (int) pow(2.0, (double) lhap) ;
    ndip = grabgeasy(genos, gtypes, idnum, mreg, lhap) ;
    if (ndip<=0) continue ;
    phaseit(haplos, genos, ndip, lhap) ;
    nhap = ndip*2 ;
    for (i=0; i<nhap; i++) {
     hpat[i] = codeit2x(haplos+i*lhap, lhap) ;
    }
    countcat(hpat, nhap, hcount, npow) ;

    for (ihap=0; ihap<npow; ihap++)  {  
     if ((lhap==1) && (ihap>0)) continue ;

     a = mreg[0] ;

     if (hcount[ihap]<10) continue ; // 10 should be parameter
     if (hcount[ihap]>(nhap-10)) continue ; // 10 should be parameter

     neq = load1x(ihap, hpat, gtypes, idnum, ndip, NULL, gobs, iscase) ;
     loadcc(ch2, gobs, iscase, neq) ;
     if (neq == 0) continue ;
     ivmaxmin(gobs, neq, &kmax, &kmin) ;
     if (kmax>2) fatalx("bad genotype\n") ;
     vsp(cx, ch2, 0.5, 4) ;
     yscore = conchi(cx, 2, 2) ;          
     if (yscore<hapthresh[lhap-1]) continue ;
     printf("scorehap:") ;
     for (k=0; k<lhap; k++) { 
      printf(" %4d",mreg[k]) ;
     }
     printf("  %4d", ihap) ;
     printf(" %9.3f", yscore) ;
     for (k=0; k<lhap; k++) { 
      b = mreg[k] ;
      cupt = snpmarkers[b] ;
      printf(" %12s",cupt -> ID) ;
     }
     printf("\n") ;
     printmat(ch2, 2, 2) ;

     sc = scoretable + nscoretable*scrowsize ;
     sc[0] = nnint(1000*yscore) ;
     sc[1] = lhap ;  // # markers ;
     sc[2] = ihap ;  // haplotype pattern
     copyiarr(na, sc+3, 2) ;
     copyiarr(nb, sc+5, 2) ;
     for (k=0; k<lhap; k++) {  
      sc[7+k] = mreg[k] ;
     }
     ++nscoretable ;
     printscoretable(ofile, scoretable, nscoretable, scrowsize, nscoretable-1) ;
    }
   }
  }
  nscoretable = idedupe(scoretable, nscoretable, scrowsize) ;
 }

   fclose(ofile) ;
   printf("## end of run\n") ;

   return 0 ;

}
void
loadp(SNP *parcupt, int *na, int *nb) 
{
   Indiv *indx ;
   SNP *cupt ;
   int i, num, g ;
   int cc[2], *nn ;

   num = parcupt -> markernum ;
   cupt = snpmarkers[num] ;

   ivclear(na, 1, 2) ;
   ivclear(nb, 1, 2) ;

   if (parcupt -> ngtypes == 0) return ;
   
  for (i=0; i<parnumindivs; i++) {  
   g = parcupt -> gtypes[i] ;
   if (g<0) continue ;
   splitc2(cc, g) ;
   indx = parindivmarkers[i] ;
   if (indx -> affstatus) nn = nb ;
   else nn = na ;
   ivvp(nn,nn, cc, 2) ;
  }
}
void
loadpx(int pat, int *patarr, int *gtypes, int n,  int *na, int *nb) 
{
   int i, num, g, typ ;
   int cc[2], *nn ;

   if (n <0) { 
    ivzero(na,2) ;
    ivzero(nb,2) ;
    return ;
   }

   ivclear(na, 1, 2) ;
   ivclear(nb, 1, 2) ;

   if (n == 0) return ;
   
  for (i=0; i<n; i+=2) {  

   g = 0 ;
   if (patarr[i] == pat) ++g ;
   if (patarr[i+1] == pat) ++g ;
   splitc2(cc, g) ;
   typ = gtypes[i/2] ;
   if (typ <= 1) continue ;
   switch (typ) { 
    case 10:  
     nn = na ; break ;
    case 11:  
     nn = nb ; break ;
    default: 
     fatalx("bad gtype: %d %d\n",i, typ) ;
   }
   ivvp(nn, nn, cc, 2) ;
  }
}


void splitc2(int *nn, int g) 
{
  ivzero(nn,2) ;
  if (g<0) return ;
  nn[1] = g ; nn[0] = 2-g ;
  return ;
}

void readcommands(int argc, char **argv) 

{
  int i ;
  char *parname = NULL ;
  phandle *ph ;
  char str[5000]  ;
  char *tempname ;
  int n ;

  while ((i = getopt (argc, argv, "p:vV")) != -1) {

    switch (i)
      {

      case 'p':
	parname = strdup(optarg) ;
	break;

      case 'v':
	printf("version: %s\n", WVERSION) ; 
	break; 

      case 'V':
	verbose = YES ;
	break; 

      case '?':
	printf ("Usage: bad params.... \n") ;
	fatalx("bad params\n") ;
      }
  }

         
   pcheck(parname,'p') ;
   printf("parameter file: %s\n", parname) ;
   ph = openpars(parname) ;
   dostrsub(ph) ;
   writepars(ph) ;

   getstring(ph, "output:", &outputname) ;
   getstring(ph, "genotypename:", &genotypename) ;
   getstring(ph, "snpname:", &snpname) ;
   getstring(ph, "snpoutfilename:", &snpoutfilename) ;
   getstring(ph, "hapfilename:", &hapfilename) ;
   getstring(ph, "indivname:", &indivname) ;
   getstring(ph, "tagfilehead:", &tagfilehead) ;
   getstring(ph, "taglistname:", &taglistname) ;
   getstring(ph, "parent_genotypename:", &pargenotypename) ;
   getstring(ph, "parent_indivname:", &parindivname) ;
   getint(ph, "lotag:", &lotag) ;
   getint(ph, "hitag:", &hitag) ;
   getint(ph, "lopos:", &lopos) ;
   getint(ph, "hipos:", &hipos) ;
   getint(ph, "chrom:", &xchrom) ;
   getint(ph,"maxlhap:", &maxlhap) ;


   closepars(ph) ;

}


void setfc(SNP **snpm, int numsnps) 
// also allocates fchrom , lchrom
{
  int i, j, chrom ;
  SNP *cupt ;
  double dis, totdis, pos1, pos2 ;

  ZALLOC(fchrom, 25, int) ;
  ZALLOC(lchrom, 25, int) ;

  ivclear(fchrom, 999999, 25) ;
  ivclear(lchrom, -999999, 25) ;
  
/* initialize real marker array */
  for (i=0; i<numsnps; i++) {
    cupt = snpm[i] ;
    if (cupt -> isfake) continue ;
    if (cupt -> ignore) continue ;
    chrom = cupt -> chrom ;
    if ((noxdata) && (chrom==23)) continue  ;
    fchrom[chrom] = MIN(fchrom[chrom],i) ;
    lchrom[chrom] = MAX(lchrom[chrom],i) ;
  }

  totdis = 0.0 ;
  for (j=1; j<=23; j++) {  
   if (lchrom[j]<0) continue ;
   cupt = snpm[fchrom[j]]; 
   pos1 = cupt -> genpos ;
   cupt = snpm[lchrom[j]]; 
   pos2 = cupt -> genpos ;
   dis = pos2 - pos1 ;
   totdis += dis ;
   printf("chrom: %5d  first: %5d  last:  %3d", 
     j, fchrom[j], lchrom[j]) ;
   printf(" distance: %9.3f\n", dis) ;
  }
  printf("total distance: %9.3f\n", totdis) ;
}

void
gethaps(char ***haplist, int *haplen, int maxn, int maxl, char *haplname) 
{
 char line[MAXSTR] ;
 char *spt[MAXFF], *sx ;
 int nsplit, num=0, k, ind, i, t ;
 SNP *cupt ;
 int tagmax ;

 FILE *fff ;

 ivzero(haplen, maxn) ;

 openit(haplname, &fff, "r") ;
 while (fgets(line, MAXSTR, fff) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   if (nsplit < 1) continue ;
   sx = spt[0] ;
   if (sx[0] == '#') {  
    freeup(spt, nsplit) ;
    continue ;
   }
   haplen[num] = nsplit ;
   for (k=0; k<nsplit; k++)  {  
    haplist[num][k] = strdup(spt[k]) ;
   }
   ++num ;
 }
 fclose(fff) ;
}



void
gettags(SNP **snpm, int numsnps, char *tlistname) 
{
 char line[MAXSTR] ;
 char *spt[MAXFF], *sx ;
 int nsplit, num=0, k, ind, i, t ;
 SNP *cupt ;
 int tagmax ;

 FILE *fff ;

 for (i=0; i<numsnps; i++) {  
  snpm[i] -> tagnumber = -1 ;
 } 
 openit(tlistname, &fff, "r") ;
 while (fgets(line, MAXSTR, fff) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   if (nsplit < 2) continue ;
   sx = spt[0] ;
   if (sx[0] == '#') {  
    freeup(spt, nsplit) ;
    continue ;
   }
    ind = snpindex(snpm, numsnps, sx)  ;
    if (ind<0) fatalx("(gettags) snp: %s not found \n", sx) ;
    cupt = snpm[ind] ;
    t = cupt -> tagnumber = atoi(spt[1]) - lotag ;
    if ((t<0) || (t>= numtags)) fatalx("bad tag: %s\n", line) ;
    freeup(spt, nsplit) ;
    continue ;
 }
 fclose(fff) ;
// now fill in missing tags 
 tagmax = 0 ;
 for (i=0; i<numsnps; i++) {  
  cupt = snpm[i] ;
  if (cupt -> ignore) continue ;
  tagmax = MAX(tagmax, cupt -> tagnumber) ;
  cupt -> tagnumber = tagmax ;
  if (verbose)
   printf("tagnumber: %20s %3d %3d\n", cupt -> ID, tagmax, tagmax+lotag) ;
 } 
}

int
load1x(int pat, int *patarr, int *gtypes, int *idnum, int n,
  double **gams, int *gobs, int *iscase) 
{
  Indiv *indx ;
  int i, neq = 0, g, x ;
  double *gg, anc ;

  for (i=0; i< n; i++) {  
   if (gtypes[i] >= 10) continue ;
   x = idnum[i] - parnumindivs ;
   if (x<0) fatalx("logic bug\n") ;
   indx = indivmarkers[x] ;
   if (indx -> ignore) continue ;
   g = 0 ;
   if (patarr[2*i] == pat) ++g ;
   if (patarr[2*i+1] == pat) ++g ;
   if (gams != NULL) {
    gg = indx -> totgamms ;
    anc = 0.5*(gg[1]+2.0*gg[2]) ;
    if (anc < 0.0) continue ;
    copyarr(gg, gams[neq], 3) ;
    bal1(gams[neq], 3) ;
   }
   gobs[neq] = g ;
   iscase[neq] = indx -> affstatus ;
   ++neq ;
  }
  printf("yy %d %d\n", neq, intsum(iscase, neq) ) ;
  return neq ;
}

int
load1(double **gams, int *gobs, int *iscase, SNP *cupt) 
{
  Indiv *indx ;
  int i, neq = 0, g ;
  double *gg, anc ;

  if (cupt -> ignore) return 0 ;
  if (cupt -> ngtypes == 0) return 0 ;

  for (i=0; i< numindivs; i++) {  
   indx = indivmarkers[i] ;
   if (indx -> ignore) continue ;
   g = cupt -> gtypes[i] ; 
   if (g<0) continue ;
   if (g>2) { 
    printf("bad... %s %s\n", cupt -> ID, indx -> ID) ;
    printf("%d %d %d\n", 
      cupt -> ngtypes, cupt -> ignore, indx -> ignore) ;
    printf("%d %d\n", 
      cupt -> markernum, indx -> idnum) ;
    fatalx("bad genotype %d\n", g) ;
   }
   gg = indx -> totgamms ;
   anc = 0.5*(gg[1]+2.0*gg[2]) ;
   if (anc < 0.0) continue ;
   copyarr(gg, gams[neq], 3) ;
   bal1(gams[neq], 3) ;
   gobs[neq] = g ;
   iscase[neq] = indx -> affstatus ;
   ++neq ;
  }
  return neq ;
}
void iswap(int *a, int *b)
{
   int t ;
   t = *a ;
   *a = *b ;
   *b = t ;
}

void xcalc(int x, double pa, double pb, double *X)
{

  double PopA_freq, PopB_freq;
  static int ncall=0 ;

  ++ncall ;
  vclear(X, 1.0, 3) ;
  

  if (x<0) {  
   return ; 
  }

  PopA_freq   =  pa ;
  PopB_freq =  pb ;

  if ((PopA_freq<0) || (PopB_freq<0)) 
    fatalx("marker freqs not set\n") ;

    switch(x) {
      
    case 2:
      X[0] = PopA_freq * PopA_freq;
      X[1] = PopA_freq * PopB_freq;
      X[2] = PopB_freq * PopB_freq;
      break;
      
    case 1:
      X[0] = 2 * PopA_freq * (1 - PopA_freq) ;
      X[1] = PopB_freq * (1 - PopA_freq) + 
	(1 - PopB_freq) * PopA_freq ;
      X[2] =  2 * PopB_freq * (1 - PopB_freq) ;
      break;
      
    case 0:
      X[0] = (1 - PopA_freq) * (1 - PopA_freq);
      X[1] = (1 - PopA_freq) * (1 - PopB_freq);
      X[2] = (1 - PopB_freq) * (1 - PopB_freq);
      break;
  
    default:
      fatalx("(xcalc) bad genotype %d\n", x) ;
      break;
    }
}


double
 logrx(double **gams, int *gobs, int *iscase, int neq) 
// setup for logistic regression 
// no big data structures  needed. 
{

   double *eq, *xeq, **rhs ;
   double *gpt, anc, *rr, *ee ;
   double ysc[2], ychi, ans[3] ;  
   int nv = 3, g, i ;
   static int ncall=0 ;

   ++ncall ;

   if (gams == NULL) nv = 2 ;
   rhs = initarray_2Ddouble(neq, 2, 0.0);

   ZALLOC(eq, neq*nv, double) ;
   ZALLOC(xeq, neq*nv, double) ;

   for (i=0; i<neq; i++) {
    g = gobs[i] ;
    rr = rhs[i] ;
    ee = eq+i*nv ;
    rr[1] = (double) g ;
    rr[0] = (double) (2-g) ;
    if (gams != NULL) {
     gpt = gams[i] ;
     anc = 0.5*(gpt[1]+2.0*gpt[2]) ;

     ee[0] = 1.0 ;
     ee[1] = anc ;
     ee[2] = (double) iscase[i] ;
    }
    else {  
     ee[0] = 1.0 ;
     ee[1] = (double) iscase[i] ;
    }
   }

    squish(xeq, eq, neq, nv, nv-1) ;
    ysc[0] = logregressit(ans, xeq, rhs, neq, nv-1) ;
    ysc[1] = logregressit(ans, eq, rhs, neq, nv) ;

    ychi = 2.0*(ysc[1] -ysc[0]) ;

   free2D(&rhs, neq) ;
   free(eq) ;
   free(xeq) ;

   return ychi ;
}
void printscoretable(FILE *ofile, int *scoretable, int nrow, int ncol, int row)  
{ 
  int i, j, k, n, *sc, lo, hi ;
  double y ;
  SNP *cupt ;

  lo = 0; hi = nrow-1 ;
  if (row>=0) { 
   lo = hi = row ;
  }
  for (i=lo ; i <= hi; i++) {  
   sc = scoretable + i*ncol ;
   y = (double) sc[0] ;
   n = sc[1] ;
   if (n<=0) continue ;
   if ((n+2) >= ncol) fatalx("bad row %d\n", n) ;
   fprintf(ofile,"%9.3f", y/1000.0) ;
/**
   fprintf(ofile,"   %3d %3d ", sc[3], sc[4]) ;
   fprintf(ofile,"   %3d %3d ", sc[5], sc[6]) ;
   for (j=0; j<n; j++) {
    fprintf(ofile," %4d", sc[j+7]) ;
   }
*/
   for (j=0; j<n; j++) {
    k = sc[j+7] ;
    cupt = snpmarkers[k] ;
    fprintf(ofile," %12s", cupt -> ID) ;
   }
   fprintf(ofile," %3d %3d ", sc[1], sc[2]) ;
   fprintf(ofile,"\n") ;
  }
  fflush(ofile) ;
}

void
phaseit(int *haplos, int *genos, int ndip, int len) 
{
  int *base1, *base2 ;
  int i, j, g, xx[2] ;

  ZALLOC(base1, len, int) ;
  ZALLOC(base2, len, int) ;

  ivclear(base1, 1, len) ;
  ivclear(base2, 2, len) ;

  for (i=0; i<ndip; i++)  {
   for (j=0; j<len; j++)   {
    g= genos[i*len+j] ;
    switch (g)  { 
     case 0: 
      xx[0] = xx[1] = 1 ; break ;
     case 1: 
      xx[0] = xx[1] = 9 ; break ;
     case 2: 
      xx[0] = xx[1] = 2 ; break ;
     default: 
      xx[0] = xx[1] = 0 ; break ;
    }
    haplos[2*i*len+j] = xx[0] ;
    haplos[2*i*len+len+j] = xx[1] ;
   }
  }
  emdoit(haplos, haplos, ndip*2, len, base1, base2)  ;

  free(base1) ;
  free(base2) ;


}

int codeit2x(int *ttt, int len)
{
  int i, k = 0, t ;

  for (i=0; i<len; i++)  {
   t = ttt[i]-1 ;
   if (t<0) fatalx("badcodeit2x\n") ;
   if (t>1) fatalx("badcodeit2x\n") ;
   k = k*2+t ;
  }
  return k ;
}

void uncodeit2x(int *ttt, int len, int kk)
{
  int i, k ;
  k = kk ;
  for (i=0; i<len; i++)  {
   ttt[len-i-1] = k%2 ;
   k /= 2 ;
  }
}

double logscor(double **gams, int *gobs, int *iscase, int neq, double pa, double pe) 
{
   int  j ; 
   double yscore, ggr[3] ;
   
   for (j=0; j<neq; j++) {  
     xcalc(gobs[j], pa, pe, ggr) ;
     vvt(gams[j], gams[j], ggr, 3) ;
     bal1(gams[j],3) ;
   }
   yscore = logrx(gams, gobs, iscase, neq) ;
   return yscore ;
}
double logsceasy(int *gobs, int *iscase, int neq) 
{
   int  j ; 
   double yscore, ggr[3] ;
   
   yscore = logrx(NULL, gobs, iscase, neq) ;
   return yscore ;
}


int settags(SNP *cupt) 
{
   static int tagnum = -1 ;
   int tag, j ;
   Indiv *indx ;
   double *gpt ;

   tag = cupt -> tagnumber ;
   if (tag == tagnum) return 0 ;
    tagnum = tag ;   
    for (j=0; j<numindivs; j++) { 
     indx = indivmarkers[j] ;
     if (indx -> ignore) continue ;
     gpt = gammavecs[tag][j] ;
     copyarr(gpt, indx -> totgamms, 3) ;
    }
}

void printsnpsx(char *snpoutfilename, SNP **snpm, int num, int printfake, int printvalids) 
// print markernumbers       
{ 
  int i, ppos ;
  SNP *cupt  ;
  char ss[10] ;
  FILE *xfile ;
  int numv ;  

  if ((snpoutfilename != NULL) && (strcmp(snpoutfilename, "NULL") == 0))  return ;
  if (snpoutfilename != NULL)  {
   openit(snpoutfilename, &xfile, "w") ;
  }
  else xfile = stdout ;

  fflush(stdout) ;
  for (i = 0; i <num; ++i) {  
     cupt = snpm[i] ;
     if (!printfake && (ignoresnp(cupt))) continue ;
     fprintf(xfile, "%4d %20s %3d ", i, cupt->ID, cupt -> chrom) ;
     fprintf(xfile, "%15.6f %15.0f",  cupt -> genpos, cupt -> physpos) ;
     fprintf(xfile, " %5d ", cupt -> popA_nn[0]) ;
     fprintf(xfile, "%5d ", cupt -> popA_nn[1]) ;
     fprintf(xfile, "%5d ", cupt -> popB_nn[0]) ;
     fprintf(xfile, "%5d", cupt -> popB_nn[1]) ;
     if (!printvalids) { 
      fprintf(xfile, "\n") ;
      continue ;
     }
     numv = numvalidgtypes(cupt) ;
     fprintf(xfile, "%6d",numv) ;
     fprintf(xfile, " %d %d %d\n", cupt -> ignore, cupt -> isfake, cupt -> isrfake) ;
  }
  if (snpoutfilename != NULL) 
    fclose(xfile) ;
}

int
grabgeasy(int *genos, int *gtypes, int *idnum, int *mreg, int n) 
{
  int i, t, a, x, z, g, totnindivs, kmax, kmin ;
  int *tgenos, *tgtypes ;
  SNP *cupt ;
  Indiv *indx ;

  totnindivs = parnumindivs + numindivs ;
  ZALLOC(tgenos, totnindivs*n, int) ;
  ZALLOC(tgtypes, totnindivs, int) ;
  ivclear(tgtypes, -1, totnindivs) ;
  for (a=0; a<n; a++) {  
   x=mreg[a] ;

   cupt = snpmarkers[x] ;
   if (cupt -> ngtypes == 0) {
     free(tgenos) ;
     free(tgtypes) ;
     printf("zz2 %d\n", x) ;
     return 0 ;
   }
   for (i=0; i<numindivs; i++) { 
    indx = indivmarkers[i] ;
    tgtypes[i] = indx->affstatus ;
    if (indx -> ignore) tgtypes[t] = -1 ; 
    tgenos[i*n+a] = cupt -> gtypes[i] ;  
   }
  }
  z = 0 ;
  for (i=0; i<totnindivs; i++) { 
   if (tgtypes[i]<0) continue ;
   ivmaxmin(tgenos+i*n, n, &kmax, &kmin) ;
   if (kmin<0) continue ;
   gtypes[z] = tgtypes[i] ;
   idnum[z] = i ;
   copyiarr(tgenos+i*n, genos+z*n, n) ;
   ++z ;
  }
  free(tgenos) ;
  free(tgtypes) ;
  return z ;
}
void
loadcc(double *ch2, int *gobs, int *iscase, int neq) 
{
   double *cc ; 
   int i, g, c ;

   vzero(ch2, 4) ;
   for (i=0; i<neq; i++) { 
    g = gobs[i] ;
    c = iscase[i] ;
    cc = ch2 ; 
    if (c==1) cc = ch2+2 ;
    cc[0] += g ;
    cc[1] += 2-g ;
   }
}


