#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <nicklib.h>
#include <getpars.h>

#include "mcio.h"  
#include "regsubs.h" 
#include "logsubs.h" 

int useparentals = NO ;

extern int verbose ;
extern SNP **snpmarkers ;
extern int numsnps ;
extern Indiv **indivmarkers;
extern int numindivs  ;
extern int numgvals ;  

void calcY(SNP **snpm, int i, int j, double *X, int numstates) ;
void calcX(SNP **snpm, int i, int j, double *X, int numstates) ;

int cnvt4(int x)  ;

double
logrmark(SNP *cupt, Indiv **indivmarkers, int numindivs) 
{
  int nv, iscase ; 
  int snpnum ; 
  int *gtypes ;
  Indiv *indx ;
  double *eq, **rhs, anc, ee[3], rr[2], ysc[3], ans[3] ;
  double *xeq, *srhs ; /*for simple regression */
  char **idnames ;
  char *popAname = "PopA" ;
  char *popBname = "PopB" ;
  double ychi, ggr[3], gg[3] ;
  int neq, g, ind, nvar, ppos ;
  
  if (cupt -> gtypes == NULL) return 0.0 ;
  if (cupt ->ngtypes == 0) return 0.0 ;

  gtypes = cupt -> gtypes ;
  snpnum = cupt -> markernum;  
  nv = 3;  

  ZALLOC(eq, numindivs*nv, double) ;
  ZALLOC(xeq, numindivs*nv, double) ;
  ZALLOC(srhs, numindivs, double) ;
  ZALLOC(idnames, numindivs, char *) ;
  rhs = initarray_2Ddouble(numindivs, 2, 0.0);

  neq = 0 ;
  for (ind =0; ind <numindivs; ++ind) {  
   indx = indivmarkers[ind] ;
   if (indx->ignore) continue ;
   g = gtypes[ind] ;
   if (g<0) continue ;
   if (numgvals == 4) g = cnvt4(g) ;
   anc = indx -> thetatrue ;
   if (anc<0.0) continue ;

   if (cupt -> isrfake) {  
    calcY(snpmarkers, ind, snpnum, ggr, 3) ;
    vvt(ggr, ggr, indx -> totgamms, 3) ;
    bal1(ggr, 3) ;
    anc = 0.5*(ggr[1] + 2.0*ggr[2]) ;
   }

   iscase = indx -> affstatus ;
   rr[1] = (double) g ;  
   rr[0] = (double) (2-g) ;  // not X
   srhs[neq] = rr[1]/asum(rr,2) ;
   ee[0] = 1.0 ;
   ee[1] = anc ;
   ee[2] = (double) iscase ;
   copyarr(ee, eq+neq*nv, nv) ;
   copyarr(rr, rhs[neq], 2) ;
   idnames[neq] = indx -> ID ;
   ++neq ;
  }

  if (useparentals) {  
   ee[0] = 1.0 ;  
   ee[1] = 0.0 ; 
   ee[2] = 0.0 ;
   rr[0] = cupt -> popA_nn[1] ;
   rr[1] = cupt -> popA_nn[0] ;
   srhs[neq] = rr[1]/asum(rr,2) ;
   copyarr(ee, eq+neq*nv, nv) ;
   copyarr(rr, rhs[neq], 2) ;
   idnames[neq] = popAname ;
   ++neq ;

   ee[1] = 1.0 ;
   rr[0] = cupt -> popB_nn[1] ;
   rr[1] = cupt -> popB_nn[0] ;
   srhs[neq] = rr[1]/asum(rr,2) ;
   copyarr(ee, eq+neq*nv, nv) ;
   copyarr(rr, rhs[neq], 2) ;
   idnames[neq] = popBname ;
   ++neq ;
  }

  if (verbose) 
   printlreg(eq, rhs, idnames, neq, nv) ;
  for (nvar=1; nvar <= nv; nvar++) { 
   squish(xeq, eq, neq, nv, nvar) ;
   ysc[nvar-1] = regressit(ans, xeq, srhs, neq, nvar) ;
   printmat(ans, 1, nvar) ;
  }

  ppos = cupt -> physpos ;
  printf("marker: %20s ppos: %12d   score1: %9.3f score2: %9.3f",  
    cupt -> ID, ppos, ysc[1]-ysc[0], ysc[2]-ysc[1]) ;
  printf("\n") ;
  printf("simple regression coeffs:\n") ;  
  printmat(ans, 1, nv) ;

  squish(xeq, eq, neq, nv, nv-1) ;
  ysc[0] = logregressit(ans, xeq, rhs, neq, nv-1) ;
  ysc[1] = logregressit(ans, eq, rhs, neq, nv) ;

  ychi = 2.0*(ysc[1] -ysc[0]) ;
  printf("%20s neq: %4d logistic_score: %9.3f\n", cupt -> ID, neq, ychi) ;
  printmat(ans, 1, nv) ;

  free(eq) ; 
  free(xeq) ; 
  free2D(&rhs, numindivs) ;
  free(srhs) ;
  free(idnames) ;
  return ychi ;
}

void printlreg(double *eq, double **rhs, char **idnames, int neq, int nv) 
{
  int i, j ;

  printf("##lreg  neq: %5d   nv:  %5d\n", neq, nv) ;
  for (i=0; i<neq; i++)  {  
   for (j=0; j<nv; j++)   {  
    printf("%12.6f ", eq[i*nv+j]) ;
   }
   printf("   %6.0f %6.0f", rhs[i][0], rhs[i][1]) ;

   if (idnames != NULL)  {
    printf(" %s", idnames[i]) ;
   }

   printf("\n") ;
  }
}

double
emscore(SNP *cupt, Indiv **indivmarkers, int numindivs, 
  double *pa, double *pe, double *ppa, double *ppe) 
{

  double ytot, ycase, ycontrol ;

  ytot = emscorex(cupt, indivmarkers, numindivs, 2, pa, pe) ;
  ycase = emscorex(cupt, indivmarkers, numindivs, 1, pa, pe) ;
  ycontrol = emscorex(cupt, indivmarkers, numindivs, 0, ppa, ppe) ;

  return 2.0*(ycase + ycontrol-ytot) ;

}

double
emscorex(SNP *cupt, Indiv **indivmarkers, int numindivs, int mode,
  double *pa, double *pe) 
// MLE scoring very similar to freqcheck
{
  int nv, iscase ; 
  int snpnum ; 
  int *gtypes ;
  Indiv *indx ;
  int na[2], nb[2], *gobs ;  
  double **gams ;
  double ylike, anc, ggr[3] ;
  int neq, g, ind ;
  int *indlist ;
  static int ncall = 0 ;

  ++ncall ;
  
  if (cupt -> gtypes == NULL) return 0.0 ;
  if (cupt ->ngtypes == 0) return 0.0 ;

  gtypes = cupt -> gtypes ;
  snpnum = cupt -> markernum;  
  nv = 3;  

  gams = initarray_2Ddouble(numindivs, 3, 0.0);
  ZALLOC(gobs, numindivs, int) ;

  neq = 0 ;
  for (ind =0; ind <numindivs; ++ind) {  
   indx = indivmarkers[ind] ;
   if (indx->ignore) continue ;
   g = gtypes[ind] ;
   if (g<0) continue ;
   anc = indx -> thetatrue ;
   if (anc<0.0) continue ;
   iscase = indx -> affstatus ;
   if ((mode<2) && (iscase != mode)) continue ;
   copyarr(indx -> totgamms, gams[neq], 3) ;
   if (cupt -> isrfake) {  
    calcY(snpmarkers, ind, snpnum, ggr, 3) ;
    vvt(gams[neq], ggr, gams[neq], 3) ;
   }
   bal1(gams[neq], 3) ;
   gobs[neq] = g ;
   ++neq ;
  }

  if (mode==1) {  
   ivclear(na, 0, 2) ;
   ivclear(nb, 0, 2) ;
  }

  else { 
   na[0] = cupt -> popA_nn[1] ;
   na[1] = cupt -> popA_nn[0] ;
   nb[0] = cupt -> popB_nn[1] ;
   nb[1] = cupt -> popB_nn[0] ;
  }

  printf("xpest: mode %d\n", mode) ;
  ylike = xpest(gams, gobs, na, nb,  
   neq, pa, pe) ;
  printf("xpestout: mode %d %12.3f\n", mode, ylike) ;

  free2D(&gams, numindivs) ;
  free(gobs) ;
  return ylike ;
}

// copied from mcsubs.c this should be moved to a common code block
void calcY(SNP **snpm, int i, int j, double *X, int numstates)
{
   int t ;
   SNP *cupt ;

   cupt = snpm[j] ;
   t = cupt -> isrfake ;
   cupt -> isrfake = NO ;
   calcX(snpm,  i, j, X, numstates) ;
   cupt -> isrfake = t ;
}

void calcX(SNP **snpm, int i, int j, double *X, int numstates)
// no magic in logreg
{
  int x,k;
  double PopA_freq, PopB_freq;
  SNP *cupt ;
  Indiv *indx ;

  if ((j>=numsnps) || (j<0)) fatalx("(calcX) bounds check failure %d\n",j) ;

  vzero(X,3) ;

  cupt = snpm[j] ;
  if ((cupt->ignore) || (cupt -> isrfake)) { 
   vclear(X, 1.0, numstates) ;
   return ;
  }
  if (cupt -> ngtypes == 0) {  
   vclear(X, 1.0, numstates) ;
   return ;
  }
  
  x = cupt -> gtypes[i] ;

  if (x<0) {  
   vclear(X, 1.0, numstates) ;
   return ; 
  }

  PopA_freq = cupt -> adm_PopA_freq;
  PopB_freq = cupt -> adm_PopB_freq;

  if ((PopA_freq<0) || (PopB_freq<0)) 
    fatalx("marker freqs %s not set\n", cupt -> ID) ;

  if (numstates == 3) {
    switch(x) {
      
    case 2:
      X[0] = PopA_freq * PopA_freq;
      X[1] = PopA_freq * PopB_freq;
      X[2] = PopB_freq * PopB_freq;
      break;
      
    case 1:
      X[0] = 2 * PopA_freq * (1 - PopA_freq) ;
      X[1] = PopB_freq * (1 - PopA_freq) + 
	(1 - PopB_freq) * PopA_freq ;
      X[2] =  2 * PopB_freq * (1 - PopB_freq) ;
      break;
      
    case 0:
      X[0] = (1 - PopA_freq) * (1 - PopA_freq);
      X[1] = (1 - PopA_freq) * (1 - PopB_freq);
      X[2] = (1 - PopB_freq) * (1 - PopB_freq);
      break;
  
    default:
      fatalx(" %s %d %d genotype unset\n", cupt->ID, i, x) ;
      break;
    }
  }
  else if (numstates == 2) {
    switch(x) {

    case 2:
      X[0] = PopA_freq;
      X[1] = PopB_freq;
      break;

    case 0:
      X[0] = (1 - PopA_freq);
      X[1] = (1 - PopB_freq);
      break;
  
    default:
      printf("zzzunset %d %d %d\n", i, x, j) ;
      indx = indivmarkers[i] ;
      printf(" gender: %c  chrom: %d\n", 
        indx -> gender, cupt -> chrom) ;
      fatalx(" %s %d %d genotype unset\n", cupt->ID, i , x) ;
      break;
    }
  }
}

double
xpscore(SNP *cupt, Indiv **indivmarkers, int numindivs, 
   double pm , double pr) 
{
  int nv, iscase ; 
  int snpnum ; 
  int *gtypes ;
  Indiv *indx ;
  int na[2], nb[2], *gobs ;  
  double **gams ;
  int *iscasex ;
  double ylike, anc, ggr[3] ;
  int neq, g, ind ;
  static int ncall = 0 ;
  double yscore, pmesh, rmesh  ; // LOD  

  ++ncall ;
  
  if (cupt -> gtypes == NULL) return 0.0 ;
  if (cupt ->ngtypes == 0) return 0.0 ;

  gtypes = cupt -> gtypes ;
  snpnum = cupt -> markernum;  
  nv = 3;  

  gams = initarray_2Ddouble(numindivs, 3, 0.0);
  ZALLOC(gobs, numindivs, int) ;
  ZALLOC(iscasex, numindivs, int) ;

  neq = 0 ;
  for (ind =0; ind <numindivs; ++ind) {  
   indx = indivmarkers[ind] ;
   if (indx->ignore) continue ;
   g = gtypes[ind] ;
   if (g<0) continue ;
   anc = indx -> thetatrue ;
   if (anc<0.0) continue ;
   iscase = indx -> affstatus ;
   if ((iscase>1) || (iscase<0)) continue ;
   iscasex[neq] = iscase ; 
   copyarr(indx -> totgamms, gams[neq], 3) ;
   if (cupt -> isrfake) {  
    calcY(snpmarkers, ind, snpnum, ggr, 3) ;
    vvt(gams[neq], ggr, gams[neq], 3) ;
   }
   bal1(gams[neq], 3) ;
   gobs[neq] = g ;
   ++neq ;
  }

   ivclear(na, 1, 2) ;
   ivclear(nb, 1, 2) ;

  pmesh = pm ;
  rmesh = pr ;

  yscore  = xpbayes(gams, gobs, na, nb,  
   iscasex,  neq, pmesh, rmesh ) ;

  free2D(&gams, numindivs) ;
  free(gobs) ;
  free(iscasex) ;
  return yscore ;
}


int cnvt4(int x) 
{

 if (x<0) return x ;
 if (x==0) return 0 ;
 if (x==3) return 2 ;
 return 1 ;

}
