#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <nicklib.h>
#include <getpars.h>

#include "arms.h"  
#include "toysim.h" 
#include "mcio.h"  
#include "mclm.h"  
#include "qsubs.h" 
#include "badpairs.h" 
#include "gammas.h" 
#include "tlest.h" 
#include "mcmcpars.h" 

#define WVERSION   "6210"
//AT: 070605: Incorporating Nick's 2923 changes
//AT:061105: Incorporating Nick's 2940 and 2950 version changes
//AT:060305: Incorporating Nick's changes from 2920 version
//AT: Incorporating the changes made to the output format 
// leave1 a bit more careful about unvisited markers
// leave1 scratchdir now created if needed
// keeptrashfiles added
// 2D toysim now supported
// bug in printindivs fixed
// markerlistname -> badsnpname
// indx -> ignore bug fixed
// NULL switches off printindivs, printsnps
// printgall added
// substantial code added to support isrfake
// checkindiv added (reich test)
// check added for whether unzip worked
// lmsctop writes out SNP
// error trapping for nreal, nindiv added
// magic stuff (external probs added)
// return value checked (leave1)
// rpower (how much of entropy recovered) added
// markername code added in printindivs  
// isrfake check added to sampfreqx 
// lmsieve allows us to make fake risks from real data. 
// isrfake bug fixed in mcio.c 
// nickscore published with lmscore    
// fflush added in bigiter loop
// thpars added in readco
// tiny bug fixed in fatalx message
// serious bug fix: ignored individuals used for neil score  
// print lod score by individual for given marker
// print genotype for given marker
// check for gtypes NULL prints warning in printindivs
// no data for marker => ignore set 
// subtle bug in sampprbb fixed.  Implements "ray sampling"
// xlgamma added but now removed
// bounds loop allocation bug fixed
// fake printed for rfake markers in pubxscores
// bug in pubxname output fixed 
// readgdata added to mcio (not needed for mcmc) 
// modifications made in mcsubs for rfake 
// ggr is "gamma" including observations under rfake marker
// bug fixed in accumulation of (a,g) counts for rfake
// no threshold for rfakes in computation of score in mclm.c 
// numvalids -> admutils.c                                  
// alldata restored in readcommands()                                 
// LOD score at marker supports weighted model 
// release 7 will support QTL (DONE)
// markersim not set => markernum used
// printsnps changed to print valid case/control genotypes
// xranbeta (sampgw.c) fixed using rejection sampling method 
// fancyxtheta switched on  (NO => thxp0 = -1) 
// metropolis sampling for sampprb (sampgw.c)
// ray sampling done correctly for beta priors (samprb)
// prior introduced on gamma hyperprior 
// extensive changes to mclm.c  (oldlmmode = NO) 
// lmbayes calculation added.
// local score (detailed output -> localoutfilename)
// BUG found (mk2from2 returned value treated as int on mcmc.c; Caused stack problems
// more care taken with nopopsmode  and cleaninit
// redundant code (ppp) in mclm.c cleaned up
// print lambda in printmodelscores (mclm)
// Bug found in 2D scoring.  xlam incorrectly calculated so that R(a=0) was not 1
// freqcheck modified if no controls (leave1.c)
// leave1mode = NO switches out doleave1  if checkit
// dupmode = NO switches out dupcheck   if checkit
// int printtoygenoout invalids not written
// badpairs, killsnps implemented  .  goodsnpname added
// bug fixed.  Last interval of chromosome not counted for lambda estimation
// extensive support for rdismode  
// bug fixed (gpsum, gpnum) 
// loodpdis now doesn't load zero distances 
// fixed packmode bugs in toysim
// small fixups in pubfreqs printgall
// small bug (printing lmscore)
// bug fixed (calling pubgamms)
// bad bug (betaprior xsubb2 fixed) 
// paramnames compatible with smartsrc release
// some code cleanup thanks to mac compiler
// check made for invalid geno value in getgenos
// fastdup added (automatic in checkit mode)
// small bug fix in fastdupcheck
// (March 07 new round of bigfixes (Arti) )                           
// loclip hiclip controls clipping  

/** 

 for rpower at each MCMC iteration we compute the entropy of ethnicity
summed over all individuals (X) and the sum of the entropy of the gammas
at each marker (G).  We show the average of 1-X/G.  This will be between
0 and 1 in practice although can be < 0 in theory.  

When entropy is 0 we get 1. When entropy is G we get 0.

*/


#define TAUINIT 100.0

double  ltmean = 1, ltsig = 0.5 ;
double  lxmean = 1.5, lxsig = 0.5 ;

/**  priors on  log10 (a+b), log10 (a+b+c) in sampprb, samprbb */

double tauinit = TAUINIT ;

char *thetafilename  = NULL ;
char *lambdafilename = NULL ;
char *freqfilename = NULL ;
char *indoutfilename = NULL ;
char *snpoutfilename = NULL ;
char *localoutfilename = NULL ;
char *badpairsname = NULL ;
char *snpscorename = NULL ;


char *magicfilename = NULL ;
char *magicname = NULL ;  //  name of marker
int magicnum = -1 ;   
char *ethnicfilename = NULL;
char *genooutfilename = NULL;
char *gammafiles = NULL ; // for dumpgams, readgams

static int cleaninit = YES ;
static int checkit = NO ;  
int  fastdup = NO ;
int fastdupnum = 10 ;
int  leave1mode = YES ;  // only used for checkit 
int  dupmode = YES ;  // only used for checkit 
extern int  rdismode  ;  // reestimate distances  
double rdis = 100.0 ; // distance at 1 cM exponential  (very soft prior)
int  usephyspos = NO ;  // genetic map ignored on input 

int  dumpgammas = NO ;
int  corrgammas = NO ;
extern int packmode ;
static int printmut = NO ;  

int debug = NO ;
int verbose = NO ;
int oldlmmode = NO ;
int randombadpairs = NO ;
int markerpub = -1 ;
/** 
 for debug style info
 also gamms at marker
 see leave1 for example
*/

/** 
 good idea on new data 
 checkit = YES 
 numbigiter = 0  
*/

int alldata = YES ; /* used in toy stuff */

#define MAXSTR  512

SNP **snpmarkers ;
SNPMC **snpmcmarkers, **modmcmarkers ; 

int nummark, numsnps ;

Indiv **indivmarkers;

int numindivs, numindok, numindxok ;


/** parameters */

  char *parname = NULL ;
  int lmmodel  = NO ;
  int lmchrom = -1 ;
  double lmthresh = 2.0 ;  
  int lmdetails = NO ;  
  int twomodel = NO ;
  int lmnumx = 13 ;
  double lmmax = 1.5 ;
  int lmlobase = -1 ; 
  int lmhibase = 999999999 ;
  int keeptrashfiles = NO ;
  
  char *trashdir = "/var/tmp" ;

  char *infile=NULL ;
  char *scorename=NULL ;
  char *score2name=NULL ;
  char  *genotypename = NULL ;
  char *snpname = NULL ;
  char *indivname = NULL ;

  char  unknowngender = 'U' ;

  FILE *scorefile=NULL ;
  FILE *score2file=NULL ;
  char *badsnpname = NULL ;
  char *goodsnpname = NULL ;   // for badpairs processing only 

  int mincasenum = 1 ;
  int maxmark, xchrom ; 
  int *fchrom, *lchrom ; /* first last numbers on chromosome */
  int runmode=0 ;
  int theta2mode = NO ;
  int lambda2mode = 0 ; 
  int noxdata = NO ;
  int numlagprint = 10 ;
/* 
   0 one lambda 
   1 alpha max 
   2 alpha 0 
*/
   
  int splittau = YES ;
  int tlreest = YES ;
  int nopopsmode = NO ;
  int popfreqmode = YES ;
  int pubfmodern = NO ;
/* in pubfreq modern or ancestral... */
  int details = YES ;
  int dotoysim = NO ;
  int allmale = NO ;  /* for PC sims */
  int numbigiter = 5 ;
  int numburn = 1 ; 

  double taulmean = 2.0 ;
  double taulsdev = 0.5 ;
  double tauscal[2] ;
  double pflat = 0.0 ;
 /* used to flatten initial probabilities */

  double yerr ;
  double riskarr[MAXNUMR] ;
  double risk2arr[MAXNUMR] ;
  double criskarr[MAXNUMR] ;
  double epithresh = 3.0 ;
  double epirisk[3] ;  
  char  *epinames = NULL ;
  double modelsc[MAXNUMR] ;
  double *qbasemodelsc ;
  double tmodelsc[MAXNUMR] ;
  double wrisk[MAXNUMR] ;  //  weights
  double lrisk = -1 ;
  int numrisks, qnumrisks ; 
  double risksim  = -99.0 ;
  double rhosim   = -99.0 ;
  int markernum = -1 ; 
  int markersim = -1 ; 
  double sim2dvals[4] ;  
  int sim2d_caseonly = NO ;
  int qtmode = NO ;
  int probitmode = YES ;
/* model scores are used as scores against qtrbase model */
  double *qtrho, *qtrisks ;  
  

// for 2d risk model 
// we specify allele prob (PopA, PopB) and 
// rlambda, rmu for 2d risk model  
// no data for controls at risk allele for sim2d

  char *markername = NULL ;
  int allcases = NO ;
  int usecontrols = YES ;
int lmsieve = NO ;   
/**    
   if yes then risksim is used with overdominant model to  
   retain cases preferentially.   
   See dolmsieve  
*/ 
#define  NPUBXA  10

  int pubxa[NPUBXA], npubxa=-1 ;
  char *pubxname = NULL ;
  char *pubxindname = NULL ;

  int ncases = -1, ncontrols = -1 ;
  int simnumindivs = - 1 ;
  int emiter = 30 ;
  int reestnumiter = 1 ;
  double allele_scale_fac = -1 ;
  double fakespacing = 0.01 ; //Morgans

/* end parameters */ 
  double plo = 0.001 ;
  double phi = 0.999 ;

  double ltaulo = 0.5 ;
  double ltauhi = 6.0 ; 

  double *zstat, *zstatsum ;
  double *wr = NULL ;


void readcommands(int argc, char **argv) ;

int countfake(SNP **snpmark, int numsnp) ;
int countnofake(SNP **snpmark, int numsnp) ;
int countrfake(SNP **snpmark, int numsnp) ;

void setug(Indiv **indn, int num, char gender) ;
void clearsnp(SNP *cupt) ;
void clearind(Indiv **indm, int num) ;
void setfc(SNP **snpm, int numsnps)  ;

void pubaux(FILE *fff, int iter)  ;
void pubscoresx (double *basescore, double *zstat) ;

void ckit(int k)  ;

void docalcXX(int x, double *X, double **XX,
 double p, double pp, int numstates) ;

int nextreal (int start, int end, int inc)  ;

double 
domc1(double *totscore, double *zstat, SNP **snpmarkers, SNPMC **snpmcmarkers, 
 int numsnps, int numindivs, 
 double *riskarr, double *risk2arr, double *criskarr, 
 double *modelscores,  double *wrisk, 
 int numrisks, 
 double *tave, double *lave, double *xtave, double *xlave,
 int pubxindiv, double **pubgams)  ;

void sampfreqx(double *tscal, SNP **tsnpmarkers, SNPMC **snpmcmarkers, 
  int numsnps, int mode)  ;

void sampfreqy(double *tscal, SNP **tsnpmarkers, SNPMC **snpmcmarkers, 
  int numsnps, int mode)  ;

/** 
 sampfreqx samples allele frequencies and tau
 sampfreqy same but for isrfake markers     
*/
 


void dosampfreq(double **xnn, double **ynn, int n, double *palpha, 
 double *r1, double *r2, int taureestmode) ;

void addpstats(SNP *tcupt, SNPMC *cuptmc)   ;

double xranbeta(double xa, double xb, double lo, double hi) ;

double sampit1c(double *xn, double alpha, double pp)  ;

double sampprt(double *ptau, double *x1, double *x2, int n, 
  double lmean, double lsig)   ;

void sampprb(double *pa, double *pb, double *qq, double *x, 
 int n, int niter, double ltmean, double ltsig, int mode) ;
void sampprbb(double *pc, double *pa, double *pb, double *x1, double *x2,
 int n, int niter, double lxmean, double lxsig) ;
void sampprg(double *pa, double *pb, double *x, int n, int niter, 
 int mlmode) ;

void samppdis(double *xflat, double *xdis, double *priordis, 
 int n, double lo, double hi)  ;


void prints1(double x) ;
void fprints1(FILE *fff, double x) ;

void tcalcX(SNP *cupt, int i, double *X, int numstates) ;



double xtaulike(double tau, double *x1, double *x2, int n) ;
double taulike(double tau, SNP **snpm, SNP **tsnpm, int n, int popmode) ;

double tauliked(double tau, SNP **snpm, SNP **tsnpm, int n,
  double *yd1, double *yd2, int popmode) ;



void estglob(double *xx, int n, int iter) ;


void dostat(char *legend, double *ss, int n)  ;

void setinit(SNP **snpm, Indiv *indx) ;

void printindivs() ;

void prmstat(int iter, double *xx, int n) ;  

void
esttheta(double *qq, double *tt, double *ttx, int n, int niter, int xchrom)  ;


void initindiv(Indiv **indm, int numindivs)  ;

void
pubindivs(double *ti, double *ti2, double *txi, double *txi2, 
     double *li, double *li2, double *lxi, double *lxi2, int n)  ;
void pubfreqs(SNPMC **snpmcmarkers, int n) ;

void
calcmsdev(double ysum, double ysum2, double num, double *mean, double *sdev)  ;


void mk2from3(double *d, double *c, double p, double pp) ;
void mk2from2(double *d, double *c) ;

void gx(double *x, double y, double tau) ;

double mufun(double theta) ;
double mufun1(double theta) ;

double
thetaxsum(Indiv **indivmarkers, int n)  ;

void printmutab(double *txsa, int n)  ;

void
pubgammas(double **totgammas, int pubx, 
 int fc, int lc, char *pubxname)  ;

void
 emsimple( SNPMC **snpmcmarkers, 
 int numsnps, int numindivs, int numiter, double *zmax, double *zmin) ;

void
scalecounts(SNP **snpm, int numsnps, double fac) ;

void scalitx(int *nn, double fac, double p) ;


void printtmeans(double * tmean) ;

void countgtypes(SNP *cupt, int *xcnt)  ;

void dophyscheck (SNP ** snpm, int numsnps) ;
void dohwcheck (SNP ** snpm, int numsnps, int vmode)  ;
void dohetxcheck (SNP ** snpm, Indiv **indivm, int numsnps)  ;
void checkindiv (SNP **snpm, Indiv **indm, int numsnps, int numindivs)  ;

void checkgeno() ;
void checkdup(SNP **snpmark, int numsnp, int numindivs) ;
void leave1(SNP **snpmarkers, int numsnps, int numindivs, 
 double lrisk, char *trashdir);
void indstatus(char *sx, Indiv *indx) ;
int sqz(double *xout, double *xin, double *xp, int n) ;
void printgall (char *pubxname) ;
void dolmsieve() ; 
int numvalidgt(Indiv **indivmarkers, SNP *cupt) ; 
void corrgamma (SNP **snpm,  int numsnps, int numindivs, 
  double winlen, double scorethresh, double epithresh)  ;

void pubdis (SNP **snp, SNPMC **snpmc, int n) ;
void setgphys(SNP **snpm, int nsnps)  ; 
double zprobg(double a, double x)  ;
  

int main(int argc, char **argv)
{
  SNPMC **mcmpt ;    
  int numstatesg = 3;
  int numstates ;
  int i, j, k, n, minct,maxct,numpermutes=0, numsc;

  SNP currmarker;
  SNP *cupt, *tcupt ;
  SNPMC *cumcpt, *ttmcpt ; 
  Indiv *indx ;

  double **gammas, **totgammas ;
  double **alphas, **betas ;
  int  nmark=0, chrom ;
  double *basescore ;
  double *topscore, *botscore ;
  double fac, totfac, tmp, tmp2, tmp4, zmax, zmin ; 
  double ylike, ybase, y, y1, y2  ;
  int numit = 0 ;

  double bestpos[MAXCH+1] ;
  double bestphys[MAXCH+1] ;
  int bestmark[MAXCH+1] ;
  int numnofake, numfake, numrfake ;
  int ncnt ;
  double p0, pp0, p1, pp1 ;
  double gg[3], dd[4], dd2[4], p, pp, xa, xb, q ;

  double *xscore, *facarray ; 
  double avelogfac, sumfac ;
  double totscore ;

  double txsum ; /* for mufun etc */
  double *txsa ;
  int bigiter, reestiter ;
  int indiv, x, fc, lc ;

#define NSTATS  3
#define MSTATS  18
#define MMEAN   11
  double *mstat ; /* for monitoring */
  int mprint ;
  double **stats, tstat[NSTATS] ;
  double dis, lpos ;
  double tmeans[MMEAN] ; /* posterior means of theta etc */

/** 
 ethnic is sequence of hidden states 
 for individuals 
 tscore is 
 log-factor in favor of disease allele 
 totscore is accumulation
 ax, bx aax bbx are beta dist. parameters (real markers)
 xc[marker] is 9x9 count matrix 
 counting 3*e + g where g is genotype e is ethnicity
*/

  double tave, lave, xtave[2], xlave[2] ;
  double *ti, *txi, *li, *lxi ;
  double *ti2, *txi2, *li2, *lxi2 ;
  int mode ;

  int nignore ;
  double ccr[3], yransc, ysc, qmu ;  

  vclear(wrisk, 1.0, MAXNUMR) ; 
  numrisks = 1 ;

  ZALLOC(qtrho, MAXNUMR, double) ;
  ZALLOC(qtrisks, MAXNUMR, double) ;
  qtrho[0] = qtrisks[0] = -99.0 ; /* silly value */
  
  epirisk[0]  = epirisk[1] = 1.0 ; 
  epirisk[2]  = 4.0 ;
  
  packmode = NO ;

  readcommands(argc,argv) ;
  printf("## ANCESTRYMAP version: %s\n", WVERSION) ;
  if (qtmode) {
   numrisks = qnumrisks ;
   if (risksim < -10.0) risksim = 0.0 ;
   if (rhosim < -10.0)  rhosim = 0.0 ;
  }
  else { 
   if (risksim < -10.0) risksim = 1.0 ;
  }
  if (parname==NULL) return 0 ;
  if (numrisks==0) fatalx("no risk model\n") ;

  if (keeptrashfiles) {
   printf("trash files will be kept\n") ;
  }

  if (rdismode == YES) packmode = YES ;

  numsnps = 
    getsnps(snpname, &snpmarkers, fakespacing, badsnpname, &nignore, numrisks) ;


  setfc(snpmarkers, numsnps) ;

  if (usephyspos) setgphys(snpmarkers, numsnps) ;


  if (simnumindivs<0) {
   numindivs = getindivs(indivname, &indivmarkers) ;
   printf("calling setstatus\n") ;
   setstatus(indivmarkers, numindivs, "Case") ;

   if (unknowngender != 'U') 
    setug(indivmarkers, numindivs, unknowngender) ;


   getgenos(genotypename, snpmarkers, indivmarkers, 
    numsnps, numindivs, nignore) ;
  


   if (badpairsname != NULL) {
    if (snpscorename != NULL)  {
     getsnpsc(snpscorename, snpmarkers, numsnps); 
     dogood(goodsnpname, snpmarkers, numsnps) ;
    }
    else {
     loadbadpsc(snpmarkers, numsnps, randombadpairs, goodsnpname) ;
    }
    dobadpairs(badpairsname, snpmarkers, numsnps) ;
    x = killsnps(indivmarkers, snpmarkers, numsnps, mincasenum) ;
    printf("badpairs removed\n") ; 
    printf("snps killed: %d\n", x) ;
   }
   if (debug) printindivs() ;
  }


  rmsnps(snpmarkers, numsnps, nignore) ;  //  rid ignorable snps
     printsnps(snpoutfilename, snpmarkers, numsnps, 
      indivmarkers, NO, YES) ;

  if (rdismode) setepath(snpmarkers, numsnps) ;  //  sets packepath,  cupt -> ebuff  

  if (lmmodel) {  
    printf("lmchrom: %d\n", lmchrom) ; 
    if((lmchrom == 0) || (lmchrom > 23))
      fatalx("Invalid value for the parameter lmchrom %d\n",lmchrom);

    printf("lmchrom: %d\n", lmchrom) ;
    lrisk = riskarr[0] ;
    setlm(snpmarkers, numsnps, lmchrom) ;
    if (numrisks>1) fatalx("If the lmmodel mode is set cannot have numrisks >1\n") ;
  }

  if (qtmode) { 
   if (qtrho[0] < -10.0) vclear(qtrho, qtrbase, numrisks) ;
   if (qtrisks[0] < -10.0) fatalx("qtrisk not set\n") ;       
   ZALLOC(qbasemodelsc, numrisks, double) ;
  }
  else probitmode = 0 ;

  
  if (simnumindivs>0)  {
   dotoysim = YES ;
   numindivs = simnumindivs ;
   for (i=0; i<numsnps; i++) { 
    cupt = snpmarkers[i] ;
    if (cupt -> isfake) continue ;
    cupt -> ngtypes = numindivs ;
    ZALLOC(cupt -> gtypes, numindivs, int) ;
    ivclear(cupt -> gtypes, 0, numindivs) ;
// bugfix else mkgtypes in toysim.c always sets gtypes to -1
   }
  }

  if (simnumindivs>0)  {
   ZALLOC(indivmarkers, numindivs, Indiv *) ;
   for (i=0; i < numindivs; i++) {  
    ZALLOC(indivmarkers[i], 1, Indiv) ;
    indx = indivmarkers[i] ;
    sprintf(indx->ID,"%s:%d","toyindiv",i) ;
   }
   clearind(indivmarkers, numindivs) ;
   if (ncases<0) {  
    ncases = simnumindivs/2 ;
    ncontrols = simnumindivs - ncases ;
    mode = 1 ;
   }
   else mode = 2 ;
   printf("calling genindiv; ncases:  %d\n", ncases) ;
  }
  if (qtmode) {  
   ncases = ncases+ncontrols ;
   ncontrols = 0 ;
  }
  if (simnumindivs>0) {
   genindiv(indivmarkers, ncases, ncontrols, mode, allmale) ;
  }


  if (pubxindname != NULL) {  
   pubxindiv = indindex(indivmarkers, numindivs, pubxindname) ;
   if (pubxindiv < 0) 
    fatalx("pubxindname:  %s not found\n", pubxindname) ;
  }

  if (markername != NULL)  {  
   markernum = snpindex(snpmarkers, numsnps, markername) ;
   if (markernum<0) fatalx("markername %s not found\n",markername) ;
  }

  if (markernum>=0) {
    if (markernum>=numsnps) fatalx("markernum out of range!\n") ; 
   cupt = snpmarkers[markernum] ;
   printf("markername: %s\n", cupt -> ID) ;
   if (cupt -> ignore) fatalx("... but set to ignore!\n") ;
   if ((dotoysim==YES) && (markersim<0)) markersim = markernum ;
  }

  if (dotoysim) { 
   printf("generating toy data markersim: %d\n", markersim) ;
   if (epinames==NULL) {
    dosim(snpmarkers,  indivmarkers, numsnps, numindivs, risksim, rhosim,
     markersim, tauscal) ;
   }
   if (epinames!=NULL) {
    doepisim(snpmarkers,  indivmarkers, numsnps, numindivs, epirisk, rhosim,
     epinames, tauscal) ;
   }
  }



  if ((dotoysim) || (genooutfilename != NULL))
    {
      printdata(genooutfilename,indoutfilename,snpmarkers,
       indivmarkers,numsnps,numindivs, NO);
    }

  
  if((twomodel < 0 ) || (twomodel > 1))
    fatalx("Invalid value for the boolean parameter twomodel %d\n", twomodel) ;
    
  if ((twomodel == 1) && (numrisks != 2)) { 
   fatalx("twomodel parameter is only supported when we have 2 risk models %d\n", twomodel) ;
  }

  if (lrisk<=0) lrisk = riskarr[0] ;

/** 
 risk used by leave1
*/


  if (risk2arr[0] < 0.0) {  
   vvt(risk2arr, riskarr, riskarr, numrisks) ;
  }

  printf("\n") ;
/**
  printf("%6s %9s %9s %9s\n",  
   "", "risk1", "risk2", "crisk") ;

  for (k=0; k< numrisks; k++)  {  
   printf("%6s %9.3f %9.3f %9.3f\n",  
    "model:", riskarr[k], risk2arr[k], criskarr[k]) ;
  }
  printf("\n\n") ;
*/

  fc = fchrom[23] ;
  lc = lchrom[23] ;

  if (noxdata) {  
   ivmaxmin(lchrom+1, 22, &numsnps, NULL) ;
   ++numsnps ;
   fchrom[23] = 999999 ;
   lchrom[23] = -999999 ;
  }

  numfake = countfake(snpmarkers, numsnps) ;
  numnofake = countnofake(snpmarkers, numsnps) ;
  numrfake = countrfake(snpmarkers, numsnps) ;

  if (numrfake>0) 
   printf("numrfake: %d\n", numrfake) ;


  if (magicname != NULL)  {  
   magicnum = snpindex(snpmarkers, numsnps, magicname) ;
   if (magicnum<0) fatalx("magicname %s not found\n") ;
  }

  if (magicnum>=0) {
   cupt = snpmarkers[markernum] ;
   printf("magicname: %s\n", cupt -> ID) ;
   setmagic(magicfilename, indivmarkers, numindivs) ;
  }

  if (lmsieve) dolmsieve() ;

  printf("emiter: %d\n", emiter) ;
  printf("reestiter: %d\n", reestnumiter) ;
  if (checkit) {
    printf("trashdir: %s\n", trashdir) ;
    dohetxcheck(snpmarkers, indivmarkers, numsnps) ;
    fastdup = YES ;
  }
  else {
   leave1mode = NO ;
   dupmode = NO ; 
  }

  if (markerpub>=0) printf("markerpub:  %d\n", markerpub) ;
  
  if (numrisks < 1) fatalx("bad risk\n") ;

  if (scorename != NULL)
    openit(scorename, &scorefile, "w") ;
  else scorefile = stdout ;
  
  if (score2name != NULL) {
    printf("auxout: %s\n",score2name) ;
    openit(score2name, &score2file, "w") ;
  }
  


// set theta, lambda to sensible.  //  that is clip
  for (indiv = 0 ; indiv <numindivs; indiv++)  {
    indx = indivmarkers[indiv] ;
    setinit(snpmarkers, indx) ;
  }


  numsc = 0 ; /* number of markers scored */
  totfac = 0.0 ;

/** mcmc stuff */
  ZALLOC(snpmcmarkers, numsnps, SNPMC *) ; 
  if (pubfmodern) 
    ZALLOC(modmcmarkers, numsnps, SNPMC *) ; 
  for (i=0; i< numsnps; i++)  {  
     ZALLOC(snpmcmarkers[i], 1, SNPMC) ; 
     cumcpt = snpmcmarkers[i] ;
     cumcpt -> ap1 =  cumcpt -> ap2 = 0 ;
     cumcpt -> cp1 =  cumcpt -> cp2 = 0 ;
     vzero(cumcpt -> gammasum, 2) ;      
     vzero(cumcpt -> gammanum, 2) ;      
     cumcpt -> pubfmodern = NO ; 
     if (pubfmodern) {   
       ZALLOC(modmcmarkers[i], 1, SNPMC) ;  
       modmcmarkers[i] = snpmcmarkers[i] ; 
       modmcmarkers[i] -> pubfmodern = YES ; 
     } 
/* moments from montecarlo */
  }
  cleargpsum(snpmarkers, numsnps) ;


  for (i=0; i < numsnps ; i++)  {  

   cupt = snpmarkers[i] ;
   if (cupt -> isfake) continue ;
   if (cupt -> ignore) continue ;

   tmp = cupt -> PopA_freq ;
   if (tmp==0.0) fatalx("bad PopA_freq %d %d %d %x\n",
     i, cupt -> isfake, cupt -> ignore, cupt) ;

   tmp = cupt -> adm_PopA_freq ;
   if (tmp==0.0) fatalx("bad adm_PopA_freq %d %d %d %x\n",
     i, cupt -> isfake, cupt -> ignore, cupt) ;

   tmp = cupt -> PopB_freq ;
   if (tmp==0.0) fatalx("bad PopB_freq %d %d %d %x\n",
     i, cupt -> isfake, cupt -> ignore, cupt) ;

   tmp = cupt -> adm_PopB_freq ;
   if (tmp==0.0) fatalx("bad adm_PopB_freq %d %d %d %x\n",
     i, cupt -> isfake, cupt -> ignore, cupt) ;

  }


  ZALLOC(xscore, numsnps, double) ;
  ZALLOC(facarray, numsnps, double) ;
  vzero(tmodelsc, MAXNUMR) ;

  vzero(facarray, numsnps) ;
  ZALLOC(zstat, numsnps, double) ;
  ZALLOC(zstatsum, numsnps, double) ;

  initindiv(indivmarkers, numindivs) ;
  ZALLOC(ti, numindivs, double) ;
  ZALLOC(txi, numindivs, double) ;
  ZALLOC(li, numindivs, double) ;
  ZALLOC(lxi, numindivs, double) ;

  ZALLOC(ti2, numindivs, double) ;
  ZALLOC(txi2, numindivs, double) ;
  ZALLOC(li2, numindivs, double) ;
  ZALLOC(lxi2, numindivs, double) ;

  ncases = ncontrols = 0 ;
   numindok = numindxok = 0 ;
   for (indiv = 0 ; indiv <numindivs; indiv++)  {
    
    indx = indivmarkers[indiv] ;
    if (allcases) indx->affstatus = YES ;
    if ((indx->affstatus == NO) && (usecontrols == NO))
      indx-> ignore = YES ;
    if (indx -> ignore) continue ;
    ++numindok ;
    if (!indx -> affstatus)  ++ ncontrols ;
    if (indx -> affstatus)  ++ ncases ;
   }

   printf("###COUNTS\n");
   printf("Num of fake Markers: %d  Num of real Markers: %d Spacing between fake markers: %9.3f\n",numfake, numnofake, fakespacing) ;

   printf("Num of Markers: %d   Num of Samples:  %d\n",numsnps, numindivs) ;
   printf("Num of Cases:  %d Num of Controls:  %d   Num of Ignored Samples: %d\n", ncases, ncontrols, numindivs-numindok) ;


   fflush(stdout) ;
   if (ncontrols == 0) fprintf(stderr, "*** warning  no controls\n") ;
   if (ncases == 0) fatalx("no cases\n") ;

   zmax = zmin = 0; 

   if (allele_scale_fac>=0)  {
    printf("allele counts scaled by %9.3f\n", allele_scale_fac) ;

    scalecounts(snpmarkers, numsnps, allele_scale_fac) ;

   }

   if (nopopsmode) { 
      printf("nopopsmode\n") ;
   }
   
   ZALLOC(stats, NSTATS, double *) ; 
 
   for (k=0; k< NSTATS; k++) { 
     ZALLOC(stats[k], numbigiter, double) ; 
   } 
   ZALLOC(txsa, numbigiter, double) ;

   if (cleaninit) {
    for (indiv=0; indiv<numindivs; indiv++) {  
// actually set by emsimple but this deals with emiters = 0
       indx = indivmarkers[indiv] ;
       y = a1/(a1+b1) ;
       indx -> theta_mode = indx -> Xtheta_mode = y ;
    }

    vclear(tauscal, TAUINIT, 2) ;
    for (i=0; i<numsnps; i++)  {  
     cupt = snpmarkers[i] ;
     if (cupt -> isfake) continue ;
     if (cupt -> ignore) continue ;  


     if (nopopsmode) {  
      ivclear(cupt -> popA_nn, 1, 2) ;
      ivclear(cupt -> popB_nn, 1, 2) ;
     }

    if (cupt -> popA_nn[0] >= 0) {  
     xa = (double) cupt -> popA_nn[0]  ;
     xb = (double) cupt -> popA_nn[1]  ;
     p = xa/(xa+xb)  ;
     cupt->PopA_freq = p ;
    }
    else p = cupt->PopA_freq ;

    gx(& (cupt -> adm_PopA_freq), p,  tauscal[0]) ;
    
    if (cupt -> popB_nn[0] >= 0) {  
     xa = (double) cupt -> popB_nn[0]  ;
     xb = (double) cupt -> popB_nn[1]  ;
     p = xa/(xa+xb)  ;
     cupt->PopB_freq = p ;
    }
    else p = cupt->PopB_freq ;

    gx(& (cupt -> adm_PopB_freq), p, tauscal[1]) ;
   }

  if (fastdup)  {  
   if (fastdupnum > 0) setfastdupnum(fastdupnum) ;
   fastdupcheck(snpmarkers, indivmarkers, numsnps, numindivs) ;  
  }

   if (checkit) {
    checkgeno() ;
    dophyscheck(snpmarkers, numsnps) ;
    dohwcheck(snpmarkers, numsnps, YES) ;
    checkdup(snpmarkers, numsnps, numindivs) ;
/**
    pubfmodern = NO ;
*/
   }
   fflush(stdout) ;

  if ((indoutfilename != NULL) || (details)) printindivs() ;
  if ((snpoutfilename != NULL) || (details)) 
     printsnps(snpoutfilename, snpmarkers, numsnps, 
      indivmarkers, NO, YES) ;

   if (numbigiter == 0) { 
    printf("## no iterations:  end of run\n") ;
    return 0 ;
   }

  if (probitmode) {  
    doprobit(indivmarkers, numindivs) ;
  }

  if (qtmode) {  
   y1 = y2 = 0.0 ;
   n = 0 ;
   for (i=0; i<numindivs; i++) { 
    indx = indivmarkers[i] ;
    if (indx->ignore) continue ;
    y = indx -> qval ;
    y1 += y ;
    y2 += y*y ;
    ++n ;
   }
   y = y1/(double) n  ;
   printf("qval mean: %9.3f\n", y) ;
   y = (y2/(double) n) - y*y ;
   printf("qval var: %9.3f\n", y) ;
  }

   emsimple( snpmcmarkers, numsnps,  numindivs, emiter, &zmax, &zmin)  ;


   for (i=0; i<numsnps; i++)  {  

    cupt = snpmarkers[i] ;
    if (cupt -> isfake) continue ;
    if (cupt -> ignore) continue ;  
    if (pflat <= 0.0) continue ;

    p = cupt -> adm_PopA_freq ;  
    q = 1-p ;
    p += pflat;
    q += pflat;
    cupt -> adm_PopA_freq  = p /(p+q) ;

    p = cupt -> adm_PopB_freq ;  
    q = 1-p ;
    p += pflat;
    q += pflat;
    cupt -> adm_PopB_freq  = p /(p+q) ;
   }
  }

  if (checkit) checkindiv(snpmarkers, indivmarkers, numsnps, numindivs) ;

  if (numindxok==0) numindxok = 1 ;

  muval1 = muval/ (double) numindxok ;
  printf("muval1: %9.3f\n", muval1) ;
/**
 overall we bias choice of state by exp(-mu*ave(thetax))
 = exp ( -mu/numindxok  \sum theta_i^X 
*/

/* compute posterior averages */
  if ((pubxindiv>=0) || (npubxa>0)) {  
   gammas = initarray_2Ddouble(numsnps, 3, 0.0);
   totgammas = initarray_2Ddouble(numsnps, 3, 0.0);
  }
  else gammas = totgammas = NULL ;
  


 printf("neil0: %9.3f %9.3f\n",zmax, zmin) ;

  cupt = snpmarkers[numsnps-1] ;
  if (cupt -> chrom < 23) {
    printf("no X data!\n") ;
    noxdata = YES ;
  }

  vzero(tmeans, MMEAN) ;
  ZALLOC(mstat, MSTATS, double) ;
  mprint = 10 ; 
  if (rdismode) ++mprint ;  

  if (pubxname != NULL) unlink(pubxname) ;  

  sumfac = avelogfac = 0.0 ;

   x = 0 ;
  for (i=0; i<numsnps; i++) {  
   cupt = snpmarkers[i] ;
   cupt -> score = 0.0 ;  // not needed but good style 
   if (cupt -> isfake) continue ;
   if (cupt -> ignore) continue ;
   if (numvalidgt(indivmarkers, cupt) == 0) { 
     cupt -> ignore = YES ; 
     cupt -> ngtypes = 0 ; 
     if (x<999) printf("*** warning.  No data for %s\n", cupt -> ID) ; 
   } 
  }

  cleartg(indivmarkers, numindivs) ; // zero totgamms
  setestdis(snpmarkers, numsnps) ;

  for (bigiter = 1-numburn ; bigiter <= numbigiter; ++bigiter)  {

    clearestdis(snpmarkers, numsnps) ;
    fflush(stdout) ; 
/* 
 until bigiter=1 this is "burn in"
 numburn 0 acceptable 
*/

    if (qtmode) {  
     setccr(thp1, thp2, qtrho, qtrisks, numrisks) ;
    }

    if (bigiter==1)  {  

      cleartg(indivmarkers, numindivs) ;
      cleargpsum(snpmarkers, numsnps) ;

      for (i=0; i<numsnps; i++)  {  
        cupt = snpmarkers[i] ;
        cumcpt = snpmcmarkers[i] ;
        vzero(cumcpt -> gammasum,2) ;
        vzero(cumcpt -> gammanum,2) ;
        cumcpt -> rpowersum = 0.0 ;
        if (twomodel > 0) { 
          vzero(cupt -> totmodelscores, numrisks) ;
        }
      }
     if (lmmodel) { 
       clearmodelscores(snpmarkers, numsnps, YES) ;
     }
    }
    
     if ((bigiter == numbigiter) && (npubxa>0)) {
      for (i=0; i<npubxa; i++) {  
       k = pubxa[i] ;
       domc1(xscore, zstat, snpmarkers, snpmcmarkers, 
        numsnps, numindivs, 
        riskarr, risk2arr, criskarr, modelsc, wr, numrisks, 
        &tave, &lave, xtave, xlave, k, gammas) ;  

        ivmaxmin(fchrom, 24, NULL, &fc) ;
        ivmaxmin(lchrom, 24, &lc, NULL) ;
        pubgammas(gammas, k, fc, lc, pubxname) ;
      }
     }

   ylike =     domc1(xscore, zstat, snpmarkers, snpmcmarkers, 
        numsnps, numindivs, 
        riskarr, risk2arr, criskarr, modelsc, wr, numrisks, 
        &tave, &lave, xtave, xlave, pubxindiv, gammas) ;  

     setgpsum(snpmarkers, numsnps) ;

     if (bigiter == (1-numburn)) {
      vmaxmin(zstat, numsnps, &zmax, &zmin) ;
      printf("neil1: %9.3f %9.3f\n",zmax, zmin) ;
     }
     

     if (pubxindiv>=0) {  
      for (j=0; j<numsnps; ++j)  {
        if (bigiter==1) { 
          vzero(totgammas[j], 3) ;
        }
        vvp(totgammas[j], totgammas[j], gammas[j], 3) ;
      }
     }


    if (tlreest)  {
       txsum = thetaxsum(indivmarkers, numindivs) ;
       estglob(mstat, MSTATS, bigiter) ;
       if (rdismode) { 
        q = rdis ;
        estflat(&q, snpmarkers, numsnps) ;
        rdis = q ; 
        printf("flat: %4d %9.3f\n",bigiter, q) ;
       }
       mstat[6] = tauscal[0]  ;         
       mstat[7] = tauscal[1]  ;         
     }

    if (bigiter == (1-numburn)) ybase = ylike ;
    ylike -= ybase ;


    vclip(xscore, xscore,  loclip, hiclip, numsnps) ; 
    vpow10(xscore, xscore, numsnps) ;

    if (bigiter >= 1) {
     vvp(facarray, facarray, xscore, numsnps) ;

     vpow10(modelsc, modelsc, numrisks) ;
     vvp(tmodelsc, tmodelsc, modelsc, numrisks) ;
     if (lmmodel) addmodelscores(snpmarkers, numsnps) ;

     if (twomodel > 0) { 
      for (i=0; i<numsnps; i++) {  
       cupt = snpmarkers[i] ;
       if (cupt -> ignore) continue ;
       vvp(cupt -> totmodelscores, cupt -> totmodelscores, 
        cupt -> modelscores, numrisks) ;
      }
     }

/** 
 accumulate zstats 
*/
     vvp(zstatsum, zstatsum, zstat, numsnps) ;
     mstat[10] = mstat[0] ;
     vvp(tmeans, tmeans, mstat+10, 8) ;
     vvp(tmeans+8, tmeans +8, tauscal, 2) ;
     tmeans[10] += rdis ;

     if (qtmode) {  
      setqco(ccr, thp1, thp2) ;
      qmu = thp1/(thp1+thp2) ;
      yransc = qqltot(indivmarkers, numindivs, ccr, qmu) ;
      for (i=0; i<numrisks; i++)  {  
        setqcox(ccr, thp1, thp2, qtrho[i], 0.0) ;
        ysc = qqltot(indivmarkers, numindivs, ccr, qmu) - yransc ;
        qbasemodelsc[i] += exp(ysc) ;
      }
     }
//   bigiter >= 1
    }

    for (i= numsnps-1; i >= 0; i--)  {  
      cupt = snpmarkers[i] ;
      if (cupt -> ignore) { 
	xscore[i] = 0.0 ;  
	continue ; 
      } 
      if ((cupt -> isfake) && (!(cupt -> isrfake))) continue ;
      xscore[i] = 0.0 ;
      if (cupt -> ngtypes == 0) continue ; 
      cumcpt = snpmcmarkers[i] ;
      p  = cupt -> adm_PopA_freq ;
      pp = cupt -> adm_PopB_freq ;
      mk2from3(dd, cumcpt -> xc, p , pp) ;
      mk2from2(dd2, cumcpt -> xd) ;
      /* put in extra check */ 
      vvp(cumcpt -> xd, dd, dd2, 4) ;
    }

    tmp = asum(xscore, numsnps) / (double) numfake ;
    tmp = MAX(tmp, 1.0e-20) ; 
//  tmp4 = log(100.0) ;  /* Added by AT to keep linux happy */ 
    tmp2 = log10(tmp) ;
    if (bigiter>=1) {  
     sumfac += tmp ;
     avelogfac = log10(sumfac) - log10 ((double) bigiter) ;
    }
    if (rdismode == NO) {
     mstat[8] = tmp2 ;
     mstat[9] = avelogfac ;
    }
    else {  
     mstat[8] = rdis ;
     mstat[9] = tmp2 ;
     mstat[10] = avelogfac ;
    }
    prmstat(bigiter, mstat, mprint) ;

    for (reestiter = 1;  reestiter <= reestnumiter; ++reestiter) {

     mode = NO ;
     if  ((bigiter >= 1)  && (reestiter==1)) mode = YES ;
     sampfreqx(tauscal, snpmarkers, snpmcmarkers, numsnps, mode)  ;
     if (numrfake>0) { 
      sampfreqy(tauscal, snpmarkers, snpmcmarkers, numsnps, mode)  ; 
      /* does NOT reestimate tau */ 
     } 
    }

   if (verbose) {
    printf("bigiter: %5d %15.3f %9.3f %9.3f %9.3f" ,
     bigiter, ylike, tmp2, tauscal[0], tauscal[1]) ;
    printf(" %9.3f %9.3f  %9.3f %9.3f  %9.3f %9.3f", 
     tave, lave, xtave[0], xlave[0], xtave[1], xlave[1]) ;
    printf("\n") ;
   }

/**
   if (bigiter >= 1) 
   pubaux(score2file, bigiter) ;
*/

  
   tstat[0] = ylike ;  tstat[1] = tmp2;  tstat[2] = tauscal[0] ;
   if (bigiter >= 1) {      
    for (k=0; k<NSTATS; k++)  {
     stats[k][bigiter-1] = tstat[k] ;
    }
    txsa[bigiter-1] = thetaxsum(indivmarkers, numindivs) / (double) numindxok ;
   }
   for (indiv = 0; indiv < numindivs; ++indiv)  {  
    indx = indivmarkers[indiv] ;
    if (bigiter >= 1)  {
     y = indx ->  theta_mode ;
     ti[indiv]  += y ; 
     ti2[indiv]  += y*y ; 
     y = indx ->  Xtheta_mode ;
     txi[indiv]  += y ; 
     txi2[indiv]  += y*y ; 
     y = indx ->  lambda_mode ;
     li[indiv]  += y ; 
     li2[indiv]  += y*y ; 
     y = indx ->  Xlambda_mode ;
     lxi[indiv]  += y ; 
     lxi2[indiv]  += y*y ; 
    }
   }
   fflush(stdout) ;
   fflush(scorefile) ;
   /* end of bigiter loop */
  }
  y = asum(txsa, numbigiter) / (double) numbigiter ;
  printf("average thetax: %9.3f\n", y) ;
  if (printmut) 
   printmutab(txsa, numbigiter) ;

  vst(tmodelsc, tmodelsc, 1.0/ (double) numbigiter, numrisks) ;
  vlog10(tmodelsc, tmodelsc, numrisks) ;

  if (qtmode) {
   vst(qbasemodelsc, qbasemodelsc, 1.0/ (double) numbigiter, numrisks) ;
   vlog10(qbasemodelsc, qbasemodelsc, numrisks) ;
  }
  
  vst(tmeans, tmeans, 1.0/ (double) numbigiter, MMEAN) ;
  printtmeans(tmeans) ;
  rdis = tmeans[10] ;

  freearms() ;

printf("\n\n") ;
  printf("%20s\n","###GENOME_WIDE SCORE FOR ALL THE MODELS");
  if (qtmode) { 
    printf("%6s %9s %9s %9s %9s  %9s\n",  
     "", "qtrho", "qtrisk", "base sc", "score", "diff") ;

   for (k=0; k< numrisks; k++)  {  
    y1 = qbasemodelsc[k] ;
    y2 = tmodelsc[k] ;

    printf("%6s %9.3f %9.3f %9.3f %9.3f  %9.3f",  
    "model:", qtrho[k], qtrisks[k], y1, y2, y2-y1) ;
    if (wr != NULL) printf("    %9.3f", wr[k]) ;
     printf("\n") ;
   }
  }

  else {
  printf("##risk1 and risk2 are the increased risk due to having one or two population A ancestry alleles, and  crisk: risk for controls\n"); 
  printf("%6s %9s %9s %9s   %9s\n",  
   "", "risk1", "risk2", "crisk", "score") ;


  for (k=0; k< numrisks; k++)  {  
   printf("%6s %9.3f %9.3f %9.3f   %9.3f",  
    "model:", riskarr[k], risk2arr[k], criskarr[k], tmodelsc[k]) ;
   if (wr != NULL) printf("    %9.3f", wr[k]) ;
    printf("\n") ;
  }
 }
  printf("\n\n") ;


  

  if (details) {
   pubindivs(ti, ti2, txi, txi2, 
     li, li2, lxi, lxi2, numbigiter) ;
   mcmpt = snpmcmarkers ;
   if (pubfmodern) mcmpt = modmcmarkers ;
   pubfreqs(mcmpt, numbigiter) ;
   if (rdismode) pubdis(snpmarkers, snpmcmarkers, numsnps) ;
  }

  vst(facarray, facarray, 1.0/ (double) numbigiter, numsnps) ; 
  vst(zstatsum, zstatsum, 1.0/ (double) numbigiter, numsnps) ; 
  vlog10(xscore, facarray, numsnps) ; 
 
  if (lmmodel) printmodelscores(xscore, snpmarkers, numsnps, numbigiter) ;

  printf("\n###LAG AND CORRELATIONS\n") ;

  if (numbigiter>2) {
   dostat("llike", stats[0], numbigiter) ;
   dostat("log10fac", stats[1], numbigiter) ;
   vclip(stats[1], stats[1],  loclip, hiclip, numbigiter) ; 
   vpow10(stats[1], stats[1], numbigiter) ;
   dostat("factor", stats[1], numbigiter) ;
   vlog10(stats[2], stats[2], numbigiter) ;
   dostat("log tauscal", stats[2], numbigiter) ;
  }

  if (twomodel > 0) {  
   y = 1.0/ (double) numbigiter ;
   for (i=0; i< numsnps; i++)  {  
    cupt = snpmarkers[i] ;
    vst(cupt -> totmodelscores, cupt -> totmodelscores, 
      y, numrisks) ;
   }
  }

  fflush(stdout) ; 
  pubscoresx(xscore, zstatsum) ;
  

  if (scorename != NULL) 
   fclose(scorefile);

  if (score2name != NULL) 
   fclose(score2file);

  y = (double) numbigiter ;

   vst(ti, ti, 1.0/y, numindivs) ;
   vst(txi, txi, 1.0/y, numindivs) ;
   vst(li, li, 1.0/y, numindivs) ;
   vst(lxi, lxi, 1.0/y, numindivs) ;

   for (indiv=0; indiv < numindivs; ++indiv) {  
    indx = indivmarkers [indiv] ;
    indx -> theta_mode = ti[indiv];
    indx -> Xtheta_mode = txi[indiv];
    indx -> lambda_mode = li[indiv];
    indx -> Xlambda_mode = lxi[indiv];
   }

  for (i=0; i<numsnps; i++) {  
    cupt = snpmarkers[i] ;
    if (cupt -> isfake) continue ;
    if (cupt -> ignore) continue ;
    cumcpt = snpmcmarkers[i] ;
    cupt -> adm_PopA_freq = cumcpt -> ap1 / y ;
    cupt -> adm_PopB_freq = cumcpt -> cp1 / y ;
  }


  if (checkit)  {  
   leave1(snpmarkers, numsnps, numindivs, lrisk, trashdir) ;
  }

  if (pubxindiv>=0) {  
   ivmaxmin(fchrom, 24, NULL, &fc) ;
   ivmaxmin(lchrom, 24, &lc, NULL) ;
   pubgammas(totgammas, pubxindiv, fc, lc, pubxname) ;
  }
  else printgall(pubxname) ;

  if (corrgammas) { 
   dumpgammas = YES ;
  }   

  if (dumpgammas) {
   dumpgams(snpmarkers, numsnps, numindivs, trashdir); 
   printf ("gammas dumped\n") ;
   if (corrgammas) 
    corrgamma (snpmarkers,  numsnps,  numindivs, 
     0.1, 2.0, epithresh) ;
  }

  printf("##end of run\n") ;
  return 0 ;
}

void checkgeno() 
{
  int i, k, x ;
  int n, nbad, ngood ;
  SNP *cupt ;

  printf("\n### CHECKGENO RESULTS FOLLOW:\n");
  nbad = ngood = 0 ;
  for (i=0; i<numsnps; i++) {  
   cupt = snpmarkers[i] ;
   if (cupt -> isfake) continue ;
   if (cupt -> ignore) continue ;
   n = cupt -> ngtypes ;
   if (n<=0) continue ;
   for (k=0; k<n; k++) {
    x = getgtypes(cupt, k) ; 
    if (x>2) {  
     printf("bad genotype: %s %4d %4d\n", cupt->ID, k,x) ;
     putgtypes(cupt, k, -1) ;
     ++nbad ;
    }
    else  {
     ++ngood ;
    }
   }
  }
  printf("Num good genotypes: %d  Num bad genotypes:  %d\n\n",ngood, nbad) ;
 
/**
  if (nbad>0) fatalx("checkgeno fails\n") ;
*/
}

void
pubgammas(double **gammas, int pubx, int fc, int lc, char *fname)  
{
  SNP *cupt ;
  double gg[3], y ;
  Indiv *indx ;
  int  i ;
  FILE *fff ;

  if (pubx>=numindivs) {   
    printf("warning.  gammas wanted for individual out of range!\n") ; 
    return ;  
  } 

  if (gammas==NULL) return ;
  if (lc<0) return ;

  if (fname != NULL) openit(fname, &fff, "a") ;
  else fff = stdout ;

/** 
 open for appending 
*/


  indx = indivmarkers[pubx]  ;
  fprintf(fff, "#Gammas for individual %d  ID: %s\n",  
   pubx, indx -> ID) ;

  fprintf(fff,"%4s %15s %4s %8s %8s %8s %8s\n","SNP_Index","SNP_ID","Chr_Num","Gen_Pos","G[0]","G[1]","G[2]");
  for (i=fc ; i<= lc; i++) {   
   cupt = snpmarkers[i] ;
   if (cupt -> ignore) continue ;
   gg[0] = gammas[i][0] ;
   gg[1] = gammas[i][1] ;
   gg[2] = gammas[i][2] ;
   y = asum(gg, 3) + 1.0E-20 ;
   vst(gg, gg, 1.0/y, 3) ;
   fprintf(fff, "%4d %20s %3d %9.3f   %9.3f %9.3f %9.3f\n",
    i, cupt->ID, cupt -> chrom, 
    100*cupt -> genpos, gg[0], gg[1], gg[2]) ;
  }
  if (fname != NULL) fclose(fff) ;
}
void
pubindivs(double *ti, double *ti2, double *txi, double *txi2, 
     double *li, double *li2, double *lxi, double *lxi2, int n)  
{

  /**AT: Making changes, since we don't want to print out theta_true, lambda_true,lambdaX_true and thetax_true if not in 
simulation mode, meaningless number*/

  int indiv ;
  Indiv *indx ;
  double yn, mean, sdev ;
  FILE *tfile = NULL, *lfile = NULL ;
  char status[12] ;

  if (thetafilename != NULL) openit(thetafilename, &tfile, "w") ;
  if (lambdafilename != NULL) openit(lambdafilename, &lfile, "w") ;

  yn = (double) n ;
  printf("###THETA or M, LAMBDA VALUES FOR ALL INDIVIDUALS\n") ;
  printf("##Indiv_Index: individual's internal index num, tmean and txmean: average theta and thetax\n");
  printf("##tsdev and txsdev: standard deviation for theta and thetax\n");
  printf("##lmean and lxmean: average lambda and lambdax\n");
  printf("##lsdev and lxsdev: standard deviation for lambda and lambdax\n");


  if ((tfile != NULL) && (dotoysim == YES)) {
    fprintf(tfile,"%5s %18s ", "#Num", "Indiv_Id") ;
    fprintf(tfile,"%9s %9s %9s ", "th_true", "th_mean", "th_sdev") ;
    fprintf(tfile,"%9s %9s %9s %9s\n", "thX_true", "thX_ mean", "thX_sdev","Status") ;
  }
  
    
  if ((tfile != NULL) && (dotoysim == NO)) {
    fprintf(tfile,"%5s %18s ", "##Indiv_Index", "Indiv_Id") ;
    fprintf(tfile,"%9s %9s ", "th_mean", "th_sdev") ;
    fprintf(tfile,"%9s %9s %9s\n", "thX_mean", "thX_sdev","Status") ;
  }

  if ((lfile != NULL) && (dotoysim == YES)) {
    fprintf(lfile,"%5s %18s", "##Indiv_Index", "Indiv_Id") ;
    fprintf(lfile,"%9s %9s %9s ", "lam_true", "lam_mean", "lam_sdev") ;
    fprintf(lfile,"%9s %9s %9s\n", "lamX_true", "lamX_mean", "lamX_sdev") ;
  }
  if ((lfile != NULL) && (dotoysim == NO)) {
    fprintf(lfile,"%5s %18s", "##Indiv_Index","Indiv_Id") ;
    fprintf(lfile,"%9s %9s ","lam_mean", "lam_sdev") ;
    fprintf(lfile,"%9s %9s\n","lamX_mean", "lamX_sdev") ;
  }

  printf("%4s %15s %4s","Num","Indiv_ID","Gender") ;
  printf("%9s %9s  ", "tmean", "tsdev") ;
  printf("%9s %9s    ", "txmean", "txsdev") ;
  printf("%9s %9s  ", "lmean", "lsdev") ;
  printf("%9s %9s\n\n", "lxmean", "lxsdev") ;


  for (indiv = 0; indiv < numindivs; ++indiv)  {  
   indx = indivmarkers [indiv] ;
   if (indx -> ignore) continue ;
   printf("%4d %25s %5c ", indiv,  indx -> ID, indx -> gender) ;
   calcmsdev(ti[indiv], ti2[indiv], yn, &mean, &sdev) ;
   printf("%9.3f %9.3f  ", mean, sdev) ;
   if ((tfile != NULL) && (dotoysim == YES)) {
    fprintf(tfile,"%5d %25s ", indiv, indx->ID) ;
    fprintf(tfile,"%9.3f %9.3f %9.3f ", indx -> thetatrue, mean, sdev) ;
   }
   if ((tfile != NULL) && (dotoysim == NO)) {
    fprintf(tfile,"%5d %25s ", indiv, indx->ID) ;
    fprintf(tfile,"%9.3f %9.3f ",mean, sdev) ;
   }
   calcmsdev(txi[indiv], txi2[indiv], yn, &mean, &sdev) ;
   printf("%9.3f %9.3f    ", mean, sdev) ;
   if ((tfile != NULL) && (dotoysim == YES)) {
    fprintf(tfile,"%9.3f %9.3f %9.3f", indx -> Xthetatrue, mean, sdev) ;
    indstatus(status, indx) ;
    fprintf(tfile," %7s\n", status) ;
   }
   if ((tfile != NULL) && (dotoysim == NO)) {
    fprintf(tfile,"%9.3f %9.3f",mean, sdev) ;
    indstatus(status, indx) ;
    fprintf(tfile," %7s\n", status) ;
   }
   
   calcmsdev(li[indiv], li2[indiv], yn, &mean, &sdev) ;
   printf("%9.3f %9.3f  ", mean, sdev) ;
   if ((lfile != NULL) && (dotoysim == YES)){
    fprintf(lfile,"%5d %25s", indiv,indx->ID) ;
    fprintf(lfile,"%9.3f %9.3f %9.3f ", indx -> lambdatrue, mean, sdev) ;
   }
   if ((lfile != NULL) && (dotoysim == NO)){
    fprintf(lfile,"%5d %25s", indiv, indx->ID) ;
    fprintf(lfile,"%9.3f %9.3f ",mean, sdev) ;
   }
   calcmsdev(lxi[indiv], lxi2[indiv], yn, &mean, &sdev) ;
   printf("%9.3f %9.3f\n", mean, sdev) ;
   if ((lfile != NULL) && (dotoysim == YES))
    fprintf(lfile,"%9.3f %9.3f %9.3f\n", indx -> Xlambdatrue, mean, sdev) ;
   if ((lfile != NULL) && (dotoysim == NO))
    fprintf(lfile,"%9.3f %9.3f\n", mean, sdev) ;
  }
  printf("\n\n") ;

  if (tfile != NULL) fclose (tfile) ;
  if (lfile != NULL) fclose (lfile) ;

}  

void pubfreqs(SNPMC **snpmcmarkers, int n) 
{
  /**AT: Making changes, since we don't want to print out theta_true, lambda_true,lambdaX_true and thetax_true if not in 
simulation mode, meaningless number*/
// side effect set adm_PopA_freq adm_PopB_freq to mean estimate from mcmc

  SNP *cupt, *tcupt ;
  SNPMC *cumcpt ;
  int i, chrom, j ; 
  double yn, p1, p2, mean, sdev ;
  double a, b , rmean, rvar, rsig, t ;
  double amean, ramean, adev ;
  double cmean, rcmean, cdev ;
  double atrue , ctrue ;
  FILE *pfile = NULL ;
  int xcnt[4] ;

  yn = (double) n ;
  printf("\n");
  printf("###ALLELE FREQUENCY ESTIMATES WITH STANDARD ERROR\n") ;
  printf("##SNP_Index: marker internal index num\n");
  printf("##amean and bmean are the average reference allele frequency for population A and B\n");
  printf("##asdev and bsdev are the corresponding standard deviation\n\n");
  printf("%4s %5s %15s %7s %7s %9s %9s\n","SNP_Index","Chr_Num","SNP_ID","amean","asdev","bmean","bsdev");

  
  if (freqfilename != NULL) openit(freqfilename, &pfile, "w") ;
  if (pfile != NULL) fprintf(pfile, "##freqs\n") ;

  if ((pfile != NULL) && (dotoysim == YES)) { 
   fprintf(pfile, "%4s %15s %5s ","#SNP_Index", "SNP_ID", "Chr_Num") ;
   fprintf(pfile, "%9s %9s %9s %9s  ",
    "atrue", "anaive", "amean", "asdev") ;
   fprintf(pfile, "%9s %9s %9s %9s\n",
    "btrue", "bnaive", "bmean", "bsdev") ;
  }

  if ((pfile != NULL) && (dotoysim == NO)) { 
    fprintf(pfile,"###ALLELE FREQUENCY ESTIMATES WITH STANDARD ERROR\n") ;
    fprintf(pfile,"##SNP_Index: marker internal index num\n");
    fprintf(pfile,"##amean and bmean are the average reference allele frequency for population A and B\n");
    fprintf(pfile,"##asdev and bsdev are the corresponding standard deviation\n\n");
    fprintf(pfile,"%4s %15s %5s %9s %9s %9s %9s %9s %9s %9s %9s\n","SNP_Index","SNP_ID","Chr_Num","atrue","anaive","amean","asdev","btrue","bnaive","bmean","bsdev");


  }


  for (i=0; i<numsnps; i++) { 
   cupt = snpmarkers[i] ;
   if (cupt -> isfake) continue ;
   if (cupt -> ignore) continue ;

   if (pubfmodern) {
    atrue = cupt -> PopAftrue ;
    ctrue = cupt -> PopBftrue ;
   }
   else {
    atrue = cupt -> adm_PopAftrue ;
    ctrue = cupt -> adm_PopBftrue ;
   }

   a = cupt -> popA_nn[0]  ;
   b = cupt -> popA_nn[1]  ;
   t = a+b ;
   ramean = a/t ;  

   a = cupt -> popB_nn[0]  ;
   b = cupt -> popB_nn[1]  ;
   t = a+b ;
   rcmean = a/t ;  

   chrom = cupt -> chrom ;
   if (pfile != NULL) fprintf(pfile, "%9d %15s %5d ",i, cupt -> ID, chrom) ;

   cumcpt = snpmcmarkers[i] ;
   p1 = cumcpt -> ap1 ;
   p2 = cumcpt -> ap2 ;
   calcmsdev(p1, p2, yn, &mean, &sdev) ;
   printf("%5d %3d ",i, chrom) ;
   printf("%20s ", cupt -> ID) ;
   printf("%9.3f %9.3f    ", mean, sdev) ;
   cupt -> adm_PopA_freq = mean ;
   if (pfile != NULL) {  
    fprintf(pfile, "%9.3f %9.3f %9.3f %9.3f  ", 
     atrue, ramean, mean, sdev) ; 
   }
   p1 = cumcpt -> cp1 ;
   p2 = cumcpt -> cp2 ;
   calcmsdev(p1, p2, yn, &mean, &sdev) ;
   printf("%9.3f %9.3f", mean, sdev) ;
   cupt -> adm_PopB_freq = mean ;
   printf("\n") ;
   fflush(stdout) ;
   if (pfile != NULL) {  
    fprintf(pfile, "%9.3f %9.3f %9.3f %9.3f\n", 
     ctrue, rcmean, mean, sdev) ; 
   }
  }
  if (pfile == NULL) return ;
  fprintf(pfile,"end_of_freqs\n") ;
  fprintf(pfile,"%5s %3s %15s %12s %12s %12s %12s\n","SNP_Index","Chr","SNP_ID","Ctl_ref","Ctl_vart","Case_ref","Case_vart");
  for (i=0; i<numsnps; i++) { 
   cupt = snpmarkers[i] ;
   if (cupt -> isfake) continue ;
   if (cupt -> ignore) continue ;
   if (cupt -> gtypes == NULL) continue ;
   countgtypes(cupt, xcnt) ;
   chrom = cupt -> chrom ;
   fprintf(pfile, "%5d %3d ",i, chrom) ;
   fprintf(pfile, "%20s ", cupt -> ID) ;
   for (j=0; j<=3 ; ++j) {
    fprintf(pfile, "%12d", xcnt[j]) ;
   }
   fprintf(pfile, "\n") ;
  }
  fprintf(pfile,"end_of_gtype count\n") ;

  if (pfile != NULL) fclose (pfile) ;
}
void
countgtypes(SNP *cupt, int *xcnt)  
{
   int k, numstates, ignore, refcnt, varcnt, a ;
   int maxcnt ;
   Indiv *indx ;

   ivzero(xcnt, 4) ;
   if (cupt -> isfake) return ;
   if (cupt -> ignore) return ;
   if (cupt -> gtypes == NULL) return ;
   if (cupt -> ngtypes == 0) return ;
   for (k=0; k< numindivs; ++k) {  
    indx = indivmarkers[k] ;
    gettln(cupt, indx, NULL, NULL, &numstates, &ignore) ;
    if (ignore == YES) continue ;
    a = 0 ;
    if (indx -> affstatus) a = 2 ;  
    refcnt = getgtypes(cupt, k) ;
    if (refcnt<0) continue ;
    maxcnt = 2  ;
    if (numstates==2) {  
     maxcnt = 1 ;
     refcnt /= 2 ;
    }
    varcnt = maxcnt-refcnt ;
    xcnt[a] += refcnt ;
    xcnt[a+1] += varcnt ;
   }
}

void
calcmsdev(double ysum, double ysum2, double num, double *mean, double *sdev)  
/* mean and sdev from  moments about 0 */
{
  double p1, p2, m, sd ;  
  m = p1 =  ysum/num ;
  p2 =  ysum2/num ;
  p2 -= p1*p1 ;
  p2 += 1.0e-10 ;
  sd = sqrt(p2) ;

  *mean = m ;  
  *sdev = sd ;

}

void pubaux(FILE *fff, int iter) 
{
  int indiv ;
  Indiv *indx ;
  double t1, t2 ;
  if (fff == NULL) return ;
  for (indiv = 0; indiv <numindivs; ++indiv)  {
   indx = indivmarkers[indiv] ;
   if (indx -> ignore) continue ;
   t1 = indx -> theta_mode ;
   t2 = indx -> Xtheta_mode ;
   fprintf(fff,"%9.3f %9.3f %5d %5d\n",t1,t2,indiv,iter) ;
  }
  fflush(fff) ;
}

void pubscoresx (double *basescore,  double *zscore) 
{
   int i, nmark=0 ;
   int chrom ;
   SNP *cupt ;
   SNPMC *cumcpt ;
   double be[MAXCH+1], bz[MAXCH+1], bw[MAXCH+1] ;
   double xf[MAXCH+1], xfn[MAXCH+1] ;
   double *w1 = NULL, *w2 ;
   double xm1, xm2, txm1, txm2 ;
   double xa[2], xb[2] ;
   int n1, n2 ;
   FILE *lfff ;

/* to score by chromosome */
   double totfac, y, yt ;
   int numsc = 0, nfake ;
   static int nbad = 0  ;
   double bestnick, bestneil, worstneil ;
   double bestx1, bestx2, best12, best21 ;

   if (twomodel > 0) { 
    n1 = twomodel ;
    n2 = numrisks - n1 ;
    ZALLOC(w1, n1, double) ;
    ZALLOC(w2, n2, double) ;
    y = asum(wrisk, n1) ;
    vst(w1, wrisk, 1.0/y, n1) ;
    y = asum(wrisk+n1, n2) ;
    vst(w2, wrisk+n1, 1.0/y, n2) ;
   }
   vclear(be, -100.0, MAXCH+1) ;
   vclear(bz, -100.0, MAXCH+1) ;
   vclear(bw,  100.0, MAXCH+1) ;
   if (localoutfilename != NULL)  {  
     openit(localoutfilename, &lfff, "w") ;
   }
   else lfff = stdout ;
   if (details) {
     fprintf(lfff,"\n###SCORES FOR EACH MARKER (fakes used for global score)\n") ;
     fprintf(lfff,"##LGS: locus genome statistic score, CCS:case control statistic\n");
     fprintf(lfff,"%12s %6s %20s", "###SNP_Index", "Chr","SNP_ID") ;
     fprintf(lfff,"%12s %10s", "Phys_Pos", "Gen_Pos" ) ;
     fprintf(lfff," %9s  %9s", "LGS", "CCS") ;
   }

   if (details && (!(twomodel>0))) {  
    fprintf(lfff,"      ") ;
    fprintf(lfff," %9s %10s  %9s\n", "G(case)", "G(control)", "rpower") ;
   }

   if (details && ((twomodel>0))) {   
     fprintf(lfff,"      %9s  %9s  %9s", "Model1", "Model2", "Diff") ; 
     fprintf(lfff,"\n") ;    
   } 

   
//  fprintf(lfff,"\n") ;    // NJP: extra blank line deleted here
   
   totfac = txm1 = txm2 = 0.0 ;

   vzero(xf, MAXCH+1) ;
   vzero(xfn, MAXCH+1) ;
   
   nfake = 0 ;
  bestx1 = bestx2 = best12 = best21 = -500.0 ;
   for (i=0; i<numsnps; i++) {

    cupt = snpmarkers[i] ;  
    cumcpt = snpmcmarkers[i] ;  

    if (cupt -> ignore) continue ; 

    if (twomodel > 0) {  
     xm1 = vdot(w1, cupt -> totmodelscores, n1) ;
     xm2 = vdot(w2, cupt -> totmodelscores +n1, n2) ;
     if ((cupt -> isfake) && (!(cupt -> isrfake))) { 
      ++nfake ;
      txm1 += xm1 ;
      txm2 += xm2 ;
     }
     xm1 = log10(xm1 + 1.0e-20) ;
     xm2 = log10(xm2 + 1.0e-20) ;
     bestx1 = MAX(bestx1, xm1) ;
     bestx2 = MAX(bestx2, xm2) ;
     best12 = MAX(best12, xm1 - xm2) ;
     best21 = MAX(best21, xm2 - xm1) ;
    }

    chrom = cupt -> chrom ;

    if (chrom<0) fatalx("bad chrom\n") ;
    if (chrom>23) fatalx("bad chrom\n") ;

     ++nmark ;

    if ((basescore[i] < -50)  && (nbad < 10)) {  
      ++nbad ;
      printf("zzbad %5d %3d %20s", i, chrom, cupt->ID) ;
    }

    if (details) {
     fprintf(lfff,"%12d %6d %20s", i, chrom, cupt->ID) ;
     fprintf(lfff," %9.0f %12.3f", 
      cupt->physpos, cupt->genpos) ;
     fprints1(lfff,basescore[i]) ;
     fprints1(lfff,zscore[i]) ;
     fprintf(lfff,"   ") ;
     if (twomodel > 0) {  
      fprints1(lfff,xm1) ;
      fprints1(lfff,xm2) ;
      fprints1(lfff,xm1-xm2) ;
     }
     else {  
      vsp(xb, cumcpt -> gammanum, 1.0e-8, 2) ;
      vvd(xa, cumcpt -> gammasum, xb, 2) ;
      fprints1(lfff,xa[0]) ;        
      fprints1(lfff,xa[1]) ;        
      y = cumcpt -> rpowersum / (double) numbigiter ;
      fprints1(lfff,y) ;
     }
     fprintf(lfff,"\n") ;
     fflush(lfff) ;
    }
/** 
 now print genpos estimates 
*/
    
    if (!(cupt -> isfake )) continue ; 
    if (cupt -> isrfake) continue ;
    if (cupt -> ignore) continue ;
    y = MIN(basescore[i], hiclip) ;
    y = MAX(basescore[i], loclip) ;
    yt = pow10(y) ;
    totfac += yt ;  
    xf[chrom] += yt ;
    xfn[chrom] += 1 ;
    ++numsc ;
    be[chrom] = MAX(be[chrom], basescore[i]) ;
    bz[chrom] = MAX(bz[chrom], zscore[i]) ;
    bw[chrom] = MIN(bw[chrom], zscore[i]) ;
   }

  totfac /= (double) numsc ;
  totfac = MAX(totfac, 1.0e-20) ;
   
   bestnick = bestneil = -100.0 ;
   worstneil = 100.0 ;
   printf("\n###SCORES FOR EACH CHROMOSOME\n");
   printf("##LGS_MAX: Maximum locus genome statistic score\n");
   printf("##CCS_MAX and CCS_MIN are the maximum and minimum case control statistic scores\n");
   printf("##LGS_LOCAL: log likelihood of the locus genome statistic score obtained by averaging over all the markers on that chromosome\n");
   printf("%10s %8s  %8s %8s  %8s", 
    "Chr_Num", "LGS_MAX", "CCS_MAX", "CCS_MIN", "LGS_LOCAL") ;
   printf("\n\n") ;
   for (chrom=1; chrom<= 23; ++chrom) {  
     y = xf[chrom] ; yt = xfn[chrom] ;
     if (be[chrom] < -50.0) continue ;
     if (yt<0.1) continue ;
     y = y/yt ;
     y  = MAX(y, 1.0e-20) ;
     printf("%5d %6.2f  %6.2f %6.2f  %6.2f", 
        chrom, be[chrom], bz[chrom], bw[chrom], log10(y)) ;
     printf("\n") ;
     fflush(stdout) ;
     bestnick = MAX(bestnick, be[chrom]) ;
     bestneil = MAX(bestneil, bz[chrom]) ;
     worstneil = MIN(worstneil, bw[chrom]) ;
   }

   /*   printf("nickgen:  %9.3f %9.3f\n\n", log10(totfac), bestnick) ;*/
   printf("\n###BESTSCORES: Maximum genome-wide score for the locus-genome statistic (LGS_MAX), and the maximum and minimum genome-wide scores for the case-control statistic (CCS_MAX and CCS_MIN)\n");
   printf("bestscores:  %9.3f %9.3f %9.3f\n", 
    bestnick, bestneil, worstneil) ;


   if (twomodel > 0) { 
     txm1 /= (double) nfake ;
     txm2 /= (double) nfake ;
     txm1 = log10(txm1) ;
     txm2 = log10(txm2) ;
     printf("twomodel (best local): ") ;  
     prints1(bestx1) ;
     prints1(bestx2) ;
     prints1(best12) ;
     prints1(best21) ;
     printf("\n") ;
     printf("twomodel (global): ") ;  
     prints1(txm1) ;
     prints1(txm2) ;
     prints1(txm1-txm2) ;
     printf("\n") ;
   }

   printf("\n###GENOME LOG FACTOR: log-likelihood of the locus genome statistic averaged over all the markers in the genome\n");
   printf("genome log-factor: %9.3f\n", log10(totfac) ) ;

   if (w1 != NULL) { 
    free(w1) ;
    free(w2) ;
   }
}

void prints1(double x) 
{
  if (x < -500) printf(" %9s ", "") ; 
  else printf(" %9.3f ",x) ;
}

void fprints1(FILE *fff, double x) 
{
  if (x < -500) fprintf(fff," %9s ", "") ; 
  else fprintf(fff," %9.3f ",x) ;
}


void readcommands(int argc, char **argv) 

{
  int i ;
  phandle *ph ;
  char str[5000]  ;
  char *tempname ;
  int seed = 0  ;
  int fancyxtheta = YES ;
  int pubx[2] ;
  int n, junk ;
  double pp[3] ;
  double qq[2],rr[2],uu[2];

/* set defaults */

  vclear(riskarr, -1.0, MAXNUMR) ;
  vclear(risk2arr, -1.0, MAXNUMR) ;
  riskarr[0] = 2.0 ;  numrisks = 1 ;
  vclear(criskarr, 1.0, MAXNUMR) ;

  maxmark = xchrom = -1 ;
  pubx[0] = 0; pubx[1] = 22 ;
  vclear(tauscal, TAUINIT, 2) ;

  while ((i = getopt (argc, argv, "p:vV")) != -1) {

    switch (i)
      {

      case 'p':
	parname = strdup(optarg) ;
	break;

      case 'v':
	printf("version: %s\n", WVERSION) ; 
	break; 

      case 'V':
	verbose = YES ;
	break; 

      case '?':
	printf ("Usage: bad params.... \n") ;
	fatalx("bad params\n") ;
      }
  }

   if (parname==NULL) { 
    fprintf(stderr, "no parameters\n") ;
    return ;
   }
         
   pcheck(parname,'p') ;
   printf("parameter file: %s\n", parname) ;
   ph = openpars(parname) ;
   dostrsub(ph) ;
   getint(ph, "numburn:", &numburn) ;
   getint(ph, "numiters:", &numbigiter) ;
   getint(ph, "seed:", &seed) ;
   getint(ph, "maxmark:", &maxmark) ;
   getint(ph, "tlreest:", &tlreest) ;
   getint(ph, "dotoysim:", &dotoysim) ;
   getint(ph, "cleaninit:", &cleaninit) ;
   getint(ph, "checkit:", &checkit) ;
   getint(ph, "leave1mode:", &leave1mode) ;
   getint(ph, "dupmode:", &dupmode) ;
   getint(ph, "fastdup:", &fastdup) ;
   getint(ph, "fastdupnum:", &fastdupnum) ;
   getint(ph, "rdismode:", &rdismode) ;
   getint(ph, "usephyspos:", &usephyspos) ;

   getint(ph, "dumpgammas:", &dumpgammas) ;
   getint(ph, "corrgammas:", &corrgammas) ;
   getstring(ph, "gammafiles:", &gammafiles) ;

   getint(ph, "fancyxtheta:", &fancyxtheta) ;
   getint(ph, "keeptrashfiles:", &keeptrashfiles) ;
   getint(ph, "noxdata:", &noxdata) ;
   getint(ph, "printmut:", &printmut) ;
   getint(ph, "theta2mode:", &theta2mode) ;
   getint(ph, "lambda2mode:", &lambda2mode) ;
   getint(ph, "nopopsmode:", &nopopsmode) ;
   getint(ph, "packmode:", &packmode) ; // controls internals 
   getint(ph, "pubfmodern:", &pubfmodern) ;
   getint(ph, "allmale:", &allmale) ;
   getint(ph, "allcases:", &allcases) ;
   getint(ph, "usecontrols:", &usecontrols) ;
   getint(ph, "details:", &details) ;
   getint(ph, "markernum:", &markernum) ;
   getint(ph, "markerpub:", &markerpub) ;
   getint(ph, "simnumindivs:", &simnumindivs) ;
   getint(ph, "emiter:", &emiter) ;
   getint(ph, "alldata:", &alldata) ;
   getint(ph, "reestiter:", &reestnumiter) ;
   getint(ph, "numlagprint:", &numlagprint) ;
   getint(ph, "twomodels:", &twomodel) ;
   getint(ph, "lmmodel:", &lmmodel) ;
   getint(ph, "lmchrom:", &lmchrom) ;
   getint(ph, "lmnumx:", &lmnumx) ;
   getdbl(ph, "lmmax:", &lmmax) ;
   getdbl(ph, "lmthresh:", &lmthresh) ;
   getint(ph, "lmdetails:", &lmdetails) ;
   getint(ph, "lmlobase:", &lmlobase) ;
   getint(ph, "lmhibase:", &lmhibase) ;
   getint(ph, "oldlmmode:", &oldlmmode) ;
   getint(ph, "randombadpairs:", &randombadpairs) ;
   getint(ph, "qtmode:", &qtmode) ;
   getint(ph, "probitmode:", &probitmode) ;
   getdbl(ph, "qtrbase:", &qtrbase) ;
   getdblss(ph, "qtrisk:", qtrisks, &qnumrisks) ;
   getdblss(ph, "qtrho:", qtrho, &n) ;

   if ((n>0) && (n!=qnumrisks)) {  
    fatalx("qtrho, qtrisk %d %d There is a size mismatch in the risk models\n", n, qnumrisks) ;
   }

   getstring(ph, "episim:", &epinames) ;
   getdbl(ph, "epithresh:", &epithresh) ;
   getdblss(ph, "epirisk:", epirisk, &n) ;

   if (getstring(ph, "unknowngender:", &tempname) >= 0) {
     unknowngender = tempname[0] ;
     free(tempname) ;
   }

   getstring(ph, "pubxindname:", &pubxindname) ; 



   if (getints(ph, "pubx:", pubx, 2) >0) {  
    pubxindiv = pubx[0] ;
    junk = pubx[1] ;
/// upwards compatible
   }

   npubxa = getints(ph, "pubxa:", pubxa, NPUBXA) ;       

   if (getints(ph, "casecontrol:", pubx, 2) >0) {  
    ncases = pubx[0] ;
    ncontrols = pubx[1] ;
    simnumindivs = ncases + ncontrols ;
   }

   getdbl(ph, "a1:", &a1) ;
   getdbl(ph, "b1:", &b1) ;
   getdbl(ph, "p1:", &p1) ;
   getdbl(ph, "psi1:", &psi1) ;

   getdbl(ph, "taulmean:", &taulmean) ;
   getdbl(ph, "taulsdev:", &taulsdev) ;
   if(taulsdev <= 0.0)
     fatalx("taulsdev %f Please enter a valid taulsdev value\n",taulsdev);
   getdblss(ph, "risk:", riskarr, &numrisks) ;
   getdblss(ph, "controlrisk:", criskarr, &n) ;
   getdbl(ph, "lrisk:", &lrisk) ;
   getdbl(ph, "fakespacing:", &fakespacing) ;
   getdbl(ph, "loclip:", &loclip) ; 
   getdbl(ph, "hiclip:", &hiclip) ;

   if(fakespacing <= 0.0)
     fatalx("fakespacing %f Please enter a valid fakespacing value\n",fakespacing) ;

   if ((n>0) && (n!=numrisks)) {  
    fatalx("controlrisk, risk %d %d There is a size mismatch in the risk models\n", n, numrisks) ;
   }

   getdblss(ph, "risk2:", risk2arr, &n) ;
   if ((n>0) && (n!=numrisks)) {  
    fatalx("risk2, risk %d %d There is a size mismatch in the risk models\n", n, numrisks) ;
   }

   getdblss(ph, "wrisk:", wrisk, &n) ;
   if (n>0) wr = wrisk ;
   if ((n>0) && (n!=numrisks)) {  
    fatalx("wrisk, risk %d %d There is a size mismatch in the risk models\n", n, numrisks) ;
   }

   if (getdblss(ph, "tauscal:", tauscal, &n) >= 0 ) { 
    if (n==1) tauscal[1] = tauscal[0] ;
   }

   getint(ph, "lmsieve:",   &lmsieve) ; 
   getint(ph, "markersim:", &markersim) ;
   getdbl(ph, "risksim:", &risksim) ;
   getstring(ph, "markername:", &markername) ;

   getdblss(ph, "sim2dvals:", sim2dvals, &n) ; 
   if (n!=4) vclear(sim2dvals, -1.0, 4) ;
   getint(ph, "sim2d_caseonly:", &sim2d_caseonly) ; 

   getdbl(ph, "muval:", &muval) ;
   getstring(ph, "thetafilename:", &thetafilename) ;                 
   getstring(ph, "lambdafilename:", &lambdafilename) ;                 
   getstring(ph, "freqfilename:", &freqfilename) ;                 
   getint(ph, "familynames:", &familynames) ;  // pedfile stuff
   getstring(ph, "indoutfilename:", &indoutfilename) ;
   getstring(ph, "indivoutname:", &indoutfilename) ; /* changed 11/02/06 */
   getstring(ph, "snpoutfilename:", &snpoutfilename) ;
   getstring(ph, "snpoutname:", &snpoutfilename) ; /* changed 11/02/06 */
   getstring(ph, "genooutfilename:", &genooutfilename) ; 
   getstring(ph, "genotypeoutname:", &genooutfilename) ; /* changed 11/02/06 */
   getstring(ph, "localoutfilename:", &localoutfilename) ;                 
   getstring(ph, "badpairsname:", &badpairsname) ;                 
   getstring(ph, "snpscorename:", &snpscorename) ;                 
   getstring(ph, "pubxname:", &pubxname) ;                 
   getdbl(ph, "allele_scale_fac:", &allele_scale_fac) ;                 

   tempname = NULL ;
   getstring(ph, "markerlistname:", &tempname) ;   
   if (tempname != NULL) {  
    badsnpname = strdup(tempname) ;
    free(tempname) ;
    printf("markerlistname deprecated\n") ;
// for backwards compatibility
   }
   getstring(ph, "badsnpname:", &badsnpname) ;
   getstring(ph, "goodsnpname:", &goodsnpname) ;

   getstring(ph, "output:", &scorename) ;
   printf("output: %s\n", scorename) ;
   getstring(ph, "auxout:", &score2name) ;
   getstring(ph, "trashdir:", &trashdir) ;

   getstring(ph, "genotypename:", &genotypename) ;
   getstring(ph, "snpname:", &snpname) ;
   getstring(ph, "indivname:", &indivname) ;
   getstring(ph, "genotoyoutfilename:", &genooutfilename) ;     /* obsolete */
   getstring(ph, "indtoyoutfilename:", &indoutfilename) ; 

   getstring(ph, "magicname:", &magicname) ;
   getstring(ph, "magicfilename:", &magicfilename) ;
   getint(ph, "mincasenum:", &mincasenum) ;

   getdblss(ph, "thpars:", qq, &n) ;
   if (n==2) {  
     thp1 = qq[0] ;
     thp2 = qq[1] ;
   }
     
   getdblss(ph, "thxpars:", pp, &n) ;

   if (n==3) {  
     thxp1 = pp[0] ;
     thxp2 = pp[1] ;
     thxp0 = pp[2] ;
   }

   
   /**AT: Adding new variables so that we can set the initial parameters for 
      the lambda prior distributions*/

   getdblss(ph, "lampars:",rr, &n) ;
   if (n==2) {  
     lp1 = rr[0] ;
     lp2 = rr[1] ;
   }

   getdblss(ph, "lamxpars:",uu, &n) ;
   if (n==3) {  
     lxp1 = uu[0] ;
     lxp2 = uu[1] ;
   }


   printf("### THE INPUT PARAMETERS\n");
   printf("PARAMETER NAME: VALUE\n");
   writepars(ph);
   closepars(ph) ;

  if (!fancyxtheta) thxp0 = -1 ;


  if (seed==0) seed = seednum() ;
  SRAND(seed) ;

  /** AT: Will just printout all the parameters in the parameter file 
  printf("taulmean: %9.3f\n", taulmean) ;
  printf("taulsdev: %9.3f\n", taulsdev) ;
  printf("tauscal: %9.3f %9.3f\n", tauscal[0], tauscal[1]) ;


  if (badsnpname != NULL) 
   printf("badsnpname: %s\n", badsnpname) ;

  if (snpname != NULL) 
   printf("snpname: %s\n", goodsnpname) ;

  printf("seed: %d\n",seed) ;
  printf("tlreest: %d\n",tlreest) ;
  printf("splittau: %d\n",splittau) ;
  if (markersim >= 0)  {
   dotoysim = YES ;
   printf("markersim: %d\n",markersim) ;
   printf("risksim: %9.3f\n",risksim) ;
  }

  if (simnumindivs>0) { 
   printf("simnumindivs: %d\n",simnumindivs) ;
  }

  if (theta2mode) {  
   printf("theta2mode\n") ;
  }

  if (lambda2mode>0) {  
   printf("lambda2mode %d\n",lambda2mode) ;
  }

  if (cleaninit) {  
   printf("cleaninit mode\n") ;
  }

  if (allele_scale_fac>=0) {  
   printf("allele_scale_fac: %9.3f\n", allele_scale_fac) ;
  }
  if (muval != 0.0) { 

  printf("*****************\n") ;
  printf("muval: %9.3f  tmumean: %9.3f\n",muval, tmumean) ;
  printf("*****************\n") ;

  } */

}
double mufun(double tave) 
{
  double y ;
  if (muval == 0.0) return 0.0 ;
  y = -muval*(tave-tmumean) ;
  return y ;
}
double mufun1(double theta) 
{
  double y ;
  if (muval == 0.0) return 0.0 ;
  y = -muval*(theta-tmumean) / (double) numindxok;
/* muval >0 theta usually < tmumean so this is usually positive */
  return y ;
}

void estglob(double *mstat, int mlen, int iter) 
{
   double *tt, *ttx, *qv  ;
   double *aaxx, *bbxx ;
   double y1, y2, m, mx ;
   int i, n, niter=20, nxx ;
   Indiv *indx ;  
   static int ncall = 0 ; 
   int mlmode ;
  
  ++ncall ;
  mlmode = NO ;
  if (ncall<5) {
    mlmode = YES ;
  }

   ZALLOC(tt, numindivs, double) ;
   ZALLOC(ttx, numindivs, double) ;
   ZALLOC(qv, numindivs, double) ;
   ZALLOC(aaxx, numindivs, double) ;
   ZALLOC(bbxx, numindivs, double) ;

   n = 0 ;
   for (i=0; i<numindivs; i++)  {
    indx = indivmarkers[i];
    if (indx->ignore) continue ;
    tt[n] = indx->theta_mode ;
    ttx[n] = indx->Xtheta_mode ;
    qv[n] = indx -> qval ;
    ++n ;
   }


   sqz(aaxx, tt, ttx, n) ;
   nxx = sqz(bbxx, ttx, ttx, n) ;
   m =  mstat[0] = asum(tt,n)/(double) n ;
   mstat[12] = asum2(tt,n)/(double) n ;

   mx = mstat[11] = mstat[1] = 0.0 ;
   mstat[2] = 0.0 ;

   if (nxx>0) {
    mx = mstat[11] = mstat[1] = asum(bbxx,n)/(double) nxx ;
    mstat[13] = asum2(bbxx,nxx)/(double) nxx ;
    mstat[2] = corr(aaxx, bbxx, nxx) ;
   }

   mstat[12] -= m*m ;
   mstat[13] -= mx*mx ;


   esttheta(qv, tt, ttx, n, niter, xchrom) ;
// side effect thp1 thp2 reestimated

   printf("estglob theta  %5d %9.3f %9.3f   %9.3f %9.3f %9.3f\n",
    iter, thp1, thp2, thxp1, thxp2, thxp0) ;

   n = 0 ;
   for (i=0; i<numindivs; i++)  {
    indx = indivmarkers[i] ;
    if (indx->ignore) continue ;
    tt[n] = indx->lambda_mode ;
    ttx[n] = indx->Xlambda_mode ;
    ++n ;
   }

   sqz(aaxx, tt, ttx, n) ;
   nxx = sqz(bbxx, ttx, ttx, n) ;
   m = mstat[14] =  asum(tt,n)/(double) n ;
   mstat[16] = asum2(tt,n)/(double) n ;
   mstat[16] -= m*m ;

   mx = mstat[15] = mstat[17] = 0.0 ;
   mstat[5] = 0.0 ;

   sampprg(&lp1, &lp2,tt, n, niter, mlmode) ;  
   if (nxx>0) {
    mx = mstat[15]  = asum(bbxx,n)/(double) nxx ;
    mstat[17] = asum2(bbxx,nxx)/(double) nxx ;
    mstat[5] = corr(aaxx, bbxx, nxx) ;
    sampprg(&lxp1, &lxp2, bbxx, nxx, niter, mlmode) ;  
   }


   mstat[3] = lp1/lp2 ;
   mstat[4] = lxp1/lxp2 ;

   printf("estglob lambda %5d %9.3f %9.3f   %9.3f %9.3f %9.3f\n",
    iter, lp1, lp2, lxp1, lxp2, asum(tt,n)/ (double) n) ;


   free(aaxx) ;
   free(bbxx) ;

   free(ttx) ;
   free(tt) ;

}
void checkdup(SNP **snpmark, int numsnp, int numindivs)
{
#define MINMARK 100
   int i, numnofake = 0, k, j, a, b ;
   int nmatch , nnomatch  ;
   double ytot, yhit ;
   SNP *cupt ;
   Indiv *inda, *indb;
   int *xmark, *xa, *xb ;
   double yx1, yx2, yn ;
   char status1[10], status2[10];
   int t1, t2 ;

   if (dupmode == NO) { 
    printf("skipping checkdup\n") ;
    return ;
   }

   numnofake = countnofake(snpmark, numsnp) ; 
   printf("###CHECKDUP RESULTS FOLLOW:\n");
   printf("##Indiv_Id1  Indiv_Id2\n");
   printf("##Num of genotypes matched: Num of genotypes mismatched\n");
   printf("##If the status of the two individuals does not match Status1: Status2\n");

   if (numnofake < MINMARK) {  
    printf("Too few markers for running checkdup: only %d markers\n", numnofake) ;
    return ;
   }
   ZALLOC(xmark, numnofake*numindivs, int) ;
    for (i=0, k=0; i< numsnp; i++) {
      cupt = snpmark[i] ;
      if (cupt -> ignore)  continue ;
      if (cupt -> isfake) continue ;    
      ++k ;
      for (j=0; j<numindivs; ++j) {  
        xmark[j*numnofake+k] = getgtypes(cupt, j) ;
      }
     }
    yx1 = yx2 = yn = 0.0; 
    for (a=0; a<numindivs; ++a) {  
     inda = indivmarkers[a] ; 
     if (inda->ignore) continue  ;
     xa  = xmark + a*numnofake ;
     for (b=a+1; b<numindivs; ++b) {  
      indb = indivmarkers[b] ; 
      if (indb->ignore) continue  ;
      xb  = xmark + b*numnofake ;
      nmatch = nnomatch = 0 ;
      for (k=0; k<numnofake; ++k) {  
       if (xa[k] < 0) continue ;
       if (xb[k] < 0) continue ;
       if (xa[k] == xb[k]) ++nmatch  ;
       else ++nnomatch ;
      }
      ytot = (double) (nmatch + nnomatch) ;
      if (ytot< MINMARK) continue ;
      yhit = ((double) nmatch) / ytot  ; 
      if (yhit>0.75) { 
       inda = indivmarkers[a] ;
       indb = indivmarkers[b] ;
       t1 = numvalids(inda, snpmark, 0, numsnp-1) ;
       t2 = numvalids(indb, snpmark, 0, numsnp-1) ;
       printf("dup? %s %s \n match: %d mismatch: %d   %d %d\n", 
        inda->ID, indb -> ID, nmatch, nnomatch, t1, t2) ;
       /*AT: 1/20/05 adding the status field to output*/
       /* NJP added valid coun */
       strcpy(status1, "Control") ; 
       if (inda->affstatus) strcpy(status1,"Case") ; 
       if (inda->ignore) strcpy(status1,"Ignore") ; 
       strcpy(status2, "Control") ; 
       if (indb->affstatus) strcpy(status2,"Case") ; 
       if (indb->ignore) strcpy(status2,"Ignore") ; 
       if(strcmp(status1,status2))
         {
           printf("status_1: %s status_2: %s\n", 
                  status1, status2) ;
         }
      }
      yx1 += yhit ;
      yx2 += yhit*yhit ;
      ++yn ;
     }
    }
   free(xmark) ;
   if (yn < 10.0) return ;                       
   yx1 /= yn ;
   yx2 /= yn ; 
   yx2 -= yx1*yx1 ;
   printf("match prob. mean: %9.3f  s.dev: %9.3f\n", yx1, sqrt(yx2+1.0e-10)) ;
}

int
countnofake(SNP **snpmark, int numsnp) 
{
     int i, numnofake = 0 ;
     SNP *cupt ;
     for (i=0; i< numsnp; i++) {
      cupt = snpmark[i] ;
      if (cupt -> ignore)  continue ;
      if (!(cupt -> isfake)) ++numnofake ;
     }
     return numnofake ;
}

int
countfake(SNP **snpmark, int numsnp) 
{
     int i, numfake = 0 ;
     SNP *cupt ;
     for (i=0; i< numsnp; i++) {
      cupt = snpmark[i] ;
      if (cupt -> ignore)  continue ;
      if (cupt -> isrfake)  continue ;
      if (cupt -> isfake) ++numfake ;
     }
     return numfake ;
}

int
countrfake(SNP **snpmark, int numsnp) 
{
     int i, numfake = 0 ;
     SNP *cupt ;
     for (i=0; i< numsnp; i++) {
      cupt = snpmark[i] ;
      if (cupt -> ignore)  continue ;
      if (cupt -> isrfake) ++numfake ;
     }
     return numfake ;
}

void printgall (char *pubxname) 
{
   FILE *fff ;
   Indiv *indx ;
   SNP *cupt ;
   int i, idnum, bb ; 
   double gg[3], y, *gx, y1;
   double popAfreq, popBfreq ;

   if (markernum < 0) return ;
   if (pubxname != NULL) openit(pubxname, &fff, "w") ;
   else fff = stdout ;

   cupt = snpmarkers[markernum] ;
   fprintf(fff, "### Gammas for marker: %s  chrom: %d pos: %9.0f\n",
     cupt -> ID, cupt -> chrom, cupt ->physpos) ;

   popAfreq = cupt -> adm_PopA_freq ;
   popBfreq = cupt -> adm_PopB_freq ;
   fprintf(fff, "###PopA_Freq: %9.3f  PopB_Freq: %9.3f\n", popAfreq, popBfreq) ;
   fprintf(fff,"%20s %12s %12s %12s %12s\n","Indiv_ID","G[0]","G[1]","G[2]","LogScore");
   
   for (i=0; i< numindivs; i++) { 
     indx = indivmarkers[i] ;
     if (indx -> ignore) continue ;
     gx = indx -> totgamms ;
     y = asum(gx, 3) ;
     if (y <=0.0 ) continue ;
     vst(gg, gx, 1.0/y, 3) ;
     fprintf(fff, "%20s  %12.6f %12.6f %12.6f", 
      indx ->ID, gg[0], gg[1], gg[2]) ;
     /* totscore only computed for cases */
     y1 = (indx -> totscore) / (double) numbigiter ;
     y1 = MAX(y1, 1.0e-20) ;
     y = log10(y1) ;
     fprintf(fff, " %12.3f",  y) ;
     idnum = indx -> idnum ;
    if (cupt -> gtypes != NULL) {
      bb = getgtypes(cupt, idnum) ;
      fprintf(fff, " %3d", bb) ;
    }
    fprintf(fff, "\n") ;
   }
  
   if (pubxname != NULL) fclose(fff) ;
}

void dolmsieve()  
{  
  int indiv ; 
  Indiv *indx ; 
  SNP *cupt ; 
  int *gtypes = NULL ;  
  int t ; 
  double y ; 
 
  cupt = NULL ;   
  cupt = snpmarkers[markernum] ; 
  if (risksim <= 0.0) fatalx("(dolmsieve) risksim unset\n") ; 
  gtypes = cupt -> gtypes ;  
  if (gtypes == NULL) fatalx("gtypes null for marker %s\n", markername) ; 
 
  for (indiv=0; indiv <numindivs; ++indiv) {   
    indx = indivmarkers[indiv] ; 
    if (indx->affstatus != YES) continue ; 
    t = getgtypes(cupt, indiv) ;
    if (t<0) continue ;   
    if (t >= 1) continue ;  /* overdominance */ 
    y = DRAND() ; 
    if (y < 1.0/risksim) continue ; 
    printf("switching status of %s\n", indx -> ID) ; 
    indx -> affstatus = NO ; 
  } 
} 

void printindivs()  
{  
  int indiv ; 
  Indiv *indx ; 
  SNP *cupt ; 
  char ss[10] ; 
  FILE *xfile ; 
  int num ; 
  int *gtypes = NULL ;  
  int n0=0 , n1= 0, numstates, nx, ignore, t ; 
  double theta, lambda ; 
 
  cupt = NULL ;   
  if (markername != NULL) {  
   cupt = snpmarkers[markernum] ; 
   gtypes = cupt -> gtypes ;  
   if (gtypes == NULL) printf("*** warning gtypes null for marker %s\n", markername) ;
  }
   
  if ((indoutfilename != NULL) && (strcmp(indoutfilename, "NULL") == 0))  return ; 
 
  if (indoutfilename != NULL) { 
   openit(indoutfilename, &xfile, "w") ; 
  } 
  else xfile = stdout ; 
 
 
  fflush(stdout) ; 
  fprintf(xfile,"%20s %5s  %8s  %5s\n","Indiv_ID","Gender","Status","Genotype_Cnt");
  for (indiv=0; indiv <numindivs; ++indiv) {   
     indx = indivmarkers[indiv] ; 
     strcpy(ss, "Control") ; 
     if (indx->affstatus) strcpy(ss,"Case") ; 
     if (indx->ignore) strcpy(ss,"Ignore") ;
     fprintf(xfile, "%20s ", indx->ID) ; 
     fprintf(xfile, "%c %12s",indx->gender, ss) ; 
     num = numvalids(indx, snpmarkers, 0, numsnps-1) ; 
     if (gtypes != NULL) { 
      t = getgtypes(cupt, indiv) ;
      fprintf(xfile, " %2d", t) ; 
      if (t >=0) { 
       gettln(cupt, indx, &theta, &lambda, &numstates, &ignore) ; 
       nx = numstates -1 ; 
       n1 += t ; 
       n0 += (nx-t) ; 
      } 
     } 
     fprintf(xfile, " %6d", num) ; 
     fprintf(xfile, "\n") ; 
  } 
  if (gtypes != NULL)  
   fprintf(xfile, "##marker counts %5d %5d\n", n0, n1) ; 
  if (indoutfilename != NULL)  
    fclose(xfile) ; 
} 

int nextreal (int start, int end, int inc) 
/* search for real marker */
{
  int x ;
  x = start ;
  if ( (x<end) && (inc <0)) 
    fatalx("bad nextreal %d %d %d\n",start, end, inc ) ;
  if ( (x>end) && (inc >0)) 
    fatalx("bad nextreal %d %d %d\n",start, end, inc ) ;
  for (x=start; x != end ; x += inc)  {
   if (!(snpmarkers[x] -> isfake))  return x ;
  }
  if (!(snpmarkers[end] -> isfake))  return end ;
  return -1 ;
}

void ckit(int k) 
{
 printf("got here %d\n",k) ;
 fflush(stdout) ; 
}
void sampfreqx(double *tauscal, SNP **snpmarkers, SNPMC **snpmcmarkers, 
  int numsnps, int mode)  

{
 double **xnn ; 
 double **ynn ; 
 int *indxsnp ;
 double *p1, *p2, *r1, *r2 ; 
 int  i, j, nx ;
 double pp, p ;

 SNP *cupt, *tcupt ;
 SNPMC *cumcpt ;
 static int ncall = 0 ;

 ++ncall ;

 
 nx = 0 ;
 ZALLOC(indxsnp, numsnps, int) ;
 for (i=0; i<numsnps; i++)  {  
   cupt = snpmarkers[i] ;
   if (cupt->isfake) continue ;
   if (cupt->ignore) continue ;
   if (cupt->isrfake) continue ; 
   indxsnp[nx] = i ;
   ++nx ;
 }

 ZALLOC(xnn, nx, double *) ;
 ZALLOC(ynn, nx, double *) ;
 ZALLOC(p1, nx, double) ;
 ZALLOC(p2, nx, double) ;
 ZALLOC(r1, nx, double) ;
 ZALLOC(r2, nx, double) ;

 for (i=0; i<nx; i++) {  
  ZALLOC(xnn[i], 2, double) ;
  ZALLOC(ynn[i], 2, double) ;
 }

 for (i=0; i<nx; i++) {  

  j = indxsnp[i] ;
  cupt = snpmarkers[j] ;
  cumcpt = snpmcmarkers[j] ;
  p1[i] = cupt -> PopA_freq ;
  p2[i] = cupt -> adm_PopA_freq ;

  xnn[i][0]  = cupt -> popA_nn[0] ;
  xnn[i][1]  = cupt -> popA_nn[1] ;

  ynn[i][0]  = cumcpt -> xd[0];
  ynn[i][1]  = cumcpt -> xd[1];
 } 
 copyarr(p1, r1, nx) ;
 copyarr(p2, r2, nx) ;
 dosampfreq(xnn,ynn,nx, tauscal+0, r1, r2,YES) ;
 for (i=0; i<nx; i++) {  

  j = indxsnp[i] ;
  cupt = snpmarkers[j] ;
  cumcpt = snpmcmarkers[j] ;

  cupt -> PopA_freq = r1[i];
  cupt -> adm_PopA_freq = r2[i];
 }

 for (i=0; i<nx; i++) {  

  j = indxsnp[i] ;
  cupt = snpmarkers[j] ;
  cumcpt = snpmcmarkers[j] ;
  p1[i] = cupt -> PopB_freq ;
  p2[i] = cupt -> adm_PopB_freq ;

  xnn[i][0]  = cupt -> popB_nn[0] ;
  xnn[i][1]  = cupt -> popB_nn[1] ;

  ynn[i][0]  = cumcpt -> xd[2];
  ynn[i][1]  = cumcpt -> xd[3];
 } 
 copyarr(p1, r1, nx) ;
 copyarr(p2, r2, nx) ;
 dosampfreq(xnn,ynn,nx, tauscal+1, r1, r2,YES) ;
 for (i=0; i<nx; i++) {  

  j = indxsnp[i] ;
  cupt = snpmarkers[j] ;
  cumcpt = snpmcmarkers[j] ;

  cupt -> PopB_freq = r1[i];
  cupt -> adm_PopB_freq = r2[i];

 }


 if (mode == YES) {  
  for (i=0; i<nx; i++) {  
   j = indxsnp[i] ;
    tcupt = snpmarkers[j] ;
    cumcpt = snpmcmarkers[j] ;
    if (pubfmodern) cumcpt = modmcmarkers[j] ; 
    addpstats(tcupt, cumcpt) ;
  }
 }


 for (i=nx-1; i>=0; i--) {  
  free(ynn[i]) ;
  free(xnn[i]) ;
 }
 free(indxsnp) ;
 free(r2) ;
 free(r1) ;
 free(p2) ;
 free(p1) ;
 free(ynn) ;
 free(xnn) ;
}

void sampfreqy(double *tauscal, SNP **snpmarkers, SNPMC **snpmcmarkers, 
  int numsnps, int mode)  
// just like sampfreqx by isrfakes only and NO tau reest

{
 double **xnn ; 
 double **ynn ; 
 int *indxsnp ;
 double *p1, *p2, *r1, *r2 ; 
 int  i, j, nx ;
 double pp, p ;

 SNP *cupt, *tcupt ;
 SNPMC *cumcpt ;
 static int ncall = 0 ;

 ++ncall ;

 
 nx = 0 ;
 ZALLOC(indxsnp, numsnps, int) ;
 for (i=0; i<numsnps; i++)  {  
   cupt = snpmarkers[i] ;
   if (!(cupt->isrfake)) continue ;
   indxsnp[nx] = i ;
   ++nx ;
 }
 if (nx == 0) {
  free(indxsnp) ;
  return ;
 }

 ZALLOC(xnn, nx, double *) ;
 ZALLOC(ynn, nx, double *) ;
 ZALLOC(p1, nx, double) ;
 ZALLOC(p2, nx, double) ;
 ZALLOC(r1, nx, double) ;
 ZALLOC(r2, nx, double) ;

 for (i=0; i<nx; i++) {  
  ZALLOC(xnn[i], 2, double) ;
  ZALLOC(ynn[i], 2, double) ;
 }

 for (i=0; i<nx; i++) {  

  j = indxsnp[i] ;
  cupt = snpmarkers[j] ;
  cumcpt = snpmcmarkers[j] ;
  p1[i] = cupt -> PopA_freq ;
  p2[i] = cupt -> adm_PopA_freq ;

  xnn[i][0]  = cupt -> popA_nn[0] ;
  xnn[i][1]  = cupt -> popA_nn[1] ;

  ynn[i][0]  = cumcpt -> xd[0];
  ynn[i][1]  = cumcpt -> xd[1];
 } 
 copyarr(p1, r1, nx) ;
 copyarr(p2, r2, nx) ;
 dosampfreq(xnn,ynn,nx, tauscal+0, r1, r2,NO) ;
 /* and copy frequencies back */
 for (i=0; i<nx; i++) {  

  j = indxsnp[i] ;
  cupt = snpmarkers[j] ;
  cumcpt = snpmcmarkers[j] ;

  cupt -> PopA_freq = r1[i];
  cupt -> adm_PopA_freq = r2[i];
 }

 for (i=0; i<nx; i++) {  

  j = indxsnp[i] ;
  cupt = snpmarkers[j] ;
  cumcpt = snpmcmarkers[j] ;
  p1[i] = cupt -> PopB_freq ;
  p2[i] = cupt -> adm_PopB_freq ;

  xnn[i][0]  = cupt -> popB_nn[0] ;
  xnn[i][1]  = cupt -> popB_nn[1] ;

  ynn[i][0]  = cumcpt -> xd[2];
  ynn[i][1]  = cumcpt -> xd[3];

 } 
 copyarr(p1, r1, nx) ;
 copyarr(p2, r2, nx) ;
 dosampfreq(xnn,ynn,nx, tauscal+1, r1, r2,NO) ;

 for (i=0; i<nx; i++) {  

  j = indxsnp[i] ;
  cupt = snpmarkers[j] ;
  cumcpt = snpmcmarkers[j] ;

  cupt -> PopB_freq = r1[i];
  cupt -> adm_PopB_freq = r2[i];
 }

 if (mode == YES) {  
  for (i=0; i<nx; i++) {  
   j = indxsnp[i] ;
    tcupt = snpmarkers[j] ;
    cumcpt = snpmcmarkers[j] ;
    if (pubfmodern) cumcpt = modmcmarkers[j] ; 
    addpstats(tcupt, cumcpt) ;
  }
 }


 for (i=nx-1; i>=0; i--) {  
  free(ynn[i]) ;
  free(xnn[i]) ;
 }
 free(indxsnp) ;
 free(r2) ;
 free(r1) ;
 free(p2) ;
 free(p1) ;
 free(ynn) ;
 free(xnn) ;
}


void addpstats(SNP *tcupt, SNPMC *cumcpt)  
{
    double p, pp ;
    p  = tcupt -> adm_PopA_freq ;
    pp = tcupt -> adm_PopB_freq ;
    if (cumcpt -> pubfmodern) {  
      p  = tcupt -> PopA_freq ; 
      pp = tcupt -> PopB_freq ; 
    } 
    cumcpt -> ap1 += p ;
    cumcpt -> ap2 += p*p ;
    cumcpt -> cp1 += pp ;
    cumcpt -> cp2 += pp*pp ;  
}

void dosampfreq(double **xnn, double **ynn, int n, double *palpha, 
 double *r1, double *r2, int taumode) 
/** 
 r1 modern 
 r2 ancestral frequency 
 used to be doreest
*/
{

    double xa, xb, qa, qb ;
    double xn[2] ;
    double alpha, p, pe, qe, x, t ; 
    double lmean, lsdev ;
    int i, iter ; 
    static int ncall=0 ;
    int nbad =0,  ngood = 0 ;
    double lof = .001 , hif = 0.999 ; 
    ++ncall ;


    lmean = taulmean  ;  lsdev = taulsdev ;

    alpha = *palpha ;

/** 
 step 1 
 sample r2.  Conditional distribution 
 is beta 
*/

     if (nopopsmode) {  
       *palpha = TAUINIT ;
      for (i=0; i<n; i++)  {  
       xa = 1.0 +  ynn[i][0] ;
       xb = 1.0 +  ynn[i][1] ;
       p = r2[i] = xranbeta(xa, xb, lof, hif) ;
      }
      return ;
     }

     for (i=0; i<n; i++)  {  
      xa = alpha*r1[i]  + ynn[i][0] ;
      xb = alpha*(1-r1[i]) + ynn[i][1] ;
      p = r2[i] = xranbeta(xa, xb, lof, hif) ; 
/* xranbeta insists we are inside lower and upper bound */
     }
/** 
 step 2 sample alpha  
 lmean lsdev are prior mean and standard dev 
 for log10(alpha)
*/
     if (taumode) 
      sampprt(&alpha, r1, r2, n, lmean, lsdev) ;

/** 
 step 3 sample r1 
*/
    for (i=0; i<n; i++)  {  

      xn[0] = xnn[i][0] ;
      xn[1] = xnn[i][1] ;

      if (xn[0] >= 0)   {
/* xn are beta priors for r1 */
       r1[i] = sampit1c(xn, alpha, r2[i]) ;
       ++ngood ;
      }
      else ++nbad ;
    }
   *palpha = alpha ;
}

void
dostat(char *legend, double *ss, int n)  
/** 
 print some descriptive stats for 
 array of statistics 
*/
{
  double *tt, y1, ysig, y2, sig, y ;  
  double rho ;
  int k ;

  if (n<=2) return ;
  if (numlagprint == 0) return ;

  printf("\n") ;

  ZALLOC(tt, n, double) ;
  copyarr(ss, tt, n) ; 

  y1 = asum(tt, n) / (double) n ;
  vsp(tt, tt, -y1, n) ;

  y2 = asum2(tt, n)/ (double) n ;  
  ysig = sqrt(y2 + 1.0e-20) ;

  vst(tt, tt, 1.0/ysig , n) ;

  printf("%20s mean: %9.3f s.err: %9.3f\n",legend, y1, ysig) ;
  for (k=1; k  <= MIN(numlagprint, n/3); ++k) { 
   rho = vdot(tt, tt+k, n-k) ;
   y = (double) (n-k) ;
   rho /= y ;
   sig = sqrt(y) * rho ;
   printf("lag: %2d corr: %9.3f sig: %9.3f\n", k, rho, sig) ;
  }


  free(tt) ;

}
void prmstat(int iter, double *xx, int n) 
{
   static int ncall = 0 ;
   int i ;

   if (scorefile == NULL) return ;
   ++ncall ;

   if (ncall==1)  {  
    fprintf(scorefile,"%5s ", "#iter") ;
    fprintf(scorefile,"%9s ", "t mean") ;
    fprintf(scorefile,"%9s ", "tx mean") ;
    fprintf(scorefile,"%9s ", "t corr") ;
    fprintf(scorefile,"%9s ", "l mean") ;
    fprintf(scorefile,"%9s ", "lx mean") ;
    fprintf(scorefile,"%9s ", "l corr") ;
    fprintf(scorefile,"%9s ", "tau (popA)");
    fprintf(scorefile,"%9s ", "tau (popB)");
    if (rdismode)  {
     fprintf(scorefile,"%9s ", "disflat") ;
    }
    fprintf(scorefile,"%9s ", "log score");
    fprintf(scorefile,"\n") ;
   }
   fprintf(scorefile,"%5d ", iter) ;
   for (i=0; i<n;i++)  {
    fprintf(scorefile,"%9.3f ", xx[i]) ;
   }

   fprintf(scorefile,"\n") ;


}

void
esttheta(double *qv, double *tt, double *ttx, int n, int niter, int xc)  
//  actually estimates  thp1, thp2 and thxp?
{
    int i, nxx ;
    double oldp1, oldp2 ;
    static int first = YES  ;
    double *bbxx, *aaxx ;
    
    if (noxdata) {
     sampprb(&thp1, &thp2, qv, tt, n, niter, ltmean, ltsig, qtmode) ;  
     return ;
    }
  ZALLOC(aaxx, n, double) ;
  ZALLOC(bbxx, n, double) ;

  sqz(aaxx, tt, ttx, n) ;
  nxx = sqz(bbxx, ttx, ttx, n) ;

  oldp1 = thp1 ;
  oldp2 = thp2 ;
  sampprb(&thp1, &thp2, qv, tt, n, niter, ltmean, ltsig, qtmode) ;  

  if (debug && first) {
   printf("esttheta in: %12.6f %12.6f\n", oldp1, oldp2) ;
   printf("esttheta out: %12.6f %12.6f\n", thp1, thp2) ;
   for (i=0; i<n; i++) { 
    printf("%4d %9.3f\n",i, tt[i]) ;
   }
  }

  if(first)
    {
      printf("\n");
      printf("###RESULTS FOR EACH MARKOV CHAIN MONTE CARLO ITERATION\n");
      printf("##estglob theta: Iteration_Num  thp1  thp2  thxp0  thxp1  thxp2\n");
      printf("##thp1, thp2: Are parameters for the prior distribution of theta, and thxp0,thxp1,thxp2 are the same for theta on X chromosome\n");
      printf("##estglob lambda: Iteration_Num  lp1  lp2  lxp1  lxp2  ave_lambda\n");
      printf("##lp1, lp2: Are parameters for the prior distribution of lambda, and lxp1,lxp2 are the same for lambda on X chromosome\n\n");
  }

  if (thxp0<0) {
   sampprb(&thxp1, &thxp2, NULL, bbxx, nxx, niter, ltmean, ltsig, NO) ;  
  }
  else {
   if (nxx>0) {
    sampprbb(&thxp0, &thxp1, &thxp2, aaxx, bbxx, nxx, niter, lxmean, lxsig) ;
   }
  }
  if (thxp1<1.0) fatalx("bad sampprbb\n") ;
  first = NO ;
  free(aaxx) ;
  free(bbxx) ;
}

int sqz(double *xout, double *xin, double *xp, int n) 
// copy xin to xout only with non-neg xp
{
  int i, nxx=0 ;
  for (i=0; i<n; i++) {  
   if (xp[i]<0) continue ;
   xout[nxx] = xin[i] ;
   ++nxx ;
  }
  return nxx ;
}

void initindiv(Indiv **indm, int n) 
{
  Indiv *indx ;
  int i ;
  double tlo = 0.05, thi = .95, llo = 1, lhi = 20 ;
  double txlo , txhi, lxlo = 1, lxhi = 20 ;
  double *x ;
/* clip initial values */

  txlo = tlo; txhi = thi ; 
  lxlo = llo; lxhi = lhi ; 

  if (cleaninit) { 

   for (i=0; i<n; ++i)  {  

    indx = indm[i] ;
    indx -> theta_mode   = ranbeta(20,80) ;
    indx -> Xtheta_mode  = ranbeta(20,100) ;

    indx -> lambda_mode  =  rangam(lp1)/lp2 ;
    indx -> Xlambda_mode  = rangam(lxp1)/lxp2 ;

   }

  }

  for (i=0; i<n; ++i)  {  
   indx = indm[i] ;
   x = &indx->theta_mode ;
   vclip(x, x, tlo, thi, 1) ;
   x = &indx->Xtheta_mode ;
   vclip(x, x, txlo, txhi, 1) ;

   x = &indx->lambda_mode ;
   vclip(x, x, llo, lhi, 1) ;
   x = &indx->Xlambda_mode ;
   vclip(x, x, lxlo, lxhi, 1) ;
  }
}
double
thetaxsum(Indiv **indm, int n) 
{   
  Indiv *indx ;
  int i ;
  double tsum=0 ;

  for (i=0; i<n; i++)  {  
   indx = indm[i] ;
   if (indx -> ignore) continue ;
   tsum += indx -> Xtheta_mode  ;
  }


  return tsum ;

}

void
scalecounts(SNP **snpm, int numsnps, double fac) 

{
   int i ;
   SNP *cupt ;

   for (i=0; i< numsnps; i++) {  
    cupt = snpm[i] ;
    if (cupt -> isfake) continue ;
    if (cupt -> ignore) continue ;
    scalitx(cupt -> popA_nn, fac, cupt -> PopAftrue) ;
    scalitx(cupt -> popB_nn, fac, cupt -> PopBftrue) ;
   }
}

void scalitx(int *nn, double fac, double p) 
{
    int n, nx, n0, n1 ;
    double x ;

    n = nn[0] + nn[1] - 2 ;
    n = MAX(n, 0) ;
    x = fac * (double) n ;  
    nx = nnint(x) ;
    nn[0] = n0 = ranbinom(nx, p) ;
    nn[1] = n1 = nx - n0 ;        
    ++nn[0] ;
    ++nn[1] ;
}

void printmutab(double *txsa, int n)  
{
    double *a, *b; 
    double mm, y ;  
    int i ;

    ZALLOC(a, n, double) ;
    ZALLOC(b, n, double) ;

    vsp(a, txsa, -tmumean, n) ;
    for (i= 100; i <= 4000; i += 100 ) {  
     mm = (double) i ;
     vst(b, a, -mm, n) ;
     vexp(b, b, n) ;
     y = asum(b, n) / (double) n ;
     printf("mu: %9.3f  log_10 ave(e-mutx): %9.3f\n", mm, log10(y)) ;
    }




   free(a) ;
   free(b) ;

}

void
printtmeans(double *tmean) 

{
  double sig ;
  printf("\n");
  printf("###POSTERIOR ESTIMATES\n") ; 
  printf("theta  mean  %10.4f\n", tmean[0]) ;
  printf("thetax mean  %10.4f\n", tmean[1]) ;
  sig = sqrt(tmean[2]) ;
  printf("theta  var  %10.4f   sdev: %10.4f\n", tmean[2], sig) ;
/* NOT posterior mean sdev */
  sig = sqrt(tmean[3]) ;
  printf("thetax var  %10.4f   sdev: %10.4f\n", tmean[3], sig) ;

  printf("\n") ;
  printf("lambda  mean  %10.4f\n", tmean[4]) ;
  printf("lambdax mean  %10.4f\n", tmean[5]) ;
  sig = sqrt(tmean[6]) ;
  printf("lambda  var  %10.4f   sdev: %10.4f\n", tmean[6], sig) ;
/* NOT posterior mean sdev */
  sig = sqrt(tmean[7]) ;
  printf("lambdax var  %10.4f   sdev: %10.4f\n", tmean[7], sig) ;

  printf("tau (PopA) %9.3f\n", tmean[8]) ;
  printf("tau (PopB) %9.3f\n", tmean[9]) ;

  if (rdismode)  {
   printf("flat %9.3f\n", tmean[10]) ;
  }

}
void indstatus(char *sx, Indiv *indx)  
{
  if (indx->ignore) { 
   strcpy(sx,"Ignore")  ;
   return ;
  }
  if (indx->affstatus) { 
   strcpy(sx,"Case")  ;
   return ;
  }
  strcpy(sx,"Control")  ;
}

void setfc(SNP **snpm, int numsnps) 
// also allocates fchrom , lchrom
{
  int i, j, chrom ;
  SNP *cupt ;
  double dis, totdis, pos1, pos2 ;

  ZALLOC(fchrom, 25, int) ;
  ZALLOC(lchrom, 25, int) ;

  ivclear(fchrom, 999999, 25) ;
  ivclear(lchrom, -999999, 25) ;
  
/* initialize real marker array */
  for (i=0; i<numsnps; i++) {
    cupt = snpm[i] ;
    if (cupt -> isfake) continue ;
    if (cupt -> ignore) continue ;
    chrom = cupt -> chrom ;
    if ((noxdata) && (chrom==23)) continue  ;
    fchrom[chrom] = MIN(fchrom[chrom],i) ;
    lchrom[chrom] = MAX(lchrom[chrom],i) ;
  }

  totdis = 0.0 ;
   if(details)
     {
       printf("\n###GENETIC DISTANCE FOR ALL CHROMOSOMES\n");
       printf("##Chr_Num: chromosome num, First_SNP and Last_SNP: First and last markers, Gen_dist: Genetic distance\n");
       printf("%5s %9s %9s %9s\n","Chr_Num","First_SNP","Last_SNP","Gen_dist");
     }

  for (j=1; j<=23; j++) {  
   if (lchrom[j]<0) continue ;
   cupt = snpm[fchrom[j]]; 
   pos1 = cupt -> genpos ;
   cupt = snpm[lchrom[j]]; 
   pos2 = cupt -> genpos ;
   dis = pos2 - pos1 ;
   totdis += dis ;
   printf("chrom: %5d  first: %5d  last:  %3d", 
     j, fchrom[j], lchrom[j]) ;
   printf(" distance: %9.3f\n", dis) ;
  }
  printf("total distance: %9.3f\n", totdis) ;
}

void pubdis(SNP **snpm, SNPMC **snpmc, int nsnps)  
{
     int i, j ;
     SNP *cupt, *cupt2 ;
     SNPMC *cumcpt, *cumcpt2 ;
     double xpos, xpos2, gpos, gpos2 ;
     double estdis, dis, flat = rdis ;
     double xgamm, alpha, z ;

     printf("pubdis: flat: %9.3f\n", flat) ;
     for (i=0; i<nsnps; i++)  {  
    
      cupt = snpm[i] ;

      if (cupt -> isfake) continue ;
      if (cupt -> ignore) continue ; 
      cupt2 = cupt -> pcupt;      
      if (cupt2 == NULL) continue ;
      j = cupt2 -> markernum ;
      cumcpt = snpmc[i] ;
      cumcpt2 = snpmc[j] ;
      xpos = cupt -> gpsum / cupt -> gpnum ;
      xpos2 = cupt2 -> gpsum / cupt2 -> gpnum ;
      gpos = cupt -> estgenpos ;
      gpos2 = cupt2 -> estgenpos ;
      dis = xpos2 - xpos ;  
      estdis = gpos2 - gpos ;
// xgamm is supposedly styandard gamma distributed with shape alpha 
// print z  score that corresponds 
      xgamm = dis*flat ;
      alpha = estdis*flat ;
      z = zprobg(alpha, xgamm) ;  // z statistic  
      if (isnan(z)) fatalx("%12.6f %12.6f %12.6f\n", alpha, xgamm, z) ;
      printf("%15s %10.4f %10.4f ", cupt -> ID, gpos, xpos) ;
      printf(" %15s %10.4f %10.4f ", cupt2 -> ID, gpos2, xpos2) ;
      printf("   %10.4f %10.4f", estdis, dis) ;
      printf(" %9.3f", z) ;
      printf("\n") ;
     }
     printf("======= end pubdis ===========\n") ;
}

void setgphys(SNP **snpm, int nsnps)   

{
 int k ; 
 SNP *cupt ;
 double pos, gpos ;

 clearestdis(snpm, nsnps) ;
 
 for (k=0; k< nsnps ; ++k)  { 
  cupt = snpm[k] ;
  pos = cupt -> physpos ;
  gpos = pos / 1.0e8 ;
  cupt -> genpos = cupt -> estgenpos = gpos ;
 }
 setestdis(snpm, nsnps) ;
}

double zprobg(double a, double x) 
// return normal variate with appropriate tail  
// looking at gamma tail area 
// probably should be in statsubs.c 
{
  double rtail, ltail, ans, junk, mintail ;
  int nbad = 0 ;

  
 mintail = ntail(10.0) ;
 for (;;)  { 
  rtail = 2.0 ;
  if (x>a) {
   rtail = rtlg(a, x) ;  
  }  
  if (rtail < 0.99) { 
   rtail = MAX(rtail, mintail) ;
   ans = zprob(rtail) ;
   if (isnan(ans)) { printf("zprobg %9.3f\n", rtail)  ;
    printf("zzr %15.9e %9.3f\n", rtail, ans) ;
    ++nbad ;
    if (nbad==10) fatalx("no good\n") ;
    continue ;
   }
   return ans ;
  }
  ltail = ltlg(a, x) ;  
   ltail = MAX(ltail, mintail) ;
   ans = -zprob(ltail) ;
   if (isnan(ans)) { printf("zprobg %9.3f\n", ltail)  ;
    printf("zzl %15.9e %9.3f\n", ltail, ans) ;
    ++nbad ;
    if (nbad==10) fatalx("no good\n") ;
    continue ;
   }
   return ans ;
 }
}

