#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <nicklib.h>
#include <getpars.h>
#include "regsubs.h" 

#define WVERSION "1000"  

int verbose = NO ;

typedef struct { 
  int nnode ; 
  int nedge ;
  int *isleaf ; 
  int *label ; 
  int root ;  
  int *matrix ;  
  double *fst ;
} GRAPH ;

typedef struct { 
  int gnode ; 
  int glabel ;
  int isroot ;  
  int isleaf ;  
  struct NODE  *left   ;
  struct NODE  *right  ;
  struct NODE  *parent ;
  double hetleaves ;   // het value for nodes whose parent is node   
  double hetx ;  
  double timmax ;  
  double timmin ;  
  double fst ; // fst for edge to parental
} NODE ;

// version that builds tree.  

int *epair2v, *v2epair, *vmask ;

#define MAXGRAPH 100
#define MAXLABEL 200 
GRAPH *grph[100] ; 

char *infilename = "pop6.out" ;

int ngraph = 0 ;
int maxeq ;
int numleaf ;
int maxlabel = MAXLABEL ;

GRAPH *oldgg, *gg ;
double topfst ;

int getnumleaf(char *infilename) ;
int  nodenum(GRAPH *gg, int label)  ;
void addedge(GRAPH *g2, GRAPH *g1, int oldnode, int newlabel, int isleaf)   ;
void addnode(GRAPH *gg, GRAPH *g1, int oldedge, int newlabel)   ;
int edgenum (GRAPH *gpt, int lab1, int lab2) ;
void trivgraph(GRAPH *gg, int label) ;
void copygraph(GRAPH *g1, GRAPH *g2) ;
void printgraph(GRAPH *gg)  ;
int nodenum (GRAPH *gpt, int lab1) ;
void proot(GRAPH *gg, int t, int *z1) ;
int othernode(GRAPH *gg, int t, int k) ;
void nodes(GRAPH *gg, int *t1, int *t2, int k) ;
int epath(GRAPH *gg, int *ep, int lab1, int lab2) ;

double *sc2, *sc3, *sc4 ;
double *serr2, *serr3, *serr4 ;
int numsc2, numsc3, numsc4 ;
int loadhets(char *fname, double *hets, double *hsig, int numleaf) ;
int loadsc2(char *fname, double *sc, double *serr, int numleaf) ;
int loadsc3(char *fname, double *sc, double *serr, int numleaf) ;
int loadsc4(char *fname, double *sc, double *serr, int numleaf) ;
void loadee(double *ee, int *p1, int *p2,  int e)  ;
int addr4(int a,int b,int c,int d,int n)   ;

void pabcd(int *bb)  ;
void wabcd(int *bb, int a, int b, int c, int d) ;

int
sqzx(double *eq, double *eqx, double *rhs,  double *wt, double *rhsx, 
  int ne, int numv, int *vmask) ; 
double
rscor(double *rr, double *ans, double *eqx, double *rhsx, int neq, int nvv) ;
void addfudge(double *eqx, double *rhsx, int neq, int nvv) ;
void getquartet(double *sc4, double *serr4, int *pa, int *pb, int *pc, int *pd, int numleaf) ;

void destroygraph(GRAPH *g1)  ;
void wipegrph(int n)  ;
int toplabel(GRAPH *gg);

void setislab(int *islab, GRAPH *gg);
double scoregraph(GRAPH *gg)  ;
int klabs(int *islabel, int a, int b, int c, int d) ;
double missxx(int *islabel) ; // score equations not in graph 
double growtree(GRAPH *newgg, GRAPH *oldgg) ;

void readcommands(int argc, char **argv) ;

void mkrtree(NODE *rnode, GRAPH *gg, int rlabel, int plabel) ;
void getkidlabels(GRAPH *gg, int rlabel, int *lableft, int *labright, int plabel) ;
void wipenode(NODE *rnode) ;
int listnode(NODE **znode,  NODE *rnode)  ;
int listleaves( NODE *xnode, int *xx) ;
double median(double *ww, int len)  ;

void calctim(NODE *rnode, int mode)  ;
void printtim(NODE *rnode)  ;

int main(int argc, char **argv)
{
 GRAPH *g1, *g2, *tgraph1, *tgraph2, *tgraph3 ;
 int x, k, n, e, i, j, t, t1, t2, iter, numiter=2   ;
 int node1, node2 ;
 int a, b, c, d, v, numv ;
 int *pp, *p1, *p2 ;
 int *vmask, *vm ;
 double *eq, *eqx, *rhs, *rhsx, *wt ;
 int m2, m3, m4, ne, nvv, neq ;
 double *hest, *hsig ;
 double y, y1, y2 ; 
 double *ans, *rr, *ee ;
 int **abcd ; 
 double ysc, yscmin, yscthresh ;  
 int vbest ; 
 int xint = 50 ;
 double *xmax, *xmin ;
 double ymax, ymin ;
 NODE *rnode, *tnode, **znode ;
 int nznode ;
 double ww[MAXGRAPH] ;
 int x1, x2, *x1list, *x2list ;
 
 
 
 readcommands(argc, argv) ;
 printf("## infilename: %s\n", infilename) ;
 numleaf = getnumleaf(infilename) ;
 printf("numleaf: %d\n", numleaf) ;

 ZALLOC(sc2, numleaf*numleaf, double) ;
 ZALLOC(sc3, numleaf*numleaf*numleaf, double) ;
 ZALLOC(sc4, numleaf*numleaf*numleaf*numleaf, double) ;

 ZALLOC(serr2, numleaf*numleaf, double) ;
 ZALLOC(serr3, numleaf*numleaf*numleaf, double) ;
 ZALLOC(serr4, numleaf*numleaf*numleaf*numleaf, double) ;

 ZALLOC(hest, numleaf*numleaf, double) ;
 ZALLOC(hsig, numleaf*numleaf, double) ;

 loadhets(infilename, hest, hsig, numleaf) ;

 printf("hetrate: \n") ; 
 printmatw(hest, numleaf, numleaf, numleaf) ;
 printf("\n\n") ;

 numsc2 = m2 = loadsc2(infilename, sc2, serr2, numleaf) ; 
 numsc3 = m3 = loadsc3(infilename, sc3, serr3, numleaf) ; 
 numsc4 = m4 = loadsc4(infilename, sc4, serr4, numleaf) ; 

 maxeq = m2 + m3 + m4 ;  

 getquartet(sc4, serr4, &a, &b, &c, &d, numleaf) ;


 for (k=0; k< MAXGRAPH; ++k) {  
  ZALLOC(grph[k], 1, GRAPH) ;
 }
 ZALLOC(tgraph1, 1, GRAPH) ;
 ZALLOC(tgraph2, 1, GRAPH) ;
 ZALLOC(tgraph3, 1, GRAPH) ;

 g1 = grph[ngraph] ; 
 trivgraph(g1, a) ;  
 ++ngraph ; 
 printgraph(g1) ;

 oldgg=grph[ngraph-1] ; gg = grph[ngraph] ;
 addedge (gg, oldgg, a, b, YES) ;  
 ++ngraph  ;

 printgraph(gg) ;

 oldgg=grph[ngraph-1] ; gg = grph[ngraph] ;
 k = edgenum(oldgg, a, b) ;
 ++xint  ;
 addnode (gg, oldgg, k, xint) ;  
 ++ngraph  ;
 node1 = nodenum(gg, xint) ;

 printgraph(gg) ;

 oldgg=grph[ngraph-1] ; gg = grph[ngraph] ;
 ++xint ;
 addedge (gg, oldgg, node1, xint, NO) ;  
 ++ngraph  ;
 node1 = nodenum(gg, xint) ;

 oldgg=grph[ngraph-1] ; gg = grph[ngraph] ;
 addedge(gg, oldgg, node1, c, YES) ;
 ++ngraph  ;
 node1 = nodenum(gg, xint) ;

 oldgg=grph[ngraph-1] ; gg = grph[ngraph] ;
 addedge(gg, oldgg, node1, d, YES) ;
 ++ngraph  ;

 printgraph(gg) ;
 copygraph(gg, tgraph1) ; 
 wipegrph(ngraph) ; 
 ngraph = 0 ;

 y = scoregraph(tgraph1) ;      // score graph        
 printf("initial score: %9.3f\n", y) ;

 for (k=4; k<numleaf; ++k) {
  growtree(tgraph2, tgraph1) ;
  destroygraph(tgraph1) ; 
  copygraph(tgraph2, tgraph1) ; 
  destroygraph(tgraph2) ; 

  y = scoregraph(tgraph1) ;      // score graph        
  printf("score: %9.3f\n", y) ;

 }

 gg = tgraph1 ;
 printf("tree:\n") ;
 printgraph(gg) ;

 e = gg -> nedge ;
 
 ZALLOC(pp, e, int) ;
 ZALLOC(p1, e, int) ;
 ZALLOC(p2, e, int) ;
 for (a=0; a<numleaf ; a++) {
  for (b=a+1; b<numleaf ; b++) {
    epath(gg, pp, a, b) ;
    printf("%3d %3d: ", a,b) ; 
    printimat(pp, 1, e) ;
  }
 }

 v = 0 ;
 ZALLOC(v2epair, e*e, int) ;
 ZALLOC(epair2v, e*e, int) ;
 for (a=0; a<e ; a++) {  
    b = a ; 
    k = a*e+b ; 
    epair2v[k] = v ;  
    v2epair[v] = k ;
    ++v ;
 }
 for (a=0; a<e ; a++) {  
   for (b=a+1; b<e ; b++) {  
    k = a*e+b ; 
    epair2v[k] = v ;  
    v2epair[v] = k ;
    k = b*e+a ; 
    epair2v[k] = v ;  
    ++v ;
   }
 }
 numv = v ;
 maxeq += numv ;
 ZALLOC(abcd, maxeq, int *) ;
 for (i=0; i<maxeq; i++) { 
  ZALLOC(abcd[i], 4, int) ;
 }
 ZALLOC(vmask, numv, int) ;
 ZALLOC(vm, numv, int) ; // working version

 ivclear(vmask, 1, e) ;
  
 printf("maxeq: %d  numv: %d", maxeq, numv) ;
 printf("\n") ;

 ZALLOC(eq, maxeq*numv, double) ;
 ZALLOC(eqx, maxeq*numv, double) ;
 ZALLOC(rhs, maxeq, double) ;
 ZALLOC(rhsx, maxeq, double) ;
 ZALLOC(wt, maxeq, double) ;

 ne = m2 = m3 = m4 =0 ;
 for (a=0; a<numleaf ; a++) {  
   for (b=a+1; b<numleaf ; b++) {  
    k = addr4(0,0,a,b,numleaf) ;
    wabcd(abcd[ne], a, b, 99, 99) ;
    rhs[ne] = sc2[k] ; 
    wt[ne]  = serr2[k] ; 
    epath(gg, pp, a, b) ;
    loadee(eq+ne*numv, pp, pp,  e) ;
    ++ne ;
    ++m2 ; 
   }
 }
 for (a=0; a<numleaf ; a++) {  
   for (b=0; b<numleaf ; b++) {  
     for (c=b+1; c<numleaf ; c++) {  
      if (a==b) continue ;  
      if (a==c) continue ;  
      k = addr4(0,a,b,c, numleaf) ;               
      wabcd(abcd[ne], a, b, c, 99) ;
      rhs[ne] = sc3[k] ; 
      wt[ne] = serr3[k] ; 
      epath(gg, p1, a, b) ;
      epath(gg, p2, a, c) ;
      loadee(eq+ne*numv, p1, p2,  e) ;
      ++ne ;
      ++m3 ; 
     }
   }
 }
  for (a=0; a < numleaf; a++)  {  
   for (b=a+1; b < numleaf; b++)  {  
    for (c=b+1; c < numleaf; c++)  {  
     for (d=c+1; d < numleaf; d++)  {  

      k = addr4(a,b,c,d,numleaf) ;                    
      wabcd(abcd[ne], a, b, c, d) ;
      rhs[ne] = sc4[k] ; 
      wt[ne] = serr4[k] ; 
      epath(gg, p1, a, b) ;
      epath(gg, p2, c, d) ;
      loadee(eq+ne*numv, p1, p2,  e) ;
      ++ne ;
      ++m4 ; 

      k = addr4(a,c,b,d,numleaf) ;                    
      wabcd(abcd[ne], a, c, b, d) ;
      rhs[ne] = sc4[k] ; 
      wt[ne] = serr4[k] ; 
      epath(gg, p1, a, c) ;
      epath(gg, p2, b, d) ;
      loadee(eq+ne*numv, p1, p2,  e) ;
      ++ne ;
      ++m4 ; 

      k = addr4(a,d,b,c,numleaf) ;                    
      wabcd(abcd[ne], a, d, b, c) ;
      rhs[ne] = sc4[k] ; 
      wt[ne] = serr4[k] ; 
      epath(gg, p1, a, d) ;
      epath(gg, p2, b, c) ;
      loadee(eq+ne*numv, p1, p2,  e) ;
      ++ne ;
      ++m4 ; 

    }
   }
  }
 }
 ne = m2 ; 
 printf("ne: %d\n", ne) ;

 printmatw(eq, 3, numv, 15) ;
 printf("\n\n") ;
 printmatw(rhs, 1, ne, 5) ;
   
   nvv = sqzx(eq, eqx, rhs,  wt, rhsx, m2+m3+m4, numv, vmask) ;

   printmatw(eqx, ne, nvv, nvv) ;
   printf("\n\n") ;
   printmatw(rhsx, 1, ne, 5) ;
   
 ZALLOC(ans, numv, double) ; 
 ZALLOC(rr, maxeq, double) ; 
 regressit(ans, eqx, rhsx, ne,  nvv)  ;
 for (j=0; j<nvv; ++j) {  
  k=v2epair[j] ;
  a = k/e ; 
  k=v2epair[j] ;
  a = k/e ; 
  b = k%e ; 
  printf("ans: %3d %3d %9.3f", a,b, ans[j]) ;
  nodes(gg, &t1, &t2, a) ;  printf(" %3d %3d", t1, t2) ;
  nodes(gg, &t1, &t2, b) ;  printf("   %3d %3d", t1, t2) ;
  printf("\n") ;
 }
 for (i=0; i<m2+m3+m4; i++) { 
  y1 = vdot(eqx+i*nvv, ans, nvv) ; 
  rr[i] = y2 = rhsx[i]-y1 ;  
  printf("rhs res: %3d  %9.3f %9.3f   %9.3f", i, rhs[i], y2*wt[i], y2) ;
  pabcd(abcd[i]) ;
 }
 y = asum2(rr, ne) ; y = sqrt(y) ;
 printf("score: %9.3f\n", y) ;
 neq = m2 + m3 +m4 ;  

 y = asum2(rr, neq) ; y = sqrt(y) ;
 printf("gscore (edge only): %9.3f\n", y) ;

 yscthresh = 0.5*(sqrt(neq)) ;

 for (iter=1; iter<=numiter; ++iter) {  
  
  nvv = sqzx(eq, eqx, rhs,  wt, rhsx, neq, numv, vmask) ;
  regressit(ans, eqx, rhsx, neq,  nvv)  ;
  ysc = rscor(rr, ans, eqx, rhsx, neq, nvv) ;
  printf("iter start: %9.3f\n", ysc) ;
  if (ysc < yscthresh) break ;
  for (i=0; i<neq; i++) { 
   y2 = rr[i] ;
   printf("iter %d: rhs res: %3d  %9.3f %9.3f   %9.3f", iter, i, rhs[i], y2*wt[i], y2) ;
   pabcd(abcd[i]) ;
  }
  yscmin = 1.0e10 ;
  vbest = 0 ;
  for (v=0; v<numv; ++v)  {  
   if (vmask[v] == 1) continue ;
   copyiarr(vmask, vm, numv) ; 
   vm[v] = 1 ;

   fflush(stdout) ;
   nvv = sqzx(eq, eqx, rhs,  wt, rhsx, neq, numv, vm) ;
   addfudge(eqx, rhsx, neq, nvv) ;
   regressit(ans, eqx, rhsx, neq+nvv, nvv)  ;
   ysc = rscor(rr, ans, eqx, rhsx, neq, nvv) ;
   if (ysc<yscmin) {  
    yscmin = ysc ; 
    vbest = v ;
   }
  }
  printf("iter: %d gscore: %9.3f", iter, yscmin) ;  

  k=v2epair[vbest] ;
  a = k/e ; 
  b = k%e ; 
  printf(" vbest:" ) ;

  nodes(gg, &t1, &t2, a) ;  printf(" %3d %3d", t1, t2) ;
  nodes(gg, &t1, &t2, b) ;  printf("   %3d %3d", t1, t2) ;

  printf("\n") ; 
  vmask[vbest] = 1 ;
 }
 printf("final:\n") ;
 printgraph(gg) ;
 nvv = sqzx(eq, eqx, rhs,  wt, rhsx, neq, numv, vmask) ;
 addfudge(eqx, rhsx, neq, nvv) ;
 regressit(ans, eqx, rhsx, neq+nvv,  nvv)  ;
 ysc = rscor(rr, ans, eqx, rhsx, neq, nvv) ;

 oldgg = tgraph3 ;
 destroygraph(oldgg) ; 
 copygraph(gg, oldgg) ;
 ZALLOC(oldgg -> fst, e, double) ;
 t = 0 ;
 for (j=0; j<numv; ++j) {  
  if (vmask[j] == 0) continue ;
  k=v2epair[j] ;
  a = k/e ; 
  b = k%e ; 
  printf("fans: %3d %3d %12.6f", a,b, ans[t]) ;
  nodes(gg, &t1, &t2, a) ;  printf(" %3d %3d", t1, t2) ;
  nodes(gg, &t1, &t2, b) ;  printf("   %3d %3d", t1, t2) ;
  x = edgenum(oldgg, t1, t2) ;
  if (a==b) oldgg -> fst[x] = ans[t] ;
  printf("\n zz2 %d %d %d %9.3f\n", t1, t2, x, oldgg -> fst[x]) ;
  ++t ;
  printf("\n") ;
 }

 for (i=0; i<neq; i++) { 
  y1 = vdot(eqx+i*nvv, ans, nvv) ; 
  rr[i] = y2 = rhsx[i]-y1 ;  
  printf("rhs res: %3d  %9.3f %9.3f   %9.3f", i, rhs[i], y2*wt[i], y2) ;
  pabcd(abcd[i]) ;
 }

 printf("attempt to place root:\n") ;

 ZALLOC(xmax, e, double) ;
 ZALLOC(xmin, e, double) ;

 for (k=0; k<e; ++k)  {  
// assume root is on edge k   
  ymin = 1000.0; ymax = -1000.0 ;

  for (a=0; a < numleaf; a++) {  
   for (b=a+1; b < numleaf; b++) { 

    epath(gg, pp, a, b) ;
    if (pp[k] == 0) continue ;

    y = hest[a*numleaf+b] ;
    ymax = MAX(y, ymax) ;
    ymin = MIN(y, ymin) ;

   }
  }
   xmin[k] = ymin ; 
   xmax[k] = ymax ;
 }

 for (k=0; k<e; ++k) {  
  printf("hroot: %3d  %9.3f %9.3f   %9.3f\n", k, xmin[k], xmax[k], xmax[k]-xmin[k]) ;
 }
 vvm(xmax, xmax, xmin, e) ;

 vlmaxmin(xmax, e, NULL, &k) ;
 printf("best edge: %d", k) ;  
 
 nodes(gg, &t1, &t2, k) ; 
 printf("  %3d %3d\n", t1, t2) ;

// switch to rooted tree format


 xint = 100 ;
 x  = edgenum(oldgg, t1, t2) ;
 topfst  = oldgg -> fst[x] ;
 printf("yy topfst: %d %d %9.3f\n", t1, t2, topfst) ;
 addnode (tgraph2, gg, k, xint) ;  
 destroygraph(gg) ;
 gg = tgraph2 ;

 ZALLOC(rnode, 1, NODE) ;
 wipenode(rnode) ;
 mkrtree(rnode, gg, xint, -1) ;
 rnode -> isroot = YES ;

 ZALLOC(znode, gg -> nnode, NODE *) ;
 nznode = listnode(znode, rnode) ;

 ZALLOC(x1list, nznode, int) ;
 ZALLOC(x2list, nznode, int) ;

 for (k=0; k<nznode; k++) {  
  tnode = znode[k] ;
  if (tnode -> isleaf) {  
   a = tnode -> glabel ;
   printf("yy: %d %d\n", k, a) ;
   tnode -> hetleaves = hest[a*numleaf+a] ;
  } 
  else {  

   t = 0 ;  
   x1 = listleaves( (NODE *) tnode -> left, x1list) ;
   x2 = listleaves( (NODE *) tnode -> right, x2list) ;

   for (a=0; a< x1; a++) {  
    for (b=0; b< x2; b++) {  
     c = x1list[a] ;
     d = x2list[b] ;
     ww[t] = hest[c*numleaf+d] ;
     ++t ;
    }
   }
   y = median(ww, t) ;
   printf("qq: %d ", tnode -> glabel) ;
   printmat(ww, 1, t) ;
   printf("  %9.3f\n", y) ; 
   printf("\n") ;
   tnode -> hetleaves = y ;
  }

 }

 for (k=0; k<nznode; k++) {  
  tnode = znode[k] ;
  printf(" %3d:  %9.3f\n", tnode -> glabel, tnode -> hetleaves) ;
 }

 printf("maxt calculation\n") ;
 calctim(rnode, YES)  ;  
 printtim(rnode) ;

 printf("##end of run\n") ;

}
void calctim(NODE *rnode, int mode) 
{
   NODE *anode, *bnode ;
   int alabel, blabel, rlabel ;
   int k ;
   double y ;

   anode = (NODE *) rnode -> left ;
   bnode = (NODE *) rnode -> right ;  
   if (mode == YES)  {  
    if (rnode -> glabel == 100) {  
     alabel = anode -> glabel ;
     blabel = bnode -> glabel ;
     k = edgenum(oldgg, alabel, blabel) ;
     if (k < 0) fatalx("bug\n") ;
     y = oldgg -> fst[k] ;
     anode -> fst = bnode -> fst = 0.5 * y ;
     rnode -> fst = 0.0 ; 
     calctim(anode, mode) ;
     calctim(bnode, mode) ;
     return ;
    }
    if (rnode -> isleaf) return ;
    rlabel = rnode -> glabel ;
    alabel = anode -> glabel ;
    k = edgenum(oldgg, alabel, rlabel) ;
     if (k < 0) fatalx("bug\n") ;
    anode -> fst = oldgg -> fst[k] ;
    printf("zz1 %d %9.3f\n", alabel, anode -> fst) ;
    blabel = bnode -> glabel ;
    k = edgenum(oldgg, blabel, rlabel) ;
     if (k < 0) fatalx("bug\n") ;
    bnode -> fst = oldgg -> fst[k] ;
    printf("zz1 %d %9.3f\n", blabel, bnode -> fst) ;
    calctim(anode, mode) ;
    calctim(bnode, mode) ;
   }
}
void printtim(NODE *rnode) 
{
   NODE *anode, *bnode ;

   if (rnode -> isleaf) return ;  
   anode = (NODE *) rnode -> left ;
   bnode = (NODE *) rnode -> right ;  
   printf("fst: %3d %3d %9.3f" ,  rnode -> glabel, anode -> glabel, anode -> fst) ;
   printf(" %3d %3d %9.3f" ,  rnode -> glabel, bnode -> glabel, bnode -> fst) ;
   printf("\n") ;

   printtim(anode) ; 
   printtim(bnode) ;
 
   
}
void pabcd(int *bb)  
{
 int i, x ;

 for (i=0; i<4; i++)  { 
  x = bb[i] ; 
  if (x==99) continue ; 
  printf(" %3d", x) ;
 }
 printf("\n") ;
}
double
rscor(double *rrx, double *ans, double *eqx, double *rhsx, int neq, int nvv) 
{

 int i; 
 double y, y1 ;
 double *rr ;

 ZALLOC(rr, neq, double) ;
 for (i=0; i<neq; i++) { 
  y1 = vdot(eqx+i*nvv, ans, nvv) ; 
  rr[i] =  rhsx[i]-y1 ;  
 }
 y = asum2(rr, neq) ; y = sqrt(y) ;

 if (rrx != NULL) copyarr(rr, rrx, neq) ;

 free(rr) ;
 return y ;

}


void wabcd(int *bb, int a, int b, int c, int d) 
{

 bb[0] = a; 
 bb[1] = b; 
 bb[2] = c; 
 bb[3] = d; 
}

void printgraph(GRAPH *gg) 
{

 int i, j, n, e ;

  printf("## nnode: %d\n", gg -> nnode) ;
  printf("## nedge: %d\n", gg -> nedge) ;

  n = gg -> nnode ; 
  e = gg -> nedge ;
  for (i=0; i<n; i++) {  
    printf("%5d  ", gg -> label[i]) ;
    printf("%2d:  ", gg -> isleaf[i]) ;
    for (j=0; j<e; j++) {  
     printf("%3d ", gg -> matrix[i*e+j]) ;
    }  
    printf("\n") ;
  }
  printf("\n\n") ;
}
 
int  nodenum(GRAPH *gg, int label) 
{
 int i, n ;

 n = gg -> nnode ;
 for (i=0; i<n ; ++i)  {  
  if (gg -> label[i] == label) return i ; 
 }
 return -1 ;
}

int epath(GRAPH *gg, int *ep, int lab1, int lab2) 
{
  int *z1, *z2 , n, e, t, i ; 

  e = gg -> nedge ;
  ZALLOC(z1, e, int) ;
  ZALLOC(z2, e, int) ;

  t = nodenum(gg, lab1) ;  
  proot(gg, t, z1) ;

  t = nodenum(gg, lab2) ;  
  proot(gg, t, z2) ;

  for (i=0; i<e; i++) {  
   ep[i] = (z1[i] - z2[i])  ;
  }
  free(z1) ; 
  free(z2) ;
}

void
proot(GRAPH *gg, int t, int *z1) 
//  path back to root from node t
{

  int e, tt, k ;
  int niter = 0 ;  

  if (t<0) fatalx("logic bug\n") ;
 
  e = gg -> nedge ;
  ivzero(z1, e) ;
  tt = t ;
  for (;;) {  
   ++niter ;  
   if (niter> (2*e)) fatalx("looping\n") ;
   if (gg -> root == tt) return ;
   for (k=0; k<e; ++k)  {  
    if (gg ->matrix[tt*e+k] == -1) {  
      tt = othernode(gg, tt, k) ;
      z1[k] = 1 ;
    }
   }
  }
}
int othernode(GRAPH *gg, int t, int k) 
// edge k node t.  What is other node?  
{

 int i, n, e ;

 n = gg -> nnode ;
 e = gg -> nedge ;

 for (i=0; i<n; i++) {  
  if (i==t) continue ; 
  if (gg -> matrix[i*e+k] != 0) return i ;
 }

 fatalx("othernode bug\n") ;

}

void nodes(GRAPH *gg, int *t1, int *t2, int k) 
// labels at ends of edge.  t1 is upstream
{

 int i, n, e, x, x0, x1 ;

 n = gg -> nnode ;
 e = gg -> nedge ;
 
 *t1 = *t2 = -1 ;
 x0 = x1 = -1 ;

 for (i=0; i<n; i++) {  
  x = gg -> matrix[i*e+k] ; 
  if (x>0) x0 = i ;
  if (x<0) x1 = i ;
 }
 

 if ((x0>=0) && (x1>=0)) {
   *t1 = gg -> label[x0] ;
   *t2 = gg -> label[x1] ;
   return ; 
 }

 fatalx("nodes bug\n") ;

}

int edgenum (GRAPH *gpt, int lab1, int lab2) 
// return edge number joining nodes with label lab1, lab2
{

 int k, n, e, a1= 0, a2 = 0, t1, t2, i ;
 e = gpt -> nedge ;
 n = gpt -> nnode ;
 t1 = t2 = -1 ;
 for (i=0; i<n ; ++i)  {  
  if (gpt -> label[i] == lab1) t1 = i ;
  if (gpt -> label[i] == lab2) t2 = i ;
 }

 if (t1<0) return -1 ;
 if (t2<0) return -1 ;

 e = gpt -> nedge ;
 for (k=0; k<e ; ++k)  {  
  a1 = a2 = NO ;
  if (gpt -> matrix[t1*e+k] != 0)  a1 = YES ;                      
  if (gpt -> matrix[t2*e+k] != 0)  a2 = YES ; 
  if (a1 && a2)  break ;
 }
 if (k>=e) return -1 ;
 
 return k ;

}

void trivgraph(GRAPH *gg, int label) 
{
  GRAPH *xx ;

  xx = gg ;  
  xx -> nnode = 1 ; 
  xx -> nedge = 0 ;
  ZALLOC(xx -> isleaf, 1, int) ;  
  ZALLOC(xx -> label, 1, int) ;  
  xx -> isleaf[0] = YES ;
  xx -> label[0] = label ;
  xx -> root = 0 ;  
}
void
addedge(GRAPH *gg, GRAPH *g1, int oldnode, int newlabel, int isleaf)   
{
  GRAPH *xx, *x ;
  int n, e, i, j ;

  x = g1 ;
  xx = gg ;  
  n = xx -> nnode = x -> nnode + 1 ; 
  e = xx -> nedge = x -> nedge + 1 ; 
  ZALLOC(xx -> isleaf, n, int) ;  
  copyiarr(x -> isleaf, xx -> isleaf, n-1) ; xx->isleaf[n-1] = isleaf ;
  ZALLOC(xx -> label, n, int) ;  
  copyiarr(x -> label, xx -> label, n-1) ; xx->label[n-1] = newlabel ;
  xx -> root = x -> root ;  
  ZALLOC(xx -> matrix, n*e, int) ;
  for (i=0; i<n-1; i++)  {  
   for (j=0; j<e-1; j++)  {  
    xx ->matrix[i*e+j] = x->matrix[i*(e-1)+j] ;
   }
  }
  i = oldnode ; j = e-1 ; 
  xx -> matrix[i*e+j] = 1 ; // ancestor
  i = n-1 ; j = e-1 ; 
  xx -> matrix[i*e+j] = -1 ; //child
  
  printf("zzaddedge %d %d %d\n", oldnode, newlabel, isleaf) ;

  printgraph(x)  ; 
  printgraph(xx) ;
}

void
copygraph(GRAPH *g1, GRAPH *g2) 
{
  GRAPH *xx, *x ;
  int n, e, i, j ;

  xx = g2 ;  
  x = g1 ;
  n = xx -> nnode = x -> nnode  ; 
  e = xx -> nedge = x -> nedge  ; 

  ZALLOC(xx -> isleaf, n, int) ;  
  copyiarr(x -> isleaf, xx -> isleaf, n) ;
  ZALLOC(xx -> label, n, int) ;  
  copyiarr(x -> label, xx -> label, n) ; 
  xx -> root = x -> root ;  
  ZALLOC(xx -> matrix, n*e, int) ;
  copyiarr(x -> matrix, xx -> matrix, n*e) ; 
}

void wipegrph(int n) 
{ 

 int k ;  

 for (k=0; k<n; ++k) {
  destroygraph(grph[k]) ;
 }



}
void destroygraph(GRAPH *g1) 
// does NOT destroy g1 but space allocated
{
  GRAPH *xx ;

  xx = g1 ; 
  if (xx -> isleaf != NULL) free(xx->isleaf) ;
  if (xx -> label != NULL) free(xx->label) ;
  if (xx -> matrix != NULL) free(xx->matrix) ;

   xx -> isleaf = xx -> label = xx -> matrix = NULL  ;

}
void
addnode(GRAPH *gg, GRAPH *g1, int oldedge, int newlabel)   
{

  GRAPH *xx, *x ;
  int n, e, i, j, newn, newe ;


  x = g1 ;
  xx = gg ;  
  n = xx -> nnode = x -> nnode + 1 ; 
  e = xx -> nedge = x -> nedge + 1 ; 
  ZALLOC(xx -> isleaf, n, int) ;  
  copyiarr(x -> isleaf, xx -> isleaf, n-1) ; xx->isleaf[n-1] = NO ;
  ZALLOC(xx -> label, n, int) ;  
  copyiarr(x -> label, xx -> label, n-1) ; xx->label[n-1] = newlabel ;
  xx -> root = x -> root ;  
  ZALLOC(xx -> matrix, n*e, int) ;

  for (i=0; i<n-1; i++)  {  
   for (j=0; j<e-1; j++)  {  
    if (j==oldedge) continue ;
    xx ->matrix[i*e+j] = x->matrix[i*(e-1)+j] ;
   }
  }

  j = oldedge ;             
  newn = n-1 ;
  newe = e-1 ;
  xx -> matrix[newn*e+newe] = 1 ;
  xx -> matrix[newn*e+j] = -1 ;
  for (i=0; i<n-1; i++)  {  
   if (x -> matrix[i*(e-1)+j] == 1) {  
    xx -> matrix[i*e+j] = 1 ;
   }
   if (x -> matrix[i*(e-1)+j] == -1) {  
    xx -> matrix[i*e+newe] = -1 ;
   }
  }

}

#define MAXSTR 512
#define MAXFF  20

int getnumleaf(char *fname) 
{

 char line[MAXSTR+1], c ;
 char *spt[MAXFF], *sx ;
 int nsplit, n = -1 ;
 int kret, a, b  ;
 

 FILE *fff ;
 openit(fname, &fff, "r") ;
 line[MAXSTR] = '\0' ;
 while (fgets(line, MAXSTR, fff) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   if (nsplit == 0) continue ;
   sx = spt[0] ;
   kret =  strcmp(sx, "score2") ;
   if (kret != 0) { 
    freeup(spt, nsplit) ; 
    continue ;
   }
   a = atoi(spt[1]) ;
   b = atoi(spt[2]) ;
   n = MAX(n, a) ; 
   n = MAX(n, b) ;

   freeup(spt, nsplit) ;
   continue ;
 }


 fclose(fff) ;
 return (n+1) ;



}


int loadhets(char *fname, double *hets, double *hsig, int numleaf) 
{

 char line[MAXSTR+1], c ;
 char *spt[MAXFF], *sx ;
 int nsplit, num=0 ;
 int kret, a, b  ;
 

 FILE *fff ;
 openit(fname, &fff, "r") ;
 line[MAXSTR] = '\0' ;
 
 vzero(hets, numleaf*numleaf) ;
 vclear(hsig, 1.0, numleaf*numleaf) ;
 while (fgets(line, MAXSTR, fff) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   if (nsplit == 0) continue ;
   sx = spt[0] ;
   kret =  strcmp(sx, "hetrate:") ;
   if (kret != 0) { 
    freeup(spt, nsplit) ; 
    continue ;
   }
   a = atoi(spt[1]) ;
   b = atoi(spt[2]) ;
   hets[b*numleaf+a] =  hets[a*numleaf+b] = atof(spt[5]) ;
   hsig[b*numleaf+a] =  hsig[a*numleaf+b] = atof(spt[6]) ;
   ++num ;

   freeup(spt, nsplit) ;
   continue ;
 }


 fclose(fff) ;
 return num ;



}

int loadsc2(char *fname, double *sc, double *serr, int numleaf) 
{

 char line[MAXSTR+1], c ;
 char *spt[MAXFF], *sx ;
 int nsplit, num=0 ;
 int kret, a, b  ;
 

 FILE *fff ;
 openit(fname, &fff, "r") ;
 line[MAXSTR] = '\0' ;
 while (fgets(line, MAXSTR, fff) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   if (nsplit == 0) continue ;
   sx = spt[0] ;
   kret =  strcmp(sx, "score2") ;
   if (kret != 0) { 
    freeup(spt, nsplit) ; 
    continue ;
   }
   a = atoi(spt[1]) ;
   b = atoi(spt[2]) ;
   sc[a*numleaf+b] = atof(spt[3]) ;
   serr[a*numleaf+b] = atof(spt[4]) ;
   ++num ;

   freeup(spt, nsplit) ;
   continue ;
 }


 fclose(fff) ;
 return num ;



}


int loadsc3(char *fname, double *sc, double *serr, int numleaf) 
{

 char line[MAXSTR+1] ;
 char *spt[MAXFF], *sx ;
 int nsplit, num=0 ;
 int kret, a, b, c  ;
 int n1, n2 ;
 

 FILE *fff ;
 openit(fname, &fff, "r") ;
 n1 = numleaf ; 
 n2 = n1*numleaf ;
 line[MAXSTR] = '\0' ;
 while (fgets(line, MAXSTR, fff) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   if (nsplit == 0) continue ;
   sx = spt[0] ;
   kret =  strcmp(sx, "score3") ;
   if (kret != 0) { 
    freeup(spt, nsplit) ; 
    continue ;
   }
   a = atoi(spt[1]) ;
   b = atoi(spt[2]) ;
   c = atoi(spt[3]) ;
   sc[a*n2+b*n1+c] = atof(spt[4]) ;
   serr[a*n2+b*n1+c] = atof(spt[5]) ;
   ++num ;

   freeup(spt, nsplit) ;
   continue ;
 }

 fclose(fff) ;
 return num ;

}

int loadsc4(char *fname, double *sc, double *serr, int numleaf) 
{

 char line[MAXSTR+1] ;
 char *spt[MAXFF], *sx ;
 int nsplit, num=0 ;
 int kret, a, b, c, d  ;
 int n1, n2, n3 ;
 

 FILE *fff ;
 openit(fname, &fff, "r") ;
 n1 = numleaf ; 
 n2 = n1*numleaf ;
 n3 = n2*numleaf ;
 line[MAXSTR] = '\0' ;
 while (fgets(line, MAXSTR, fff) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   if (nsplit == 0) continue ;
   sx = spt[0] ;
   kret =  strcmp(sx, "score4") ;
   if (kret != 0) { 
    freeup(spt, nsplit) ; 
    continue ;
   }
   a = atoi(spt[1]) ;
   b = atoi(spt[2]) ;
   c = atoi(spt[3]) ;
   d = atoi(spt[4]) ;

   sc[a*n3+b*n2+c*n1+d] = atof(spt[5]) ;
   serr[a*n3+b*n2+c*n1+d] = atof(spt[6]) ;
   ++num ;

   freeup(spt, nsplit) ;
   continue ;
 }


 fclose(fff) ;
 return num ;


}
int
sqzx(double *eq, double *eqx, double *rhs,  double *wt, double *rhsx, 
  int ne, int numv, int *vmask) 
// squeeze out equations 
{

  int i, j, x, vv ;
  double *ee, *ff ;
  static int ncall = 0 ;
  ++ncall ;

  vv = intsum(vmask, numv) ;  

  vzero(eqx, ne*vv) ;

  for (i=0 ; i<ne; i++)  {  

   rhsx[i] = rhs[i]/wt[i] ;
   ee = eq + i*numv ;  
   ff = eqx + i*vv ;  
   

   x = 0 ;  
   for (j=0; j<numv; ++j)  {  
    if (vmask[j] == 0 ) continue ;
    ff[x] = ee[j]/wt[i] ;  
    if (ncall < -1)  {
     printf("ww %d %d %d %9.3f\n", i, j, x, ff[x]) ;
    }
    ++x ;
   }
  }
  return vv ;

}
void
loadee(double *ee, int *p1, int *p2,  int e) 
// p1 p2 are edge lists. work out product 
{

  int i, j, k, v ;

  vzero(ee, e) ;
  for (i=0; i<e; i++) {  
   if (p1[i] == 0) continue ;
    for (j=0; j<e; j++) {  
     if (p2[j] == 0) continue ;
     k = i*e+j ;
     v = epair2v[k] ;  
     ee[v] += p1[i]*p2[j] ;
    }
  }
}
int
addr4(int a,int b,int c,int d,int n)                            
{
  int t ; 
  t = a ; 
  t = t*n+b ;
  t = t*n+c ;
  t = t*n+d ;

  return t ; 

}
void
addfudge(double *eqx, double *rhsx, int neq, int nvv) 
// add small amount of prior to force matrix non-singular  
{
   double *ee ;  
   int i ;  

   for (i=0; i<nvv; i++)  {  
    rhsx[i+neq] = 0.0 ;  
    ee = eqx + (i+neq)*nvv ;
    vzero(ee, nvv) ;
    ee[i] = .0001  ;
   }
}
void
getquartet(double *sc4, double *serr4, int *pa, int *pb, int *pc, int *pd, int numleaf) 
{

  int a, b, c, d ; 
  int xa, xb, xc, xd ;
  int ind[3], k ; 
  double ysc, yscmax = -1.0e10 ;
  double xx[3], xx2[3] ;

  for (a=0; a < numleaf; a++)  {  
   for (b=a+1; b < numleaf; b++)  {  
    for (c=b+1; c < numleaf; c++)  {  
     for (d=c+1; d < numleaf; d++)  {  

      k = addr4(a,b,c,d,numleaf) ;  
      xx[0] = sc4[k]/serr4[k] ;

      k = addr4(a,c,b,d,numleaf) ;  
      xx[1] = sc4[k]/serr4[k] ;

      k = addr4(a,d,b,c,numleaf) ;  
      xx[2] = sc4[k]/serr4[k] ;

      vvt(xx, xx, xx, 3) ;

      copyarr(xx, xx2, 3) ;
      sortit(xx2, ind, 3) ;  

      ysc = xx2[1] - xx2[0] ;
      ysc -= 0.5*xx2[0] ;  // fudge factor to punish large xx2[0] 

      if (ysc>yscmax)  {
       yscmax = ysc ; 
       k = ind[0] ;  
       switch (k)  { 
     
        case 0:  
         xa = a; xb = b ; xc = c ; xd = d ;
         break ;

        case 1:  
         xa = a; xb = c ; xc = b ; xd = d ;
         break ;

        case 2:  
         xa = a; xb = d ; xc = b ; xd = c ;
         break ;

       }

      }

     }
    }
   }
  }


  *pa = xa ; *pb = xb ; *pc = xc ; *pd = xd ;  
  printf("zz %d %d %d %d %9.3f\n", xa, xb, xc, xd, yscmax) ;

}
#define MAXNODE 50   
#define MAXEDGE 50 


double growtree(GRAPH *newgg, GRAPH *gg) 
{

 GRAPH *tgraph, *xgraph ;
 int *islabel ;
 int k, xint, j, kbest, jbest, xnode ; 
 double ysc, yscmin ;


 ZALLOC(tgraph, 1, GRAPH) ;
 ZALLOC(xgraph, 1, GRAPH) ;

 yscmin = 1.0e10 ;  
 kbest = jbest = -1 ;
 tgraph = grph[ngraph] ;

 xint = toplabel(gg) + 1 ;

 ZALLOC(islabel, maxlabel, int) ;
 setislab(islabel, gg) ;
 
 for (k=0; k< gg ->nedge; k++) {

   addnode (xgraph, gg, k, xint) ;  
   
   xnode = nodenum(xgraph, xint) ; 
   if (xnode<0) fatalx("logic bug\n") ;

   for (j=0; j<numleaf; j++)  {  
    if (islabel[j] == 1) continue ;
    printgraph(xgraph) ;
    addedge(tgraph, xgraph, xnode, j, YES)  ;   
   
    printgraph(tgraph) ;
    
    ysc = scoregraph(tgraph) ;  
    printf("zzgrow %d %d %9.3f\n", k, j, ysc) ;
    if (ysc<yscmin)  {
     yscmin = ysc ;  
     kbest = k ; 
     jbest = j ;
    }
    destroygraph(tgraph) ;
   }
   destroygraph(xgraph) ;

  }

  if (kbest>=0) {
   addnode(xgraph, gg, kbest, xint) ;  
   xnode = nodenum(xgraph, xint) ; 
   addedge(newgg, xgraph, xnode, jbest, YES) ;
   ysc = scoregraph(newgg) ;
  }
  else { 
   copygraph(gg, newgg) ;
   ysc = -1.0 ;
  }

  free(tgraph) ;
  free(xgraph) ;

  return ysc ;

}

int toplabel(GRAPH *gg) 
{
 
 int n, kmax ;

 n = gg -> nnode ;  
 ivmaxmin(gg->label, n, &kmax, NULL) ;


 return kmax ;


}

void setislab(int *islab, GRAPH *gg)
{

 int n, k, a ; 

 n = gg -> nnode ;  
 ivclear(islab, -1, maxlabel) ;

 for (a=0; a<n; a++) {
  if (gg->isleaf[a]) {
   k = gg -> label[a]  ;
   islab[k] = 1 ;
  }
 }

}

double scoregraph(GRAPH *gg)  
{

 int *islabel ; 
 int a, b, c, d, k, x, ne, n, e ;
 int *pp, *p1, *p2 ;
 double *eq, *eqx, *rhs, *rhsx, *wt, *ans ;
 int maxeq, numv, v ;
 int *vmask ;
 int neq, nvv ; 
 double ysc, ybase ;

 n = gg -> nnode ;  
 e = gg -> nedge ;

 ZALLOC(islabel, maxlabel, int) ;

 ZALLOC(pp, e, int) ;
 ZALLOC(p1, e, int) ;
 ZALLOC(p2, e, int) ;

 setislab(islabel, gg) ;

 ZALLOC(epair2v, e*e, int) ;
 ZALLOC(v2epair, e*e, int) ;
 v = 0 ;

 for (a=0; a<e ; a++) {  
    b = a ; 
    k = a*e+b ; 
    epair2v[k] = v ;  
    v2epair[v] = k ;
    ++v ;
 }
 for (a=0; a<e ; a++) {  
   for (b=a+1; b<e ; b++) {  
    k = a*e+b ; 
    epair2v[k] = v ;  
    v2epair[v] = k ;
    k = b*e+a ; 
    epair2v[k] = v ;  
    ++v ;
   }
 }
 numv = v ;
 maxeq = n*n*n*n ; 
 maxeq += numv ;

 ZALLOC(eq, maxeq*numv, double) ;
 ZALLOC(eqx, maxeq*numv, double) ;
 ZALLOC(rhs, maxeq, double) ;
 ZALLOC(rhsx, maxeq, double) ;
 ZALLOC(wt, maxeq, double) ;

 ybase = missxx(islabel) ;
 printf("zz1 %9.3f\n", ybase) ;
 ne = 0 ;
 for (a=0; a<numleaf ; a++) {  
   for (b=a+1; b<numleaf ; b++) {  
    if (islabel[a] <0) continue ;
    if (islabel[b] <0) continue ;
    k = addr4(0,0,a,b,numleaf) ;
    rhs[ne] = sc2[k] ; 
    wt[ne]  = serr2[k] ; 
    epath(gg, pp, a, b) ;
    loadee(eq+ne*numv, pp, pp,  e) ;
    ++ne ;
   }
 }
 for (a=0; a<numleaf ; a++) {  
   for (b=0; b<numleaf ; b++) {  
     for (c=b+1; c<numleaf ; c++) {  
      if (a==b) continue ;  
      if (a==c) continue ;  
      if (islabel[a] <0) continue ;
      if (islabel[b] <0) continue ;
      if (islabel[c] <0) continue ;
      k = addr4(0,a,b,c, numleaf) ;               
      rhs[ne] = sc3[k] ; 
      wt[ne] = serr3[k] ; 
      epath(gg, p1, a, b) ;
      epath(gg, p2, a, c) ;
      loadee(eq+ne*numv, p1, p2,  e) ;
      ++ne ;
     }
   }
 }
  for (a=0; a < numleaf; a++)  {  
   for (b=a+1; b < numleaf; b++)  {  
    for (c=b+1; c < numleaf; c++)  {  
      for (d=c+1; d < numleaf; d++)  {  

      if (islabel[a] <0) continue ;
      if (islabel[b] <0) continue ;
      if (islabel[c] <0) continue ;
      if (islabel[d] <0) continue ;

      k = addr4(a,b,c,d,numleaf) ;                    
      rhs[ne] = sc4[k] ; 
      wt[ne] = serr4[k] ; 
      epath(gg, p1, a, b) ;
      epath(gg, p2, c, d) ;
      loadee(eq+ne*numv, p1, p2,  e) ;
      ++ne ;

      k = addr4(a,c,b,d,numleaf) ;                    
      rhs[ne] = sc4[k] ; 
      wt[ne] = serr4[k] ; 
      epath(gg, p1, a, c) ;
      epath(gg, p2, b, d) ;
      loadee(eq+ne*numv, p1, p2,  e) ;
      ++ne ;

      k = addr4(a,d,b,c,numleaf) ;                    
      rhs[ne] = sc4[k] ; 
      wt[ne] = serr4[k] ; 
      epath(gg, p1, a, d) ;
      epath(gg, p2, b, c) ;
      loadee(eq+ne*numv, p1, p2,  e) ;
      ++ne ;

    }
   }
  }
 }

 neq = ne ;
 ZALLOC(vmask, numv, int) ;
 ivclear(vmask, 1, e) ;

 nvv = sqzx(eq, eqx, rhs,  wt, rhsx, neq, numv, vmask) ;
 ZALLOC(ans, nvv, double) ;
 regressit(ans, eqx, rhsx, neq,  nvv)  ;
 ysc = rscor(NULL, ans, eqx, rhsx, neq, nvv) ;
 printf("gscore (edge pair): %9.3f\n", ysc) ;

 free(eq) ;
 free(eqx) ;
 free(rhs) ;
 free(rhsx) ; 
 free(wt) ;

 free(pp) ; 
 free(p1) ; 
 free(p2) ;

 free(epair2v) ;
 free(v2epair) ;

 free(vmask) ; 
 free(ans) ;

 free(islabel) ;

 return (ysc + ybase) ;

}

int klabs(int *islabel, int a, int b, int c, int d) 
{  
// returns 1 if any of a, b, c, d is missing

 if (islabel[a] != 1) return 1 ;
 if (islabel[b] != 1) return 1 ;
 if (islabel[c] != 1) return 1 ;
 if (islabel[d] != 1) return 1 ;

 return -1 ;


}

double missxx(int *islabel) 

{

 int a, b, c, d, k, kret ;
 double y, ysc ; 

 ysc = 0.0 ;

 for (a=0; a<numleaf ; a++) {  
   for (b=a+1; b<numleaf ; b++) {  
    kret = klabs(islabel, a, a, a, b) ;
    if (kret>0) {
     k = addr4(0,0,a,b,numleaf) ;
     y = sc2[k]/serr2[k] ; ysc += y*y ;
    }
   }
 }
 for (a=0; a<numleaf ; a++) {  
   for (b=0; b<numleaf ; b++) {  
     for (c=b+1; c<numleaf ; c++) {  
      if (a==b) continue ;  
      if (a==c) continue ;  
      kret = klabs(islabel, a, a, b, c) ;
      if (kret>0) {
       k = addr4(0,a,b,c, numleaf) ;               
       y = sc3[k]/serr3[k] ; ysc += y*y ;
      }
     }
   }
 }
  for (a=0; a < numleaf; a++)  {  
   for (b=a+1; b < numleaf; b++)  {  
    for (c=b+1; c < numleaf; c++)  {  
      for (d=c+1; d < numleaf; d++)  {  

      kret = klabs(islabel, a, b, c, d) ;
      if (kret <=0) break ;

      k = addr4(a,b,c,d,numleaf) ;                    
      y = sc4[k]/serr4[k] ; ysc += y*y ;

      k = addr4(a,c,b,d,numleaf) ;                    
      y = sc4[k]/serr4[k] ; ysc += y*y ;
      k = addr4(a,d,b,c,numleaf) ;                    

      y = sc4[k]/serr4[k] ; ysc += y*y ;

    }
   }
  }
 }

 return ysc ;
}


void readcommands(int argc, char **argv) 

{
  int i,haploid=0;
  char *parname = NULL ;
  phandle *ph ;
  char str[5000]  ;
  char *tempname ;
  int n ;

  while ((i = getopt (argc, argv, "i:p:vV")) != -1) {

    switch (i)
      {

      case 'i':
	infilename = strdup(optarg) ;
	break;

      case 'p':
	parname = strdup(optarg) ;
	break;

      case 'v':
	printf("version: %s\n", WVERSION) ; 
	break; 

      case 'V':
	verbose = YES ;
	break; 

      case '?':
	printf ("Usage: bad params.... \n") ;
	fatalx("bad params\n") ;
      }
  }

         
   if (parname == NULL) return ;
   ph = openpars(parname) ;
   dostrsub(ph) ;

   getstring(ph, "infilename:", &infilename) ;
   writepars(ph) ;
   closepars(ph) ;

}
double median(double *x, int n) 
{
  double *ww, y ;  
  int a ;

  ZALLOC(ww, n, double) ; 
  copyarr(x, ww, n) ; 

  sortit(ww, NULL, n) ;  
  a = n/2 ;  
  if (n == (2*a)) {  
   y = 0.5*(ww[a]+ww[a-1]) ;
  }
  else y =  ww[a] ;
  free(ww) ; 
  return y ;

}

void wipenode(NODE *rnode) { 
 rnode -> gnode = rnode -> glabel = -1 ; 
 rnode -> isleaf = YES ;
 rnode -> left = NULL ;
 rnode -> right = NULL ;
 rnode -> parent = NULL ;
}
void mkrtree(NODE *rnode, GRAPH *gg, int rlabel, int plabel) { 

  int a, t, s, e, n, k, tt, pp ; 
  int lableft = -1, labright = -1 ;
  NODE *tnode ;
     
  rnode -> glabel = rlabel ;  
  rnode -> isroot = NO ;  
  rnode -> isleaf = NO ;  


  if (rlabel == 100) rnode -> isroot = YES ;
  t = nodenum(gg, rlabel) ;
  rnode -> gnode = t ;
  printf("rtree:  %d  %d\n", rlabel, t) ;

  if (gg -> isleaf[t])  { 
   rnode -> isleaf = YES ;
   return ;
  }

  rnode -> isleaf = NO ;
  e = gg -> nedge ;  
  n = gg -> nnode ; 

  for (k=0; k<e; k++)  {  
   s = gg -> matrix[t*e+k] ;
   if (s == 0) continue ;
   tt = othernode(gg, t, k) ;
   pp = gg -> label[tt]  ;
   if (pp == plabel) continue ; 
   if ((s < 0) || (labright>0)) lableft = pp ;
   else  labright = pp ;
  }

  printf("rtlab:  %d  %d\n", lableft, labright) ;
  ZALLOC(tnode, 1, NODE) ;
  tnode -> parent = (struct NODE *) rnode ;
  mkrtree(tnode, gg, lableft, rlabel) ;
  rnode -> left = (struct NODE *) tnode ;

  ZALLOC(tnode, 1, NODE) ;
  tnode -> parent = (struct NODE *) rnode ;
  mkrtree(tnode, gg, labright, rlabel) ;
  rnode -> right = (struct NODE *) tnode ;
}
int
listleaves( NODE *tnode, int *xx) 
{
   int n ;
            
  n = 0 ;
 
  if (tnode -> isleaf == YES) {
   xx[0] = tnode -> glabel ;
   return 1 ;
  }
  
  n += listleaves( (NODE *) tnode -> left, xx+n) ;
  n += listleaves( (NODE *) tnode -> right, xx+n) ;

  return n ;

} 
int
listnode(NODE  **znode, NODE *tnode) 
{
   int n=0 ;
            
   znode[0] = tnode ;
   n = 1 ;
 
  if (tnode -> isleaf == YES) {
   return n ;
  }
  
  n += listnode( znode +n, (NODE *) tnode -> left  ); 
  n += listnode( znode +n, (NODE *) tnode -> right ); 

  return n ;

} 

